/*****************************************************************************
 *  Rohde&Schwarz CMW500 Wideband Radio Communication Tester instrument driver
 *  GSM Measurements
 *  Original Release: August 2008
 *  By: Jiri Kominek
 *
 *  Should you have any technical questions please contact the hotline of
 *  Rohde & Schwarz Vertriebs-GmbH Rohde & Schwarz Support Center
 *
 *  e-mail: CustomerSupport@rohde-schwarz.com
 *
 *  Modification History:
 *    see ChangeLog
 *
 *****************************************************************************/
#include "rscmwgg.h"
             
#include "time.h"
#if (defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__NT__))
    #define ANSI
#endif
#ifndef ANSI                /* UNIX compatible */
    #include <sys/timeb.h>
    #include <sys/types.h>
    #include <unistd.h>
    #define Sleep sleep
#else                       /* ANSI compatible */
        #include <windows.h>
#endif

/*****************************************************************************
 *- Value Definition and Range Tables ---------------------------------------*
 *****************************************************************************/
static ViString mevMviewBurstTypeArr[] = {"OFF","ANY","GMSK","EPSK","ACC",VI_NULL};
static ViString mevPatternTypeArr[] = {"OFF","GMSK","EPSK",VI_NULL};
static ViString switchArr[] = {"OFF","ON",VI_NULL};
static ViString modulationSchemeArr[] = {"GMSK","EPSK",VI_NULL};
static ViString PVTSectionArr[] = {"REDG","UPAR","FEDG",VI_NULL};
static ViString resultTypeArr[] = {"CURR","AVER","MAX","MIN",VI_NULL};
static ViString slotLevelArr[] = {"OFF","FULL","A1","A2","A3","A4","A5","A6",
                                  "A7",VI_NULL};
static ViString burstTypeArr[] = {"GMSK","EPSK","SYNC","FCOR","DUMM",
                                  "ACC","BCCH","TCH",VI_NULL};
static ViString dataSourceArr[] = {"ALL0","ALL1","PR9","PR11","PR15","PR16",
                                   "PR20","PR21","PR23","ALT","BERP",VI_NULL};
static ViString testSequenceArr[] = {"STANDARD","TS0","TS1","TS2","TSC0","TSC1",
                                     "TSC2","TSC3","TSC4","TSC5","TSC6","TSC7",
                                     VI_NULL};
static ViString modSchemeArr[] = {"TFS","TFS2","SM5","SM52","SM9","SM92",
                                  VI_NULL};
static ViString listModeBurstTypeArr[] = {"OFF","GMSK","EPSK","ACC", VI_NULL};
static ViString RFConnectorArr[] = {"RF1C", "RF2C", "RF3C", "RF4C", "RF1O", "RF3O", "RFAC", "RFBC", "RFAO", VI_NULL};
static ViString RFConverterArr[] = {"", "TX1", "TX2", VI_NULL};


/*****************************************************************************
 *- Callback Declarations ---------------------------------------------------*
 *****************************************************************************/
ViStatus rscmwgg_CheckStatusCallback (ViSession io);
ViStatus rscmwgg_WaitForOPCCallback (ViSession io);
ViStatus _VI_FUNC rscmwgg_RsClose (ViSession vi);

/*****************************************************************************
 *- User-Callable Functions (Exportable Functions) --------------------------*
 *****************************************************************************/

/*****************************************************************************
 * Function: rscmwgg_init
 * Purpose:  VXIplug&play required function. Calls the
 *           rscmwgg_InitWithOptions function.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_init (ViRsrc resourceName, ViBoolean IDQuery,
                                 ViBoolean resetDevice, ViSession *newInstrSession)
{
    ViStatus    error = VI_SUCCESS;

    if (newInstrSession == VI_NULL)
        {
        (void) Rs_SetErrorInfo (VI_NULL, VI_FALSE, RS_ERROR_INVALID_PARAMETER,
                          VI_ERROR_PARAMETER4, "Null address for Instrument Handle");
        checkErr( RS_ERROR_INVALID_PARAMETER);
        }

    checkErr( rscmwgg_InitWithOptions (resourceName, IDQuery, resetDevice,
                                      "QueryInstrStatus=1", newInstrSession));

Error:
    return error;
}

/*****************************************************************************
 * Function: rscmwgg_InitWithOptions
 * Purpose:  This function creates a new RS session and calls the
 *           RsInit function.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_InitWithOptions(
    ViRsrc          resourceName,
    ViBoolean       IDQuery,
    ViBoolean       resetDevice,
    ViString        optionString,
    ViSession      *newInstrSession
)
{
    ViStatus                error = VI_SUCCESS;
    ViChar                  newResourceName[RS_MAX_MESSAGE_BUF_SIZE] = "";
    ViChar                  newOptionString[RS_MAX_MESSAGE_BUF_SIZE] = "";
    ViBoolean               isLogicalName = 0;
    ViSession               instrSession = 0, rmSession = 0;
    RsAttrPropertiesPtr     *attrList=NULL;
    RsRepCapPtr             repCapTable = NULL;

    if (newInstrSession == VI_NULL)
        {
        (void) Rs_SetErrorInfo (VI_NULL, VI_FALSE, RS_ERROR_INVALID_PARAMETER,
                         VI_ERROR_PARAMETER5, "Null address for Instrument Handle");
        checkErr( RS_ERROR_INVALID_PARAMETER);
        }

    /* Prepare resource manager session handle */
    checkErr (viOpenDefaultRM (&rmSession));

    /* Establish communication session with the instrument */

    viCheckErr( viOpen (rmSession, resourceName, VI_NULL, VI_NULL, &instrSession));
    
    /* Configure VISA Formatted I/O */
    viCheckErr( viSetAttribute (instrSession, VI_ATTR_TMO_VALUE, 5000 ));
    viCheckErr( viSetBuf (instrSession, (ViUInt16) (VI_READ_BUF | VI_WRITE_BUF), 4000));
    viCheckErr( viSetAttribute (instrSession, VI_ATTR_WR_BUF_OPER_MODE, VI_FLUSH_ON_ACCESS));
    viCheckErr( viSetAttribute (instrSession, VI_ATTR_RD_BUF_OPER_MODE, VI_FLUSH_ON_ACCESS));

    checkErr( Rs_GetInfoFromResourceName (resourceName,
                                          (ViString)optionString,
                                          newResourceName,
                                          newOptionString,
                                          &isLogicalName));

    ATTRLIST(attrList,rscmwgg); 
    checkErr( Rs_SpecificDriverNew (instrSession, "rscmwgg", newOptionString, attrList));
   
    /* Build repeated capability table */
    REPCAPTABLE(repCapTable,rscmwgg);
    checkErr( Rs_BuildRepCapTable (instrSession, repCapTable));


    checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_VISA_RM_SESSION, 0, &rmSession));

    /* With remote control via Ethernet, remote control is not automatically
       set by means of a command. The instrument must be explicitly set to
       the REMOTE state, e.g. by sending the intf command &GTR
    */
    checkErr (viWrite (instrSession, (ViBuf) "&GTR\n", 5, NULL));
    
    /*- Reset instrument ----------------------------------------------------*/
    if (resetDevice == VI_TRUE)
    {
        checkErr( rscmwgg_reset (instrSession));
    }
    else  /*- Send Default Instrument Setup ---------------------------------*/
        checkErr( rscmwgg_DefaultInstrSetup (instrSession));

    /*- Identification Query ------------------------------------------------*/
    if (IDQuery == VI_TRUE)
        {
        ViChar rdBuffer[RSCMWGG_IO_BUFFER_SIZE] = "";

        checkErr( Rs_GetAttribute (instrSession, VI_NULL, (ViAttr) RSCMWGG_ATTR_ID_QUERY_RESPONSE,
                                   0, RSCMWGG_IO_BUFFER_SIZE, rdBuffer));

        if (strncmp (rdBuffer, RSCMWGG_VALID_ID_RESPONSE_STRING, strlen(RSCMWGG_VALID_ID_RESPONSE_STRING)) != 0)
            {
            viCheckErr( VI_ERROR_FAIL_ID_QUERY);
            }
        }

    if (isLogicalName == VI_FALSE)
        {
        ViInt32 oldFlag = 0;

        checkErr (Rs_GetAttributeFlags (instrSession, RS_ATTR_IO_RESOURCE_DESCRIPTOR, &oldFlag));
        checkErr (Rs_SetAttributeFlags (instrSession, RS_ATTR_IO_RESOURCE_DESCRIPTOR, oldFlag & 0xfffb | 0x0010));
        checkErr (Rs_SetAttribute (instrSession, "", RS_ATTR_IO_RESOURCE_DESCRIPTOR, 0, newResourceName));
        }
    

    if (isLogicalName == VI_TRUE)
        checkErr( Rs_ApplyDefaultSetup (instrSession));

    /* Instrument session accessible by the end user is native VISA session,
       or when simulated virtual instrument (vi) session. */
    checkErr( rscmwgg_CheckStatus (instrSession));

Error:

    if (error < VI_SUCCESS)
    {
        ViSession   rmSess = VI_NULL;

        (void) Rs_GetAttribute (instrSession, VI_NULL, RS_ATTR_VISA_RM_SESSION, 0, (ViInt32) sizeof (rmSess), &rmSess);
        
        if (instrSession != 0)
            (void) viClose (instrSession);
        (void) viClose (rmSess);
    }
    else
    {
        *newInstrSession = instrSession;
    }

    if (error < VI_SUCCESS)
        (void) Rs_Dispose (*newInstrSession);

    return error;
}

/*****************************************************************************
 * Function: Configure Generator State
 * Purpose:  Turns the generator on or off.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_ConfigureGeneratorState (ViSession instrSession,
                                                   ViInt32 instance,
                                                   ViBoolean state)
{
    ViStatus    error   = VI_SUCCESS;
    ViChar      rep_cap[5] = "";
    
    checkErr( Rs_LockSession (instrSession, VI_NULL));

    checkErr(rscmwgg_instanceRepCap (instrSession, instance, rep_cap));
    
    viCheckParm(rscmwgg_SetAttributeViBoolean(instrSession, rep_cap,
        RSCMWGG_ATTR_GENERATOR_STATE, state), 3, "State");
    
    checkErr( rscmwgg_CheckStatus (instrSession));
    
Error:
    Rs_UnlockSession(instrSession, VI_NULL);
    
    return error;   
    
}


/*****************************************************************************
 * Function: Query Generator State
 * Purpose:  Queries the generator state.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_QueryGeneratorState (ViSession instrSession,
                                               ViInt32 instance, ViInt32 *state)
{
    ViStatus    error   = VI_SUCCESS;
    ViChar      rep_cap[5] = "";
    
    checkErr( Rs_LockSession (instrSession, VI_NULL));
    checkErr(rscmwgg_instanceRepCap (instrSession, instance, rep_cap));
    
    if (state == VI_NULL)
        viCheckParm( RS_ERROR_INVALID_PARAMETER, 3, "Null address for state");

    checkErr( rscmwgg_GetAttributeViInt32 (instrSession, rep_cap,
        RSCMWGG_ATTR_QUERY_GENERATOR_STATE, state));
    
    checkErr( rscmwgg_CheckStatus (instrSession));

Error:
    (void) Rs_UnlockSession(instrSession, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: Configure Generator Signal Routing
 * Purpose:  Activates the standalone scenario, selects the outputh path and
 *           connector and define an external attenuation value.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_ConfigureGeneratorSignalRouting (ViSession instrSession,
                                                           ViInt32 instance,
                                                           ViInt32 RFConnector,
                                                           ViInt32 RFConverter,
                                                           ViReal64 externalAttenuation)
{
    ViStatus    error   = VI_SUCCESS;
    ViChar      rep_cap[5] = "";
    
    checkErr( Rs_LockSession (instrSession, VI_NULL));
    
    if (rscmwgg_invalidViInt32Range (instance, 1, 4) == VI_TRUE)
    {
        viCheckParm (RS_ERROR_INVALID_PARAMETER, 2, "Instance");
    };
    
    if (rscmwgg_invalidViInt32Range (RFConnector, RSCMWGG_VAL_CONN_RF1C, RSCMWGG_VAL_CONN_RFAO) == VI_TRUE)
    {
        viCheckParm (RS_ERROR_INVALID_PARAMETER, 3, "RX Connector");
    };
    
    if (rscmwgg_invalidViInt32Range (RFConverter, RSCMWGG_VAL_TX1, RSCMWGG_VAL_TX2) == VI_TRUE)
    {
        viCheckParm (RS_ERROR_INVALID_PARAMETER, 3, "RF Converter");
    };
    
    checkErr(rscmwgg_instanceRepCap (instrSession, instance, rep_cap));
    
    checkErr (viPrintf (instrSession, "ROUT:GSM:GEN%ld:SCEN:SAL %s,%s\n", instance, 
                                                                          RFConnectorArr[RFConnector],
                                                                          RFConverterArr[RFConverter]));
    checkErr( rscmwgg_CheckStatus (instrSession)); 

    viCheckParm(rscmwgg_SetAttributeViReal64(instrSession, rep_cap,
                                            RSCMWGG_ATTR_GENERATOR_EXTERNAL_ATTENUATION, 
                                            externalAttenuation
                                            ), 4, "External Attenuation");

    
Error:
    Rs_UnlockSession(instrSession, VI_NULL);
    
    return error;   
}

/*****************************************************************************
 * Function: Query Generator Signal Routing
 * Purpose:  Queries selected outputh path and connector.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_QueryGeneratorSignalRouting (ViSession instrSession,
                                                       ViInt32 instance,
                                                       ViInt32 *scenario,
                                                       ViChar  controller[],
                                                       ViInt32 *RFConnector,
                                                       ViInt32 *RFConverter)
{
    ViStatus    error   = VI_SUCCESS;
    ViChar      *pbuffer = NULL;
    ViChar      *p2buf = NULL;
    ViUInt32    bytesRead = 0;
    
    checkErr( Rs_LockSession (instrSession, VI_NULL));

    if (rscmwgg_invalidViInt32Range (instance, 1, 4) == VI_TRUE)
    {
        viCheckParm (RS_ERROR_INVALID_PARAMETER, 2, "Instance");
    };
    
    viCheckErr (viPrintf (instrSession, "ROUT:GSM:GEN%ld?\n", instance));
    viCheckErr(Rs_ReadDataUnknownLength (instrSession, &pbuffer, &bytesRead));
    checkErr( rscmwgg_CheckStatus (instrSession)); 
    
    p2buf = strtok (pbuffer, ",\r\n");
    if (strcmp ("SAL", p2buf) == 0)
        *scenario = 1;
    else
        *scenario = -1;

    p2buf = strtok (NULL, ",\r\n");
    sscanf (p2buf, "\"%[^\"]", controller);
    
    p2buf = strtok (NULL, ",\r\n");
    *RFConnector = rscmwgg_StringIndex (RFConnectorArr, p2buf);

    p2buf = strtok (NULL, ",\r\n");
    *RFConverter = rscmwgg_StringIndex (RFConverterArr, p2buf);
    
Error:
    if (pbuffer) free (pbuffer);
    Rs_UnlockSession(instrSession, VI_NULL);
    
    return error;  
}

/*****************************************************************************
 * Function: Configure Generator Frequency
 * Purpose:  Sets the RF carrier frequency of the GSM generator.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_ConfigureGeneratorFrequency (ViSession instrSession,
                                                       ViInt32 instance,
                                                       ViReal64 frequency)
{
    ViStatus    error   = VI_SUCCESS;
    ViChar      rep_cap[5] = "";
    
    checkErr( Rs_LockSession (instrSession, VI_NULL));

    checkErr(rscmwgg_instanceRepCap (instrSession, instance, rep_cap));

    viCheckParm(rscmwgg_SetAttributeViReal64(instrSession, rep_cap,
        RSCMWGG_ATTR_GENERATOR_FREQUENCY, frequency), 3, 
        "Frequency");

    checkErr( rscmwgg_CheckStatus (instrSession));
    
Error:
    Rs_UnlockSession(instrSession, VI_NULL);
    
    return error;   
}

/*****************************************************************************
 * Function: Configure Generator Channel
 * Purpose:  Selects the frequency band and frequency channel of the
 *           generated GSM signal.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_ConfigureGeneratorChannel (ViSession instrSession,
                                                     ViInt32 instance,
                                                     ViInt32 band,
                                                     ViInt32 channel)
{
    ViStatus    error   = VI_SUCCESS;
    ViChar      rep_cap[5] = "";
    
    checkErr( Rs_LockSession (instrSession, VI_NULL));

    checkErr(rscmwgg_instanceRepCap (instrSession, instance, rep_cap));

    viCheckParm(rscmwgg_SetAttributeViInt32(instrSession, rep_cap,
        RSCMWGG_ATTR_GENERATOR_BAND, band), 3, "Band");

    viCheckParm(rscmwgg_SetAttributeViInt32(instrSession, rep_cap,
        RSCMWGG_ATTR_GENERATOR_CHANNEL, channel), 4, "Channel");

    checkErr( rscmwgg_CheckStatus (instrSession));
    
Error:
    Rs_UnlockSession(instrSession, VI_NULL);
    
    return error;   
}

/*****************************************************************************
 * Function: Configure Generator RMS Level
 * Purpose:  Sets the "Full-Slot Level" (RMS level) of the generator.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_ConfigureGeneratorRMSLevel (ViSession instrSession,
                                                      ViInt32 instance,
                                                      ViReal64 level)
{
    ViStatus    error   = VI_SUCCESS;
    ViChar      rep_cap[5] = "";
    
    checkErr( Rs_LockSession (instrSession, VI_NULL));

    
    checkErr(rscmwgg_instanceRepCap (instrSession, instance, rep_cap));

    viCheckParm(rscmwgg_SetAttributeViReal64(instrSession, rep_cap,
        RSCMWGG_ATTR_GENERATOR_RMS_LEVEL, level), 3, "Level");
        
    checkErr( rscmwgg_CheckStatus (instrSession));
    
Error:
    Rs_UnlockSession(instrSession, VI_NULL);
    
    return error;
}

/*****************************************************************************
 * Function: Query Generator PEP
 * Purpose:  Queries the Peak Envelope Power (PEP) of the RF generator.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_QueryGeneratorPEP (ViSession instrSession,
                                             ViInt32 instance, ViReal64 *PEP)
{
    ViStatus    error   = VI_SUCCESS;
    ViChar      rep_cap[5] = "";
    
    checkErr( Rs_LockSession (instrSession, VI_NULL));
    checkErr(rscmwgg_instanceRepCap (instrSession, instance, rep_cap));
    
    if (PEP == VI_NULL)
        viCheckParm( RS_ERROR_INVALID_PARAMETER, 3, "Null address for PEP");

    checkErr( rscmwgg_GetAttributeViReal64 (instrSession, rep_cap,
        RSCMWGG_ATTR_QUERY_GENERATOR_PEP, PEP));
    
    checkErr( rscmwgg_CheckStatus (instrSession));

Error:
    (void) Rs_UnlockSession(instrSession, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: Configure Generator Slot Level
 * Purpose:  Selects the average burst power in slot 0 to 7.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_ConfigureGeneratorSlotLevel (ViSession instrSession,
                                                       ViInt32 instance,
                                                       ViInt32 slotLevel[])
{
    ViStatus    error   = VI_SUCCESS;
    ViChar      buffer[RS_MAX_MESSAGE_BUF_SIZE]="";
    ViChar      *p2buffer=buffer;
    ViInt32     i;
    
    checkErr( Rs_LockSession (instrSession, VI_NULL));

    if (rscmwgg_invalidViInt32Range (instance, 1, 4) == VI_TRUE)
    {
        viCheckParm (RS_ERROR_INVALID_PARAMETER, 2, "Instance");
    };  
    
    p2buffer += sprintf (p2buffer, "SOURce:GSM:GEN%ld:SCONfig:SLEVel ", instance);
    
    for (i=0;i<8;i++)
    {
        if (rscmwgg_invalidViInt32Range (slotLevel[i], 0, 8) == VI_TRUE)
        {
            viCheckParm (RS_ERROR_INVALID_PARAMETER, 3, "SlotLevel");
        }
        else
        {
            p2buffer += sprintf (p2buffer, "%s,", slotLevelArr[slotLevel[i]]);
        }  
    }
    
    *--p2buffer='\0';
    checkErr (viPrintf (instrSession, "%s\n", buffer));
        
    checkErr( rscmwgg_CheckStatus (instrSession));
    
Error:
    Rs_UnlockSession(instrSession, VI_NULL);
    
    return error;   
}

/*****************************************************************************
 * Function: Configure Generator Slot Burst Type
 * Purpose:  Selects the length, modulation type and bit structure of the
 *           bursts in slot 0 to 7.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_ConfigureGeneratorSlotBurstType (ViSession instrSession,
                                                           ViInt32 instance,
                                                           ViInt32 burstType[])
{
    ViStatus    error   = VI_SUCCESS;
    ViChar      buffer[RS_MAX_MESSAGE_BUF_SIZE]="";
    ViChar      *p2buffer=buffer;
    ViInt32     i;
    
    checkErr( Rs_LockSession (instrSession, VI_NULL));

    if (rscmwgg_invalidViInt32Range (instance, 1, 4) == VI_TRUE)
    {
        viCheckParm (RS_ERROR_INVALID_PARAMETER, 2, "Instance");
    };  
    
    p2buffer += sprintf (p2buffer, "SOURce:GSM:GEN%ld:SCONfig:BTYPe ", instance);
    
    for (i=0;i<8;i++)
    {
        if (rscmwgg_invalidViInt32Range (burstType[i], 0, 7) == VI_TRUE)
        {
            viCheckParm (RS_ERROR_INVALID_PARAMETER, 3, "Burst Type");
        }
        else
        {
            p2buffer += sprintf (p2buffer, "%s,", burstTypeArr[burstType[i]]);
        }  
    }
    
    *--p2buffer='\0';
    checkErr (viPrintf (instrSession, "%s\n", buffer));
        
    checkErr( rscmwgg_CheckStatus (instrSession));
    
Error:
    Rs_UnlockSession(instrSession, VI_NULL);
    
    return error;   
}

/*****************************************************************************
 * Function: Configure Generator Slot Data Source
 * Purpose:  Selects the bit sequence to be transmitted in the data fields of
 *           the bursts in slot 0 to 7. The allowed settings depend on the
 *           burst type: The data fields for frequency correction bursts,
 *           dummy bursts are not configurable.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_ConfigureGeneratorSlotDataSource (ViSession instrSession,
                                                            ViInt32 instance,
                                                            ViInt32 dataSource[])
{
    ViStatus    error   = VI_SUCCESS;
    ViChar      buffer[RS_MAX_MESSAGE_BUF_SIZE]="";
    ViChar      *p2buffer=buffer;
    ViInt32     i;
    
    checkErr( Rs_LockSession (instrSession, VI_NULL));

    if (rscmwgg_invalidViInt32Range (instance, 1, 4) == VI_TRUE)
    {
        viCheckParm (RS_ERROR_INVALID_PARAMETER, 2, "Instance");
    };  
    
    p2buffer += sprintf (p2buffer, "SOURce:GSM:GEN%ld:SCONfig:DSOurce ", instance);
    
    for (i=0;i<8;i++)
    {
        if (rscmwgg_invalidViInt32Range (dataSource[i], 0, 10) == VI_TRUE)
        {
            viCheckParm (RS_ERROR_INVALID_PARAMETER, 3, "Data Source");
        }
        else
        {
            p2buffer += sprintf (p2buffer, "%s,", dataSourceArr[dataSource[i]]);
        }  
    }
    
    *--p2buffer='\0';
    checkErr (viPrintf (instrSession, "%s\n", buffer));
        
    checkErr( rscmwgg_CheckStatus (instrSession));
    
Error:
    Rs_UnlockSession(instrSession, VI_NULL);
    
    return error;
}

/*****************************************************************************
 * Function: Configure Generator Slot Test Sequence
 * Purpose:  Sets one of the specified GSM training sequences TSC 0 to TSC 7
 *           for the burst in slot 0 to 7. The allowed settings depend on the
 *           burst typ
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_ConfigureGeneratorSlotTestSequence (ViSession instrSession,
                                                              ViInt32 instance,
                                                              ViInt32 testSequence[])
{
    ViStatus    error   = VI_SUCCESS;
    ViChar      buffer[RS_MAX_MESSAGE_BUF_SIZE]="";
    ViChar      *p2buffer=buffer;
    ViInt32     i;
    
    checkErr( Rs_LockSession (instrSession, VI_NULL));

    if (rscmwgg_invalidViInt32Range (instance, 1, 4) == VI_TRUE)
    {
        viCheckParm (RS_ERROR_INVALID_PARAMETER, 2, "Instance");
    };  
    
    p2buffer += sprintf (p2buffer, "SOURce:GSM:GEN%ld:SCONfig:TSEQuence ", instance);
    
    for (i=0;i<8;i++)
    {
        if (rscmwgg_invalidViInt32Range (testSequence[i], 0, 8) == VI_TRUE)
        {
            viCheckParm (RS_ERROR_INVALID_PARAMETER, 3, "Test Sequence");
        }
        else
        {
            p2buffer += sprintf (p2buffer, "%s,", testSequenceArr[testSequence[i]]);
        }  
    }
    
    *--p2buffer='\0';
    checkErr (viPrintf (instrSession, "%s\n", buffer));
        
    checkErr( rscmwgg_CheckStatus (instrSession));
    
Error:
    Rs_UnlockSession(instrSession, VI_NULL);
    
    return error;
}

/*****************************************************************************
 * Function: Configure Generator Slot Modulation Scheme
 * Purpose:  Selects the modulation scheme and coding for the transfer of
 *           speech data or radio blocks via traffic channels. The setting is
 *           valid for traffic channel burst types only (Configure Generator
 *           Slot Burst Type - TCH). Slot no. 0 must carry a BCCH channel
 *           with TCH/FS traffic mode, therefore the first entry (<Mode 0>)
 *           is ignored.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_ConfigureGeneratorSlotModulationScheme (ViSession instrSession,
                                                                      ViInt32 instance,
                                                                      ViInt32 modulationScheme[])
{
    ViStatus    error   = VI_SUCCESS;
    ViChar      buffer[RS_MAX_MESSAGE_BUF_SIZE]="";
    ViChar      *p2buffer=buffer;
    ViInt32     i;
    
    checkErr( Rs_LockSession (instrSession, VI_NULL));

    if (rscmwgg_invalidViInt32Range (instance, 1, 4) == VI_TRUE)
    {
        viCheckParm (RS_ERROR_INVALID_PARAMETER, 2, "Instance");
    };  
    
    p2buffer += sprintf (p2buffer, "SOUR:GSM:GEN%ld:SCON:TMOD ", instance);
    
    for (i=0;i<8;i++)
    {
        if (rscmwgg_invalidViInt32Range (modulationScheme[i], RSCMWGG_VAL_SLOT_MODE_TFS, RSCMWGG_VAL_SLOT_MODE_SM92) == VI_TRUE)
        {
            viCheckParm (RS_ERROR_INVALID_PARAMETER, 3, "Modulation Scheme");
        }
        else
        {
            p2buffer += sprintf (p2buffer, "%s,", modSchemeArr[modulationScheme[i]]);
        }  
    }
    
    *--p2buffer='\0';
    checkErr (viPrintf (instrSession, "%s\n", buffer));
        
    checkErr( rscmwgg_CheckStatus (instrSession));
    
Error:
    Rs_UnlockSession(instrSession, VI_NULL);
    
    return error;
}

/*****************************************************************************
 * Function: Configure Generator Slot Attenuations
 * Purpose:  Configures 7 relative slot levels, to be used for individual
 *           slot level settings.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_ConfigureGeneratorSlotAttenuations (ViSession instrSession,
                                                              ViInt32 instance,
                                                              ViReal64 attenuations[])
{
    ViStatus    error   = VI_SUCCESS;
    ViChar      buffer[RS_MAX_MESSAGE_BUF_SIZE]="";
    ViChar      *p2buffer=buffer;
    ViInt32     i;
    
    checkErr( Rs_LockSession (instrSession, VI_NULL));

    if (rscmwgg_invalidViInt32Range (instance, 1, 4) == VI_TRUE)
    {
        viCheckParm (RS_ERROR_INVALID_PARAMETER, 2, "Instance");
    };  
    
    p2buffer += sprintf (p2buffer, "SOURce:GSM:GEN%ld:SCONfig:SATTenuation ", instance);
    
    for (i=0;i<7;i++)
    {
        if (rscmwgg_invalidViReal64Range (attenuations[i], 0.0, 40.0) == VI_TRUE)
        {
            viCheckParm (RS_ERROR_INVALID_PARAMETER, 3, "Attenuations");
        }
        else
        {
            p2buffer += sprintf (p2buffer, "%f,", attenuations[i]);
        }  
    }
    
    *--p2buffer='\0';
    checkErr (viPrintf (instrSession, "%s\n", buffer));
        
    checkErr( rscmwgg_CheckStatus (instrSession));
    
Error:
    Rs_UnlockSession(instrSession, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: Configure Generator Slot Power Control Level
 * Purpose:  Selects the Power Control Level (PCL) that is signaled to the
 *           mobile under test. The setting is valid for traffic channel
 *           burst types only (function Configure Generator Slot Burst Type -
 *           TCH). The PCL values <TCH 1> to <TCH 7> control the output power
 *           of the mobile in slots 0 to 7; the first entry (<TCH 0>) is
 *           ignored.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_ConfigureGeneratorSlotPowerControlLevel (ViSession instrSession,
                                                                   ViInt32 instance,
                                                                   ViInt32 powerControlLevel[])
{
    ViStatus    error   = VI_SUCCESS;
    ViChar      buffer[RS_MAX_MESSAGE_BUF_SIZE]="";
    ViChar      *p2buffer=buffer;
    ViInt32     i;
    
    checkErr( Rs_LockSession (instrSession, VI_NULL));

    if (rscmwgg_invalidViInt32Range (instance, 1, 4) == VI_TRUE)
    {
        viCheckParm (RS_ERROR_INVALID_PARAMETER, 2, "Instance");
    };  
    
    p2buffer += sprintf (p2buffer, "SOUR:GSM:GEN%ld:SCON:PCL ", instance);
    
    for (i=0;i<8;i++)
    {
        if (rscmwgg_invalidViReal64Range (powerControlLevel[i], 0, 31) == VI_TRUE)
        {
            viCheckParm (RS_ERROR_INVALID_PARAMETER, 3, "Power Control Level");
        }
        else
        {
            p2buffer += sprintf (p2buffer, "%ld,", powerControlLevel[i]);
        }  
    }
    
    *--p2buffer='\0';
    checkErr (viPrintf (instrSession, "%s\n", buffer));
        
    checkErr( rscmwgg_CheckStatus (instrSession));
    
Error:
    Rs_UnlockSession(instrSession, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: rscmwgg_SetAttribute<type> Functions
 * Purpose:  These functions enable the instrument driver user to set
 *           attribute values directly.  There are typesafe versions for
 *           ViInt32, ViReal64, ViString, ViBoolean, and ViSession datatypes.
 *****************************************************************************/
/* ViInt32 */
ViStatus _VI_FUNC rscmwgg_SetAttributeViInt32 (ViSession instrSession, ViString channelName,
                                                ViUInt32 attributeId, ViInt32 value)
{
    return Rs_SetAttribute (instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, &value);
}

/* ViReal64 */
ViStatus _VI_FUNC rscmwgg_SetAttributeViReal64 (ViSession instrSession, ViString channelName,
                                                 ViUInt32 attributeId, ViReal64 value)
{
     return Rs_SetAttribute (instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, &value);
}

/* ViString */
ViStatus _VI_FUNC rscmwgg_SetAttributeViString (ViSession instrSession, ViString channelName,
                                                 ViUInt32 attributeId, ViString value)
{
    return Rs_SetAttribute (instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, value);
}

/* ViBoolean */
ViStatus _VI_FUNC rscmwgg_SetAttributeViBoolean (ViSession instrSession, ViString channelName,
                                                  ViUInt32 attributeId, ViBoolean value)
{
    return Rs_SetAttribute (instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, &value);
}

/* ViSession */
ViStatus _VI_FUNC rscmwgg_SetAttributeViSession (ViSession instrSession, ViString channelName,
                                                  ViUInt32 attributeId, ViSession value)
{
    return Rs_SetAttribute (instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, &value);
}

/*****************************************************************************
 * Function: rscmwgg_GetAttribute<type> Functions
 * Purpose:  These functions enable the instrument driver user to get
 *           attribute values directly.  There are typesafe versions for
 *           ViInt32, ViReal64, ViString, ViBoolean, and ViSession attributes.
 *****************************************************************************/

/* ViInt32 */
ViStatus _VI_FUNC rscmwgg_GetAttributeViInt32 (ViSession instrSession, ViString channelName,
                                                ViUInt32 attributeId, ViInt32 *value)
{
    return Rs_GetAttribute (instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL,
                            (ViInt32) sizeof (ViInt32), (ViInt32 *)value);
}

/* ViReal64 */
ViStatus _VI_FUNC rscmwgg_GetAttributeViReal64 (ViSession instrSession, ViString channelName,
                                                 ViUInt32 attributeId, ViReal64 *value)
{
    return Rs_GetAttribute (instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL,
                            (ViInt32) sizeof (ViReal64), (ViReal64 *)value);
}

/* ViString */
ViStatus _VI_FUNC rscmwgg_GetAttributeViString (ViSession instrSession, ViString channelName,
                                                 ViUInt32 attributeId, ViInt32 bufSize,
                                                 ViChar value[])
{
    return Rs_GetAttribute (instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL,
                            bufSize, value);
}

/* ViBoolean */
ViStatus _VI_FUNC rscmwgg_GetAttributeViBoolean (ViSession instrSession, ViString channelName,
                                                  ViUInt32 attributeId, ViBoolean *value)
{
    return Rs_GetAttribute (instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL,
                            (ViInt32) sizeof (ViBoolean), (ViBoolean *)value);
}

/* ViSession */
ViStatus _VI_FUNC rscmwgg_GetAttributeViSession (ViSession instrSession, ViString channelName,
                                                  ViUInt32 attributeId, ViSession *value)
{
    return Rs_GetAttribute (instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL,
                            (ViInt32) sizeof (ViSession), (ViSession *)value);
}


/*****************************************************************************
 * Function: Rs_GetAttributeRepCapNameId
 * Purpose:  This function returns the repeated capability id(s) that
 *           belongs to the attribute you specify. More then one id is listed
 *           in comma separated string.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_GetAttributeRepeatedCapabilityIds (ViSession instrSession,
                                                            ViUInt32 attributeID,
                                                            ViInt32 bufferSize,
                                                            ViChar _VI_FAR repeatedCapabilityIds[])
{
    ViStatus    error = VI_SUCCESS;
    ViChar      repCapNameId[RS_MAX_MESSAGE_BUF_SIZE] = "";

    if (bufferSize <= 0)
        viCheckParm(RS_ERROR_INVALID_PARAMETER, 3, "Buffer Size");

    viCheckErr (Rs_GetAttributeRepCapNameId (instrSession, attributeID, repCapNameId));

    strncpy (repeatedCapabilityIds, repCapNameId, RS_MAX_MESSAGE_BUF_SIZE);
    repeatedCapabilityIds[(bufferSize > RS_MAX_MESSAGE_BUF_SIZE) ?
        RS_MAX_MESSAGE_BUF_SIZE : bufferSize] = '\0';

Error:
    return error;
}

/*****************************************************************************
 * Function: Get Attribute Repeated Capability Id Name(s)
 * Purpose:  This function returns the repeated capability name(s) that
 *           belongs to the attribute and single repeated capability id you
 *           may specify. More then one name is listed in comma separated
 *           string.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_GetAttributeRepeatedCapabilityIdNames (ViSession instrSession,
                                                                ViUInt32 attributeID,
                                                                ViString repeatedCapabilityId,
                                                                ViInt32 bufferSize,
                                                                ViChar _VI_FAR repeatedCapabilityIdNames[])
{
    ViStatus    error = VI_SUCCESS;
    ViChar      repCapName[RS_MAX_MESSAGE_BUF_SIZE] = "";

    if (bufferSize <= 0)
        viCheckParm(RS_ERROR_INVALID_PARAMETER, 4, "Buffer Size");

    viCheckErr (Rs_GetAttributeRepCapName (instrSession,
                                           attributeID,
                                           repeatedCapabilityId,
                                           repCapName));

    strncpy (repeatedCapabilityIdNames, repCapName, RS_MAX_MESSAGE_BUF_SIZE);
    repeatedCapabilityIdNames[(bufferSize > RS_MAX_MESSAGE_BUF_SIZE) ?
        RS_MAX_MESSAGE_BUF_SIZE : bufferSize] = '\0';

Error:
    return error;
}

/*****************************************************************************
 * Function: rscmwgg_CheckAttribute<type> Functions
 * Purpose:  These functions enable the instrument driver user to check
 *           attribute values directly.  These functions check the value you
 *           specify even if you set the rscmwgg_ATTR_RANGE_CHECK
 *           attribute to VI_FALSE.  There are typesafe versions for ViInt32,
 *           ViReal64, ViString, ViBoolean, and ViSession datatypes.
 *****************************************************************************/

/* ViInt32 */
ViStatus _VI_FUNC rscmwgg_CheckAttributeViInt32 (ViSession instrSession, ViString channelName,
                                                  ViUInt32 attributeId, ViInt32 value)
{
    return Rs_CheckAttribute (instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, &value);
}

/* ViReal64 */
ViStatus _VI_FUNC rscmwgg_CheckAttributeViReal64 (ViSession instrSession, ViString channelName,
                                                   ViUInt32 attributeId, ViReal64 value)
{
    return Rs_CheckAttribute (instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, &value);
}

/* ViString */
ViStatus _VI_FUNC rscmwgg_CheckAttributeViString (ViSession instrSession, ViString channelName,
                                                   ViUInt32 attributeId, ViString value)
{
    return Rs_CheckAttribute (instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, value);
}

/* ViBoolean */
ViStatus _VI_FUNC rscmwgg_CheckAttributeViBoolean (ViSession instrSession, ViString channelName,
                                                    ViUInt32 attributeId, ViBoolean value)
{
    return Rs_CheckAttribute (instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, &value);
}

/* ViSession */
ViStatus _VI_FUNC rscmwgg_CheckAttributeViSession (ViSession instrSession, ViString channelName,
                                                    ViUInt32 attributeId, ViSession value)
{
    return Rs_CheckAttribute (instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, &value);
}

/*****************************************************************************
 * Function: Reset
 * Purpose:  This function resets the instrument.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_reset (ViSession instrSession)
{
    ViStatus    error   = VI_SUCCESS;

    checkErr( Rs_LockSession (instrSession, VI_NULL));

    if (Rs_Simulating(instrSession) == 0)                /* call only when locked */
        {
        viCheckErr( viPrintf (instrSession, "*RST\n"));
        }

    checkErr( rscmwgg_DefaultInstrSetup (instrSession));
    checkErr( rscmwgg_CheckStatus (instrSession));

Error:

    (void) Rs_UnlockSession (instrSession, VI_NULL);

    return error;
}

/*****************************************************************************
 * Function: Self Test
 * Purpose:  This function executes the instrument self-test and returns the
 *           result.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_self_test (ViSession instrSession, ViInt16 *testResult,
                                      ViChar testMessage[])
{
    ViStatus    error   = VI_SUCCESS;
    
    checkErr( Rs_LockSession (instrSession, VI_NULL));

    if (testResult == NULL)
        viCheckParm( RS_ERROR_INVALID_PARAMETER, 2, "Null address for Test Result");
    if (testMessage == NULL)
        viCheckParm( RS_ERROR_INVALID_PARAMETER, 3, "Null address for Test Message");

    if (Rs_Simulating(instrSession) == 0)                /* call only when locked */
        {
        viCheckErr( viPrintf (instrSession, "*TST?"));
        viCheckErr( rscmwgg_WaitForOPC (instrSession, 600000));  
        viCheckErr( viScanf (instrSession, "%hd", testResult));
        if (*testResult == 0)
            strcpy (testMessage, "Self-Test Passed.");
        else
            strcpy (testMessage, "Self-Test Failed.");
        }
    else
        {
        /* Simulate Self Test */
        *testResult = 0;
        strcpy (testMessage, "No error.");
        }

    checkErr( rscmwgg_CheckStatus (instrSession));

Error:
    (void) Rs_UnlockSession(instrSession, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: Self Test Results
 * Purpose:  This function executes the instrument self-test and returns the
 *           result.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_SelfTestResults (ViSession instrSession,
                                            ViInt32 bufferSize,
                                            ViChar results[])
{
    ViStatus    error   = VI_SUCCESS;
    ViChar      command[] = "DIAG:SERV:STES:RES?\n";
    
    checkErr( Rs_LockSession (instrSession, VI_NULL));

    checkErr (viWrite (instrSession, (ViBuf) command, (ViUInt32) strlen (command), NULL));
    checkErr (viRead (instrSession, (ViBuf) results, (ViUInt32) bufferSize, NULL));

    checkErr( rscmwgg_CheckStatus (instrSession));

Error:
    (void) Rs_UnlockSession(instrSession, VI_NULL);
    return error;

}

/*****************************************************************************
 * Function: Revision Query
 * Purpose:  This function returns the driver and instrument revisions.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_revision_query (ViSession instrSession, ViChar driverRev[],
                                           ViChar instrRev[])
{
    ViStatus    error   = VI_SUCCESS;

    checkErr( Rs_LockSession (instrSession, VI_NULL));

    if (driverRev == VI_NULL)
        viCheckParm( RS_ERROR_INVALID_PARAMETER, 2, "Null address for Driver Revision");
    if (instrRev == VI_NULL)
        viCheckParm( RS_ERROR_INVALID_PARAMETER, 3, "Null address for Instrument Revision");

    checkErr( Rs_GetAttribute (instrSession, VI_NULL, RSCMWGG_ATTR_SPECIFIC_DRIVER_REVISION,
                                        0, 256, driverRev));
    checkErr( Rs_GetAttribute (instrSession, "", RSCMWGG_ATTR_INSTRUMENT_FIRMWARE_REVISION,
                                        0, 256, instrRev));
    checkErr( rscmwgg_CheckStatus (instrSession));

Error:
    (void) Rs_UnlockSession(instrSession, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function:    System Version
 * Purpose:     This function queries the number of the SCPI version, which 
 *              is relevant for the instrument.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_SystemVersion (ViSession instrSession,
                                          ViInt32 length, 
                                          ViChar systemVersion[])

{
    ViStatus    error = VI_SUCCESS;
   
    checkErr( Rs_LockSession (instrSession, VI_NULL)); 
    
    viCheckParm( rscmwgg_GetAttributeViString (instrSession, "", 
                                                RSCMWGG_ATTR_SYSTEM_VERSION, length,
                                                systemVersion),
                                                2, "System Version");
    checkErr( rscmwgg_CheckStatus (instrSession));
    
Error:
    (void) Rs_UnlockSession(instrSession, VI_NULL); 
    return error;
}

/*****************************************************************************
 * Function: Error Query
 * Purpose:  This function queries the instrument error queue and returns
 *           the result.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_error_query (ViSession instrSession, ViInt32 *errCode,
                                        ViChar errMessage[])
{
    ViStatus    error   = VI_SUCCESS;
    ViChar      buffer[RSCMWGG_IO_BUFFER_SIZE] = "";
    
    checkErr( Rs_LockSession (instrSession, VI_NULL));

    if (errCode == NULL)
        viCheckParm( RS_ERROR_INVALID_PARAMETER, 2, "Null address for Error Code");
    if (errMessage == NULL)
        viCheckParm( RS_ERROR_INVALID_PARAMETER, 3, "Null address for Error Message");

    if (Rs_Simulating(instrSession) == 0)                /* call only when locked */
        {
        viCheckErr (rscmwgg_GetAttributeViString( instrSession, "", RSCMWGG_ATTR_SYSTEM_ERROR, RSCMWGG_IO_BUFFER_SIZE, buffer));
        if  (sscanf (buffer, "%ld,\"%256[^\"]", errCode, errMessage) != 2)
            viCheckErr(RS_ERROR_UNEXPECTED_RESPONSE);

        }
    else
        {
            /* Simulate Error Query */
        *errCode = 0;
        strcpy (errMessage, "No error.");
        }

Error:
    (void) Rs_UnlockSession(instrSession, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: Error List
 * Purpose:  This function reads all system messages and returns a list of 
 *           comma separated strings.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_ErrorList (ViSession instrSession,
                                      ViInt32 bufferSize, 
                                      ViChar errors[])
{
    ViStatus    error   = VI_SUCCESS;
    
    checkErr( Rs_LockSession (instrSession, VI_NULL));

    if (bufferSize < 0)
        viCheckParm( RS_ERROR_INVALID_PARAMETER, 2, "Buffer Size");
    if (errors == VI_NULL)
        viCheckParm( RS_ERROR_INVALID_PARAMETER, 3, "Null address for Errors");

    viCheckErr (rscmwgg_GetAttributeViString( instrSession, "", RSCMWGG_ATTR_SYSTEM_ALL_ERRORS, 
        bufferSize, errors));

Error:
    (void) Rs_UnlockSession(instrSession, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: Error Message
 * Purpose:  This function translates the error codes returned by this
 *           instrument driver into user-readable strings.
 *
 *           Note:  The caller can pass VI_NULL for the vi parameter.  This
 *           is useful if one of the init functions fail.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_error_message (ViSession instrSession, ViStatus errorCode,
                                          ViChar errorMessage[])
{
    ViStatus    error   = VI_SUCCESS;

    static      RsStringValueTable errorTable =
        {
        /*=CHANGE:================================================================*
            Insert instrument driver specific error codes here.  Example:

            {rscmwgg_ERROR_TOO_MANY_SAMPLES,   "Sample Count cannot exceed 512."},
         *=============================================================END=CHANGE=*/
            RSCMWGG_ERROR_CODES_AND_MSGS,
            {VI_NULL, VI_NULL}
        };

    (void) Rs_LockSession(instrSession, VI_NULL);

        /* all VISA and RS error codes are handled as well as codes in the table */
    if (errorMessage == VI_NULL)
        viCheckParm( RS_ERROR_INVALID_PARAMETER, 3, "Null address for Error Message");

    checkErr( Rs_GetSpecificDriverStatusDesc(instrSession, errorCode, errorMessage, errorTable));

Error:
    (void) Rs_UnlockSession(instrSession, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: Preset
 * Purpose:  This function initiates an instrument reset.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_Preset (ViSession instrSession)
{
    ViStatus    error   = VI_SUCCESS;
   
    checkErr( Rs_LockSession (instrSession, VI_NULL));

    viCheckErr (rscmwgg_SetAttributeViString (instrSession, "", 
                RSCMWGG_ATTR_SYSTEM_PRESET, VI_NULL));
    
    checkErr( rscmwgg_CheckStatus (instrSession));

Error:
    (void) Rs_UnlockSession(instrSession, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: Get Error and Clear Error Functions
 * Purpose:  These functions enable the instrument driver user to
 *           get or clear the error information the driver associates with the
 *           RS session.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_GetError (ViSession instrSession,
                                     ViStatus *errorCode,
                                     ViInt32 bufferSize,
                                     ViChar description[])
{
    ViStatus    error       = VI_SUCCESS;
    ViStatus    primary     = VI_SUCCESS,
                secondary   = VI_SUCCESS;
    ViChar      elaboration[RS_MAX_MESSAGE_BUF_SIZE]    = "";
    ViChar      errorMessage[RS_MAX_MESSAGE_BUF_SIZE]  = "";
    ViChar     *appendPoint = errorMessage;
    ViInt32     actualSize  = 0;
    ViBoolean   locked      = VI_FALSE;
    RsSessionPropertiesPtr  sessionProperties = Rs_ViSession (instrSession);
 
    /* Lock Session */
    checkErr( Rs_LockSession(instrSession, &locked));
 
    /* Test for nulls and acquire error data */
    if (bufferSize != 0)
    {
        if (errorCode == VI_NULL)
        {
            viCheckParm( RS_ERROR_INVALID_PARAMETER, 2, "Null address for Error");
        }
        if (sessionProperties == NULL)
        {
            primary = VI_ERROR_RSRC_NFOUND;
            secondary = 0;
        }
        else
        {
            checkErr( Rs_GetErrorInfo (instrSession, &primary, &secondary, elaboration));
        }
    }
 
    else
        {
            if (sessionProperties == NULL)
            {
                (void) Rs_UnlockSession (instrSession, &locked);
                return 1024;
            }
            else
            {
                checkErr( Rs_GetAttribute (instrSession, VI_NULL, RS_ATTR_ERROR_ELABORATION, 0, 256, elaboration));
                checkErr( Rs_GetAttribute (instrSession, VI_NULL, RS_ATTR_SECONDARY_ERROR, 0, (ViInt32) sizeof (secondary), &secondary));
                checkErr( Rs_GetAttribute (instrSession, VI_NULL, RS_ATTR_PRIMARY_ERROR, 0, (ViInt32) sizeof (primary), &primary));
            }
        }
 
    /* Format data */
    if (primary != VI_SUCCESS)
        {
        ViChar msg[256] = "";
        checkErr( rscmwgg_error_message (instrSession, primary, msg));
        appendPoint += sprintf(appendPoint, "Primary Error: (Hex 0x%08X) %s\n", (unsigned int) primary, msg);
        }
 
    if (secondary != VI_SUCCESS)
        {
        ViChar msg[256] = "";
        checkErr( rscmwgg_error_message (instrSession, secondary, msg));
        appendPoint += sprintf(appendPoint, "Secondary Error: (Hex 0x%08X) %s\n", (unsigned int) secondary, msg);
        }
 
    if (strlen (elaboration) > 0)
        {
        sprintf(appendPoint, "Elaboration: %s", elaboration);
        }
 
    actualSize = (ViInt32) (strlen(errorMessage) + 1);
 
    /* Prepare return values */
    if (bufferSize == 0)
        {
        error = (ViStatus) actualSize;
        }
    else
        {
        if (bufferSize > 0)
            {
            if (actualSize > bufferSize)
                {
                error = actualSize;
                actualSize = bufferSize;
                }
            }
        if (description == VI_NULL)
            {
            viCheckParm( RS_ERROR_INVALID_PARAMETER, 4, "Null address for Description");
            }
        memcpy(description, errorMessage, (size_t)(actualSize-1));
        description[actualSize-1] = '\0';
        }
 
    if (errorCode)
        {
        *errorCode = primary;
        }
 
Error:
    /* Unlock Session */
    (void) Rs_UnlockSession (instrSession, &locked);
    return error;
}

ViStatus _VI_FUNC rscmwgg_ClearError (ViSession instrSession)
{
    return Rs_ClearErrorInfo (instrSession);
}

/*****************************************************************************
 * Function: WriteInstrData and ReadInstrData Functions
 * Purpose:  These functions enable the instrument driver user to
 *           write and read commands directly to and from the instrument.
 *
 *           Note:  These functions bypass the RS attribute state caching.
 *                  WriteInstrData invalidates the cached values for all
 *                  attributes.
 *****************************************************************************/

ViStatus _VI_FUNC rscmwgg_WriteInstrData (ViSession instrSession, ViString writeBuffer)
{
    return Rs_WriteInstrData (instrSession, writeBuffer);
}

ViStatus _VI_FUNC rscmwgg_ReadInstrData (ViSession instrSession, ViInt32 numBytes,
                                          ViChar rdBuf[], ViUInt32 *bytesRead)
{
    return Rs_ReadInstrData (instrSession, numBytes, rdBuf, bytesRead);
}

/*****************************************************************************
 * Function: Query ViBoolean
 * Purpose:  This function queries the ViBoolean value.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_QueryViBoolean (ViSession instrSession,
                                           ViString command, 
                                           ViBoolean *value)
{
    ViStatus    error = VI_SUCCESS;
    
    checkErr( Rs_LockSession (instrSession, VI_NULL)); 
    if (strlen(command)==0)
        viCheckParm (RS_ERROR_INVALID_PARAMETER, 2, "Command");
    if (value==VI_NULL)
        viCheckParm (RS_ERROR_INVALID_PARAMETER, 3, "Value");
    viCheckErr (viQueryf (instrSession, "%s\n", "%d", command, value));
    checkErr( rscmwgg_CheckStatus (instrSession)); 
Error:
    (void) Rs_UnlockSession(instrSession, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: Query ViInt32
 * Purpose:  This function queries the ViInt32 value.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_QueryViInt32 (ViSession instrSession,
                                         ViString command, 
                                         ViInt32 *value)
{
    ViStatus    error = VI_SUCCESS;
    
    checkErr( Rs_LockSession (instrSession, VI_NULL)); 
    if (strlen(command)==0)
        viCheckParm (RS_ERROR_INVALID_PARAMETER, 2, "Command");
    if (value==VI_NULL)
        viCheckParm (RS_ERROR_INVALID_PARAMETER, 3, "Value");
    viCheckErr (viQueryf (instrSession, "%s\n", "%ld", command, value));
    checkErr( rscmwgg_CheckStatus (instrSession)); 
Error:
    (void) Rs_UnlockSession(instrSession, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: Query ViReal64
 * Purpose:  This function queries the ViReal64 value.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_QueryViReal64 (ViSession instrSession,
                                          ViString command, 
                                          ViReal64 *value)
{
    ViStatus    error = VI_SUCCESS;
    
    checkErr( Rs_LockSession (instrSession, VI_NULL)); 
    if (strlen(command)==0)
        viCheckParm (RS_ERROR_INVALID_PARAMETER, 2, "Command");
    if (value==VI_NULL)
        viCheckParm (RS_ERROR_INVALID_PARAMETER, 3, "Value");
    viCheckErr (viQueryf (instrSession, "%s\n", "%le", command, value));
    checkErr( rscmwgg_CheckStatus (instrSession)); 
Error:
    (void) Rs_UnlockSession(instrSession, VI_NULL);
    return error;
}


/*****************************************************************************
 * Function: Query ViString
 * Purpose:  This function queries the ViString value.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_QueryViString (ViSession instrSession,
                                          ViString command, 
                                          ViInt32 bufferSize,
                                          ViChar _VI_FAR value[])
{
    ViStatus    error = VI_SUCCESS;
    ViChar      *pbuffer = NULL;
    ViUInt32        bytesRead;
    
    checkErr( Rs_LockSession (instrSession, VI_NULL)); 
    if (strlen(command)==0)
        viCheckParm (RS_ERROR_INVALID_PARAMETER, 2, "Command");
    if (value==VI_NULL)
        viCheckParm (RS_ERROR_INVALID_PARAMETER, 3, "Value");
    viCheckErr (viPrintf (instrSession, "%s\n", command));
    viCheckErr(Rs_ReadDataUnknownLength (instrSession, &pbuffer, &bytesRead));  
    if (bytesRead>=(ViUInt32)bufferSize)
    {
        strncpy(value, pbuffer, (size_t) (bufferSize-1));
        value[bufferSize-1]='\0';
        error = (ViStatus) (bytesRead+1);
    } 
    else
    {
        strncpy(value, pbuffer, (size_t) bytesRead); 
        value[bytesRead]='\0';   
    }
    checkErr( rscmwgg_CheckStatus (instrSession)); 
Error:
    if (pbuffer) free (pbuffer);
    (void) Rs_UnlockSession(instrSession, VI_NULL);
    return error;
}

/*****************************************************************************
 * Function: rscmwgg_close
 * Purpose:  This function closes the instrument.
 *
 *           Note:  This function must unlock the session before calling
 *           Rs_Dispose.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_close (ViSession instrSession)
{
    ViStatus    error   = VI_SUCCESS;

    checkErr( Rs_LockSession (instrSession, VI_NULL));
    checkErr( rscmwgg_RsClose (instrSession));

Error:

    (void) Rs_UnlockSession (instrSession, VI_NULL);
    (void) Rs_Dispose (instrSession);

    return error;
}


/*****************************************************************************
 * Function: rscmwgg_RsClose
 * Purpose:  This function performs all of the drivers clean-up operations
 *           except for closing the RS session.  This function is called by
 *           rscmwgg_close or by an RS class driver.
 *
 *           Note:  This function must close the I/O session and set
 *           RS_ATTR_IO_SESSION to 0.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgg_RsClose (ViSession instrSession)
{
    ViStatus    error   = VI_SUCCESS;

    /* Do not lock here. The caller manages the lock. */

    if(instrSession != 0)
        {
        ViSession   rmSession = VI_NULL;

        checkErr( Rs_GetAttribute (instrSession, VI_NULL, RS_ATTR_VISA_RM_SESSION, 0, (ViInt32) sizeof (rmSession), &rmSession));
        
        viPrintf (instrSession, "&GTL");
        
        (void) viClose (instrSession);
        (void) viClose (rmSession);
        }
Error:    
    return error;
}

/*****************************************************************************
 *------------------------ Global Session Callbacks -------------------------*
 *****************************************************************************/

/*****************************************************************************
 * Function: Rs_CheckStatusCallback
 * Purpose:  This function queries the instrument to determine if it has
 *           encountered an error. If the instrument has encountered an
 *           error, this function returns the RS_ERROR_INSTRUMENT_SPECIFIC
 *           error code. This function is called by the
 *           Rs_CheckStatus utility function. The
 *           Rs_SetAttribute and Rs_GetAttribute functions invoke this
 *           function when the optionFlags parameter includes the
 *           RS_VAL_DIRECT_USER_CALL flag.
 *
 *           The user can disable calls to this function by setting the
 *           RSCMWGG_QUERY_INSTRUMENT_STATUS attribute to VI_FALSE.  The driver can
 *           disable the check status callback for a particular attribute by
 *           setting the RS_VAL_DONT_CHECK_STATUS flag.
 *
 *****************************************************************************/
ViStatus rscmwgg_CheckStatusCallback (ViSession instrSession)
{
    ViUInt32    ret_cnt;
    ViStatus    error   = VI_SUCCESS;
    ViInt16     esr     = 0;
    ViChar      buffer[256];
    ViInt32     code = 0;
 
    viCheckErr (viWrite (instrSession, (ViBuf) "*STB?\n", 6, NULL));
    viCheckErr (viRead (instrSession, (ViPBuf) buffer, 4, &ret_cnt));
    buffer[ret_cnt] = '\0';
    esr = (ViInt16) atol (buffer);
 
    if ((esr & IEEE_488_2_ERROR_MASK) != 0)
    {
        checkErr (Rs_SetErrorInfo (instrSession, VI_TRUE, RS_ERROR_INSTR_SPECIFIC, code, buffer));
 
        viCheckErr (RS_ERROR_INSTR_SPECIFIC);
    }
 
Error:
 
    return error;

}

/*****************************************************************************
 * Function: Rs_WaitForOPCCallback
 * Purpose:  This function waits until the instrument has finished processing
 *           all pending operations. This function is called by the
 *           Rs_WaitForOPC utility function. The RS engine invokes
 *           this function in the following two cases:
 *           - Before invoking the read callback for attributes for which the
 *             RS_VAL_WAIT_FOR_OPC_BEFORE_READS flag is set.
 *           - After invoking the write callback for attributes for which the
 *             RS_VAL_WAIT_FOR_OPC_AFTER_WRITES flag is set.
 *
 *****************************************************************************/
ViStatus rscmwgg_WaitForOPCCallback (ViSession instrSession)
{
    ViStatus    error = VI_SUCCESS;
    ViUInt16    response    = 0;
    //ViInt32     counter     = 0;
    ViInt32     opc_timeout = 0;
    clock_t     start_time = clock();
	ViReal64    opc_timeout_s = 0.0;
    
    checkErr( Rs_GetAttribute (instrSession, NULL, RS_ATTR_OPC_TIMEOUT, 0,
        (ViInt32) sizeof (opc_timeout), &opc_timeout));
	
	opc_timeout_s = (ViReal64) opc_timeout / 1000;

    /* Check STB for errors. */
    do
    {
		error = viReadSTB (instrSession, &response);

        if ((((double)(clock() - start_time))/(double)CLOCKS_PER_SEC) >= opc_timeout_s)
            return VI_ERROR_TMO;
		else
			Sleep (0);

        /* Wait while STB is non zero value. Only Error/Event Queue
           and ESB bits are taken into consideration. */
    
    } while ((response & IEEE_488_2_SRE_MASK) == 0);
    /* (check Error Queue bit in STB) */
    error = ((response & IEEE_488_2_ERROR_MASK) != 0) ? RS_ERROR_INSTRUMENT_STATUS : error;


Error:
    return error;
}
