/*****************************************************************************
 *  Rohde&Schwarz CMW500 Wideband Radio Communication Tester instrument driver
 *  GSM Measurements
 *  Original Release: February 2008
 *  By: Helena Kominkova
 *
 *  Should you have any technical questions please contact the hotline of
 *  Rohde & Schwarz Vertriebs-GmbH Rohde & Schwarz Support Center
 *
 *  e-mail: CustomerSupport@rohde-schwarz.com
 *
 *  Modification History:
 *    see ChangeLog
 *
 *****************************************************************************/
#include "rscmwgm.h"

#if (defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__NT__))
#define ANSI
#endif
#ifndef ANSI                /* UNIX compatible */
    #include <sys/timeb.h>
    #include <sys/types.h>
    #include <unistd.h>
    #define Sleep sleep
#else                       /* ANSI compatible */
#include <windows.h>
#endif

#if defined (_CVI_)
    #include <utility.h>
#endif
/*****************************************************************************
 *- Value Definition and Range Tables ---------------------------------------*
 *****************************************************************************/
static ViString mevMviewBurstTypeArr[] = {"OFF", "ANY", "GMSK", "EPSK", "ACC", "QAM16",NULL};
static ViString mevPatternTypeArr[] = {"OFF", "GMSK", "EPSK",NULL};
static ViString modulationSchemeArr[] = {"GMSK", "EPSK", "QAM16",NULL};
static ViString PVTSectionArr[] = {"REDG", "UPAR", "FEDG",NULL};
static ViString resultTypeArr[] = {"CURR", "AVER", "MAX", "MIN", "SDEV",NULL};
static ViString listModeBurstTypeArr[] = {"OFF", "GMSK", "EPSK", "ACC", NULL};
static ViString BERLoopArr[] = {"C", "SRB", NULL};
static ViString analyzerSalScenRXCArr[] = {
	"RF1C", "RF2C", "RF3C", "RF4C",
	"RFAC", "RFBC", "RF2I", "RF4I", "RFBI",
	"IQ1I", "IQ3I", "IQ5I", "IQ7I", "R11",
	"R12", "R13", "R14", "R15", "R16", "R17",
	"R18", "R21", "R22", "R23", "R24", "R25",
	"R26", "R27", "R28", "R31", "R32", "R33",
	"R34", "R35", "R36", "R37", "R38", "R41",
	"R42", "R43", "R44", "R45", "R46", "R47",
	"R48", "RA1", "RA2", "RA3", "RA4", "RA5",
	"RA6", "RA7", "RA8", "RB1", "RB2", "RB3",
	"RB4", NULL
};
static ViString analyzerSalScenRFCArr[] = {"RX1", "RX2", "RX3", "RX4", NULL};
static ViString scenarioArr[] = {"SAL", "CSP", "MAPR", NULL};
static ViString detectorTypeArr[] = {"RMS", "PEAK",NULL};
static ViString retriggerArr[] = {"OFF", "ON", "IFP", NULL};

/*****************************************************************************
 *- Callback Declarations ---------------------------------------------------*
 *****************************************************************************/
ViStatus rscmwgm_CheckStatusCallback(ViSession io);
ViStatus rscmwgm_WaitForOPCCallback(ViSession io);
ViStatus _VI_FUNC rscmwgm_RsClose(ViSession vi);

/*****************************************************************************
 *- User-Callable Functions (Exportable Functions) --------------------------*
 *****************************************************************************/

/*****************************************************************************
 * Function: rscmwgm_init
 * Purpose:  VXIplug&play required function. Calls the
 *           rscmwgm_InitWithOptions function.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_init(ViRsrc resourceName, ViBoolean IDQuery,
                               ViBoolean resetDevice, ViSession *newInstrSession)
{
	ViStatus error = VI_SUCCESS;

	if (newInstrSession == NULL)
	{
		(void)RsCore_SetErrorInfo(0, VI_FALSE, RS_ERROR_INVALID_PARAMETER,
		                          VI_ERROR_PARAMETER4, "Null address for Instrument Handle");

		checkErr(RS_ERROR_INVALID_PARAMETER);
	}

	checkErr(rscmwgm_InitWithOptions (resourceName, IDQuery, resetDevice, "", newInstrSession));

Error:
	return error;
}

/*****************************************************************************
 * Function: rscmwgm_InitWithOptions
 * Purpose:  This function creates a new RS session and calls the
 *           RsInit function.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_InitWithOptions(
	ViRsrc resourceName,
	ViBoolean IDQuery,
	ViBoolean resetDevice,
	ViString optionString,
	ViSession *newInstrSession
)
{
	ViStatus error = VI_SUCCESS;
	RsCoreAttributePtr *attrList = NULL;
	ViSession instrSession = 0;
	RsCoreSessionPtr sessionProperties = NULL;

	if (newInstrSession == NULL)
	{
		(void)RsCore_SetErrorInfo(0, VI_FALSE, RS_ERROR_INVALID_PARAMETER, VI_ERROR_PARAMETER5,
		                          "Null address for Instrument Handle");
		checkErr(RS_ERROR_INVALID_PARAMETER);
	}

	attrList = (RsCoreAttributePtr *)rscmwgm_RsCoreAttributeList;

	*newInstrSession = 0;

	checkErr(RsCore_NewSpecificDriver(
        resourceName, // VISA Resource Name
        "rscmwgm", // driver prefix
        optionString, // Init Options string, is applied to the session settings
        attrList, // List of all attributes from the rsxxx_AttrPropertiesList
        VI_FALSE, //idnModelFullName Model has the full *IDN? query name e.g. true: "CMW100A", VI_FALSE: "CMW" Used in CheckInstrument()
        0, // WriteDelay
        0, // ReadDelay
        100000, // IO Segment Size
        RS_VAL_OPCWAIT_STBPOLLING, // OPC Wait Mode
        RSCMWGM_OPC_TIMEOUT, // OPC timeout
        15000, // VISA Timeout
        600000, // Self-test timeout
        RS_VAL_BIN_FLOAT_FORMAT_SINGLE_4BYTES, // BinaryFloatNumbersFormat
        RS_VAL_BIN_INTEGER_FORMAT_INT32_4BYTES, // binaryIntegerNumbersFormat
        &instrSession));

	// Minimal allowed rscore version is 3.4.0
	checkErr(RsCore_CheckMinimalEngineVersion(instrSession, 3, 4, 0));

	checkErr(RsCore_GetRsSession(instrSession, &sessionProperties));

	// No SCPI command has been sent yet.
	// Call viClear before sending any SCPI command
	checkErr(RsCore_ViClear(instrSession));

	/* --- Here perform settings that are default for this driver,
	but can be overwritten by the optionsString settings */

	sessionProperties->autoSystErrQuery = VI_FALSE;

	// Parse option string and optionally sets the initial state of the following session attributes
	checkErr(RsCore_ApplyOptionString(instrSession, optionString));
	checkErr(RsCore_BuildRepCapTable(instrSession, rscmwgm_RsCoreRepCapTable));

	// Query *IDN? string, parse it and set the following attributes:
	// -RS_ATTR_INSTRUMENT_MANUFACTURER
	// - RS_ATTR_INSTRUMENT_MODEL - based on the idnModelFullName true: "CMW100A", false : "CMW"
	// - RS_ATTR_INSTRUMENT_FIRMWARE_REVISION
	checkErr(RsCore_QueryAndParseIDNstring(instrSession, RSCMWGM_SIMULATION_ID_QUERY, NULL));

	// Query *OPT? string, parse the options, remove the duplicates, sort them and store the result string to RS_ATTR_OPTIONS_LIST
	checkErr(RsCore_QueryAndParseOPTstring(instrSession, RSCMWGM_SIMULATION_OPT_QUERY, RS_VAL_OPTIONS_PARSING_AUTO));

	// Default Instrument Setup + optional *RST
	if (resetDevice == VI_TRUE)
	{
		checkErr(rscmwgm_reset(instrSession));
	}
	else
	{
		checkErr(rscmwgm_DefaultInstrSetup(instrSession));
	}

	if (IDQuery == VI_TRUE)
	{
		checkErr(RsCore_FitsIDNpattern(instrSession, RSCMWGM_VALID_ID_RESPONSE_STRING, NULL));
	}

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	if (error < VI_SUCCESS)
	{
		if (error == RS_ERROR_INSTRUMENT_STATUS && sessionProperties->autoSystErrQuery == VI_FALSE)
		{
			sessionProperties->autoSystErrQuery = VI_TRUE;
			(void)rscmwgm_CheckStatus(instrSession);
			sessionProperties->autoSystErrQuery = VI_FALSE;
		}

		if (instrSession != 0)
			(void)RsCore_ViClose(instrSession);
	}
	else
	{
		*newInstrSession = instrSession;
	}

	return error;
}

/*****************************************************************************
 * Function: Configure Analyzer Measure Protocol Scenario
 * Purpose:  This function activates the Measure@ProtocolTest scenario.
 *           The signal routing and analyzer settings are ignored by
 *           the measurement application. The corresponding settings have
 *           to be configured within the protocol test application used in
 *           parallel.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureAnalyzerMeasureProtocolScenario(
							ViSession instrSession,
							ViInt32   instance)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViString(instrSession, repCap, RSCMWGM_ATTR_MEASURE_PROTOCOL_SCENARIO, NULL),
			3, "Measure Protocol Scenario");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/// HIFN  This function activates the Measure@ProtocolTest scenario. The signal
/// HIFN  routing and analyzer settings are ignored by the measurement
/// HIFN  application. The corresponding settings have to be
/// HIFN     configured within the protocol test application used in parallel.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR application/String parameter selecting the protocol test application.
ViStatus _VI_FUNC rscmwgm_ConfigureAnalyzerMeasureProtocolScenarioApplication(
							ViSession instrSession,
							ViInt32   instance,
							ViString  application)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViString(instrSession, repCap, RSCMWGM_ATTR_MEASURE_PROTOCOL_SCENARIO_APPLICATION, application),
			3, "Application");

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Configure AnalyzerExternal Attenuation
 * Purpose:  This function defines an external attenuation value.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureAnalyzerExternalAttenuation(
							ViSession instrSession,
							ViInt32   instance,
							ViReal64  externalAttenuation)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViReal64(instrSession, repCap, RSCMWGM_ATTR_INPUT_EXTERNAL_ATTENUTATION, externalAttenuation),
			3, "Input External Attenutation");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Analyzer
 * Purpose:  Configures the RF input path.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureAnalyzer(ViSession instrSession,
                                            ViInt32   instance,
                                            ViReal64  expectedNominalPower,
                                            ViReal64  margin,
                                            ViReal64  centerFrequency)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViReal64(instrSession, repCap, RSCMWGM_ATTR_ANALYZER_EXPECTED_NOMINAL_POWER, expectedNominalPower),
			3, "Expected Nominal Power");

	viCheckParm(rscmwgm_SetAttributeViReal64(instrSession, repCap, RSCMWGM_ATTR_ANALYZER_USER_MARGIN, margin),
			4, "Margin");

	viCheckParm(rscmwgm_SetAttributeViReal64(instrSession, repCap, RSCMWGM_ATTR_ANALYZER_CENTER_FREQUENCY, centerFrequency),
			5, "Center Frequency");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Analyzer Stand Alone Scenario
 * Purpose:  This function activates the standalone scenario and selects the
 *           input connector for the measured RF signal.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureAnalyzerStandAloneScenario(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   RFConnector,
							ViInt32   RFConverter)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, RFConnector, RSCMWGM_VAL_ANALYZER_CONN_RF1C, RSCMWGM_VAL_ANALYZER_CONN_RB4),
			3, "RF Connector");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, RFConverter, RSCMWGM_VAL_ANALYZER_SAL_CONV_RX1, RSCMWGM_VAL_ANALYZER_SAL_CONV_RX4),
			4, "RF Convertor");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "ROUT:GSM:MEAS%ld:SCEN:SAL %s,%s", instance, analyzerSalScenRXCArr[RFConnector], analyzerSalScenRFCArr[RFConverter]);
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Configure Analyzer Combined Signal Path Scenario
 * Purpose:  This function activates the combined signal path scenario and selects a
 *           master. The master controls the signal routing settings,
 *           analyzer settings and UE signal info settings while the combined
 *           signal path scenario is active.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureAnalyzerCombinedSignalPathScenario(
							ViSession instrSession,
							ViInt32   instance,
							ViString  master)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViString(instrSession, repCap, RSCMWGM_ATTR_COMBINED_SIGNAL_PATH_SCENARIO, master),
			3, "Master");

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Configures the analyzer frequency by means of band and channel.
/// HIRET Document return value here.
/// HIRET You may use multiple lines for documentation.
/// HIPAR instrSession/Document parameter here.
/// HIPAR instance/Document parameter here.
/// HIPAR band/Document parameter here.
/// HIPAR channel/Document parameter here.
ViStatus _VI_FUNC rscmwgm_ConfigureAnalyzerChannelAndBand(ViSession instrSession,
                                                          ViInt32   instance,
                                                          ViInt32   band,
                                                          ViInt32   channel)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_ANALYZER_BAND, band),
			3, "Band");

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_ANALYZER_CHANNEL, channel),
			4, "Channel");

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Configure Analyzer Mixer Input Level
 * Purpose:  Varies the input level of the mixer in the analyzer path.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureAnalyzerMixerInputLevel(ViSession instrSession,
                                                           ViInt32   instance,
                                                           ViReal64  mixerLevelOffset)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViReal64(instrSession, repCap, RSCMWGM_ATTR_ANALYZER_MIXER_INPUT_LEVEL, mixerLevelOffset),
			3, "Mixer Level Offset");

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}


/// HIFN Specifies a positive or negative frequency offset to be added to
/// HIFN the center frequency of the configured channel.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator
/// HIPAR frequencyOffset/Sets the frequency offset.
ViStatus _VI_FUNC rscmwgm_ConfigureAnalyzerFrequencyOffset(ViSession instrSession,
                                                           ViInt32   instance,
                                                           ViReal64  frequencyOffset)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViReal64(instrSession, repCap, RSCMWGM_ATTR_ANALYZER_FREQUENCY_OFFSET, frequencyOffset),
			3, "Frequency Offset");

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Query Signal Routing
 * Purpose:  Returns the configured routing settings.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_QuerySignalRouting(ViSession instrSession,
                                             ViInt32   instance,
                                             ViInt32   *scenario,
                                             ViChar    controller[],
                                             ViInt32   *RFConnector,
                                             ViInt32   *RFConverter)
{
	ViStatus error = VI_SUCCESS;
	ViChar   *pbuffer = NULL;
	ViChar   *p2buf;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, controller),
			4, "Controller");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "ROUT:GSM:MEAS%ld", instance);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &pbuffer));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(pbuffer, ",\r\n");
	*scenario = RsCore_FindStringIndex(scenarioArr, p2buf);

	p2buf = strtok(NULL, ",\r\n");
	strcpy(controller, p2buf);

	p2buf = strtok(NULL, ",\r\n");
	*RFConnector = RsCore_FindStringIndex(analyzerSalScenRXCArr, p2buf);

	p2buf = strtok(NULL, ",\r\n");
	*RFConverter = RsCore_FindStringIndex(analyzerSalScenRFCArr, p2buf);

Error:
	if (pbuffer) free(pbuffer);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement Statistics
 * Purpose:  Configures the scope of the multi-evaluation measurement.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasStatistics(ViSession instrSession,
                                                       ViInt32   instance,
                                                       ViInt32   repetition,
                                                       ViInt32   stopCondition)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_REPETITION_MODE, repetition),
			3, "Repetition");

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_STOP_CONDITION, stopCondition),
			4, "Stop Condition");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement Statistics Count
 * Purpose:  Configures the statistics count. The statistic count (also termed
 *           statistic length) is the integer number of measurement intervals
 *           per measurement cycle (statistics cycle, single-shot measurement).
 *           The length of a measurement interval is measurement-specific.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasStatisticsCount(ViSession instrSession,
                                                            ViInt32   instance,
                                                            ViInt32   measurement,
                                                            ViInt32   statisticCount)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instance_mevMeasRepCap(instrSession, instance, measurement, repCap));

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_COUNT, statisticCount),
			4, "Statistic Count");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement Statistics Count
 * Purpose:  Configures the statistics count. The statistic count (also termed
 *           statistic length) is the integer number of measurement intervals
 *           per measurement cycle (statistics cycle, single-shot measurement).
 *           The length of a measurement interval is measurement-specific.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasMeasuredSlot(ViSession instrSession,
                                                         ViInt32   instance,
                                                         ViInt32   slotOffset,
                                                         ViInt32   no_ofSlot,
                                                         ViInt32   measuredSlot)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, slotOffset, 0, 7),
			3, "Slot Offset");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, no_ofSlot, 1, 8),
			4, "No of Slot");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, measuredSlot, 0, 7),
			5, "Measured Slot");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CONF:GSM:MEAS%ld:MEV:MSL %ld,%ld,%ld", instance, slotOffset, no_ofSlot, measuredSlot);
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement Statistics Count
 * Purpose:  Configures the statistics count. The statistic count (also termed
 *           statistic length) is the integer number of measurement intervals
 *           per measurement cycle (statistics cycle, single-shot measurement).
 *           The length of a measurement interval is measurement-specific.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasResults(ViSession instrSession,
                                                    ViInt32   instance,
                                                    ViInt32   measurement,
                                                    ViBoolean state)
{
	ViStatus   error = VI_SUCCESS;
	ViChar     repCap[RS_REPCAP_BUF_SIZE] = "";
	ViString   mevResArr[] = {
	"mevPVT",  "mevEVM", "mevMERR", "mevPERR", "mevSMFR", "mevSMT",
	"mevSSFR", "mevSST", "mevAMPM", "mevBER", "mevIQ", "mevMSC"
	           };

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, measurement, 0, 11),
			3, "Measurement");

	snprintf(repCap, RS_REPCAP_BUF_SIZE, "i%ld,%s", instance, mevResArr[measurement]);

	viCheckParm(rscmwgm_SetAttributeViBoolean(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_RESULT, state),
			4, "State");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement Statistics Count
 * Purpose:  Configures the analyzer of the R&S CMW 500 to
 *           the properties of the analyzed signal.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasParameters(ViSession instrSession,
                                                       ViInt32   instance,
                                                       ViInt32   frequencyRange,
                                                       ViBoolean signalLevelState,
                                                       ViReal64  signalLevel,
                                                       ViBoolean accessBurstMeasurement,
                                                       ViInt32   trainingSequence,
                                                       ViBoolean measure_onException)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_FREQUENCY_RANGE, frequencyRange),
			3, "Frequency Range");

	viCheckParm(rscmwgm_SetAttributeViBoolean(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_SIGNAL_LEVEL_RELATIVE_STATE, signalLevelState),
			4, "Signal Level State");
	if (signalLevelState == VI_TRUE)
	{
		viCheckParm(rscmwgm_SetAttributeViReal64(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_SIGNAL_LEVEL_RELATIVE, signalLevel),
				5, "Signal Level");
	}
	viCheckParm(rscmwgm_SetAttributeViBoolean(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_ACCESS_BURST, accessBurstMeasurement),
			6, "Access Burst Measurement");

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_TRAINING_SEQUENCE, trainingSequence),
			7, "Training Sequence");

	viCheckParm(rscmwgm_SetAttributeViBoolean(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_RESULTS_REJECTION, measure_onException),
			8, "Measure on Exception");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/// HIFN Defines a timeout for the measurement. The timer is started when the measurement is initiated via a READ or INIT command. It is not started if the measurement is initiated manually (ON/OFF key or RESTART/STOP key).
/// HIFN When the measurement has completed the first measurement cycle (first single shot), the statistical depth is reached and the timer is reset.
/// HIFN If the first measurement cycle has not been completed when the timer expires, the measurement is stopped. The measurement state changes to RDY and the reliability indicator is set to 1, indicating that a measurement timeout occurred. Still running READ, FETCh or CALCulate commands are completed, returning the available results. At least for some results, there are no values at all or the statistical depth has not been reached.
/// HIRET Document return value here.
/// HIRET You may use multiple lines for documentation.
/// HIPAR instrSession/Document parameter here.
/// HIPAR instance/Document parameter here.
/// HIPAR timeout/Document parameter here.
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasTimeout(ViSession instrSession,
                                                    ViInt32   instance,
                                                    ViReal64  timeout)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViReal64(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_TIMEOUT, timeout),
			3, "Timeout");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement Modulation Scheme
 * Purpose:  Defines the expected modulation scheme and burst type in all
 *           timeslots and adjusts the power/time template accordingly.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasModulationScheme(ViSession instrSession,
                                                             ViInt32   instance,
                                                             ViInt32   slot0,
                                                             ViInt32   slot1,
                                                             ViInt32   slot2,
                                                             ViInt32   slot3,
                                                             ViInt32   slot4,
                                                             ViInt32   slot5,
                                                             ViInt32   slot6,
                                                             ViInt32   slot7)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, slot0, 0, 5),
			3, "slot 0");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, slot1, 0, 5),
			4, "slot 1");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, slot2, 0, 5),
			5, "slot 2");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, slot3, 0, 5),
			6, "slot 3");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, slot4, 0, 5),
			7, "slot 4");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, slot5, 0, 5),
			8, "slot 5");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, slot6, 0, 5),
			9, "slot 6");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, slot7, 0, 5),
			10, "slot 7");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CONF:GSM:MEAS%ld:MEV:MVI %s,%s,%s,%s,%s,%s,%s,%s", instance,
	         mevMviewBurstTypeArr[slot0], mevMviewBurstTypeArr[slot1], mevMviewBurstTypeArr[slot2],
	         mevMviewBurstTypeArr[slot3], mevMviewBurstTypeArr[slot4], mevMviewBurstTypeArr[slot5],
	         mevMviewBurstTypeArr[slot6], mevMviewBurstTypeArr[slot7]);
	checkErr(RsCore_Write(instrSession, cmd));
	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement 16QAM NB Search
 * Purpose:  This function configures the search for 16-QAM-modulated normal
 *           bursts (with option R&S CMW-KM701).
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeas16QAMNBSearch(ViSession instrSession,
                                                          ViInt32   instance,
                                                          ViBoolean NBSearchState)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViBoolean(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_16QAM_NB_SEARCH, NBSearchState),
			3, "NB Search State");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement VAMOS Training Sequence Code
 * Purpose:  Specifies the expected VAMOS Training Sequence Code (TSC) set of
 *           the measured GSM uplink signal. With a specific TSC set
 *           selection, the R&S CMW500 will analyze bursts with this TSC set
 *           only.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasVAMOSTSC(ViSession instrSession,
                                                     ViInt32   instance,
                                                     ViInt32   TSCSet)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_VAMOS_TSC, TSCSet),
			3, "TSC Set");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}


/// HIFN This function configures modulation decoding.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR type/Selects the loop for BER tests.
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasDecode(ViSession instrSession,
                                                   ViInt32   instance,
                                                   ViInt32   type)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_DECODE, type),
			3, "Type");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement Trigger
 * Purpose:  Configures the trigger parameters.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasTrigger(ViSession instrSession,
                                                    ViInt32   instance,
                                                    ViString  source,
                                                    ViInt32   delay,
                                                    ViBoolean timeoutState,
                                                    ViReal64  timeout,
                                                    ViReal64  threshold,
                                                    ViInt32   slope)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViString(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_TRIGGER_SOURCE, source),
			3, "Source");

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_TRIGGER_DELAY, delay),
			4, "Delay");

	viCheckParm(rscmwgm_SetAttributeViBoolean(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_TRIGGER_TIMEOUT_STATE, timeoutState),
			5, "Timeout State");
	if (timeoutState == VI_TRUE)
	{
		viCheckParm(rscmwgm_SetAttributeViReal64(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_TRIGGER_TIMEOUT, timeout),
				6, "Timeout");
	}
	viCheckParm(rscmwgm_SetAttributeViReal64(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_TRIGGER_THRESHOLD, threshold),
			7, "Threshold");

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_TRIGGER_SLOPE, slope),
			8, "Slope");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Query Multi Eval Measurement Trigger Source Catalog
 * Purpose:  This function lists all trigger source values that can be set
 *           using TRIGger:GSM:MEAS<instance>:SOURce.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasTriggerSourceCatalog(ViSession instrSession,
                                                             ViInt32   instance,
                                                             ViInt32   arraySize,
                                                             ViChar    catalog[])
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, catalog), 7, "catalog");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "TRIG:GSM:MEAS%ld:MEV:CAT:SOUR?", instance);
	checkErr(RsCore_QueryViStringUnknownLengthToUserBuffer(instrSession, cmd, arraySize, catalog, NULL));
	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement Frame Method
 * Purpose:  This function configures the analyzer of the R&S CMW 500 frame
 *           synchronization method.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasFrameMethod(ViSession instrSession,
                                                        ViInt32   instance,
                                                        ViInt32   method)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_FRAME_SYNCHRONIZATION_METHOD, method),
			3, "Method");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement Frame Gap Length
 * Purpose:  This function configures the analyzer of the R&S CMW 500 frame
 *           synchronization gap length.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasFrameGapLength(ViSession instrSession,
                                                           ViInt32   instance,
                                                           ViInt32   gapLength)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_GAP_LENGTH, gapLength),
			3, "Gap Length");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement Trigger Gap
 * Purpose:  Defines a minimum duration of the power-down periods (gaps)
 *           between two triggered IF power pulses.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasTriggerGap(ViSession instrSession,
                                                       ViInt32   instance,
                                                       ViInt32   gap)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_TRIGGER_GAP, gap),
			3, "Gap");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/// HIFN Enables feature ignore initial off frames to avoid trigger timeout in access burst measurement in idle mode.
/// HIRET Document return value here.
/// HIRET You may use multiple lines for documentation.
/// HIPAR instrSession/Document parameter here.
/// HIPAR instance/Document parameter here.
/// HIPAR ignoreInitialOffFrames/Document parameter here.
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasIgnoreInitialOffFrames(
							ViSession instrSession,
							ViInt32   instance,
							ViBoolean ignoreInitialOffFrames)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViBoolean(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_IGNORE_INITIAL_OFF_FRAMES, ignoreInitialOffFrames),
			3, "Ignore Initial Off Frames");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement Frame Pattern
 * Purpose:  This function configures the analyzer of the R&S CMW 500 frame
 *           synchronization pattern.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasFramePattern(ViSession instrSession,
                                                         ViInt32   instance,
                                                         ViInt32   pattern0,
                                                         ViInt32   pattern1,
                                                         ViInt32   pattern2,
                                                         ViInt32   pattern3,
                                                         ViInt32   pattern4,
                                                         ViInt32   pattern5,
                                                         ViInt32   pattern6,
                                                         ViInt32   pattern7)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, pattern0, 0, 2),
			3, "Pattern 0");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, pattern1, 0, 2),
			4, "Pattern 1");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, pattern2, 0, 2),
			5, "Pattern 2");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, pattern3, 0, 2),
			6, "Pattern 3");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, pattern4, 0, 2),
			7, "Pattern 4");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, pattern5, 0, 2),
			8, "Pattern 5");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, pattern6, 0, 2),
			9, "Pattern 6");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, pattern7, 0, 2),
			10, "Pattern 7");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CONF:GSM:MEAS%ld:MEV:APAT %s,%s,%s,%s,%s,%s,%s,%s", instance,
	         mevPatternTypeArr[pattern0], mevPatternTypeArr[pattern1], mevPatternTypeArr[pattern2],
	         mevPatternTypeArr[pattern3], mevPatternTypeArr[pattern4], mevPatternTypeArr[pattern5],
	         mevPatternTypeArr[pattern6], mevPatternTypeArr[pattern7]);
	checkErr(RsCore_Write(instrSession, cmd));
	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement Frame Pattern
 * Purpose:  This function configures the analyzer of the R&S CMW 500 frame
 *           synchronization pattern.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasPVT(ViSession instrSession,
                                                ViInt32   instance,
                                                ViInt32   refPowerMode,
                                                ViInt32   IFFilterBandwidth)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_REFERENCE_POWER_MODE, refPowerMode),
			3, "Ref Power Mode");

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_IF_FILTER_BANDWIDTH, IFFilterBandwidth),
			4, "IF Filter Bandwidth");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement Power vs. Time PCL
 * Purpose:  This function configures the PLC values in all timeslots.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasPVTPCL(ViSession instrSession,
                                                   ViInt32   instance,
                                                   ViInt32   arraySize,
                                                   ViInt32   PCLMode,
                                                   ViInt32   PCLValues[])
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *p2buffer = cmd;
	ViInt32  i = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, arraySize, 1, 32001),
			3, "Array Size");

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_PCL_MODE, PCLMode),
			3, "PCL Mode");

	if (PCLMode == RSCMWGM_VAL_PCL)
	{
		p2buffer += sprintf(p2buffer, "CONF:GSM:MEAS%ld:MEV:PCL ", instance);

		for (i = 0; i < 8; i++)
		{
			if (i < arraySize)
			{
				viCheckParm(RsCore_InvalidViInt32Range(instrSession, PCLValues[i], 0, 31),
						5, "PCL Values");

				p2buffer += sprintf(p2buffer, "%ld,", PCLValues[i]);
			}
			else
			{
				p2buffer += sprintf(p2buffer, "0,");
			}
		}
		*--p2buffer = '\0';

		checkErr(RsCore_Write(instrSession, cmd));
	}

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement BER
 * Purpose:  This function selects the loop for BER tests.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasBER(ViSession instrSession,
                                                ViInt32   instance,
                                                ViInt32   loop)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_BER_LOOP, loop),
			3, "Loop");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}


/*****************************************************************************
 * Function: Configure Multi Eval Measurement Threshold
 * Purpose:  This function selects the "Threshold Start" or "Threshold Run"
 *           value for BER tests.
 *           This is the maximum bit error rate in the first burst of the BER
 *           measurement.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasBERThreshold(ViSession instrSession,
                                                         ViInt32   instance,
                                                         ViInt32   thresholdType,
                                                         ViReal64  threshold)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, thresholdType, RSCMWGM_VAL_TRSH_START, RSCMWGM_VAL_TRSH_RUN),
			3, "Treshold Type");
	if (thresholdType == RSCMWGM_VAL_TRSH_START)
	{
		viCheckParm(rscmwgm_SetAttributeViReal64(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_BER_THRESHOLD_START, threshold),
				3, "Threshold Start");
	}
	else if (thresholdType == RSCMWGM_VAL_TRSH_RUN)
	{
		viCheckParm(rscmwgm_SetAttributeViReal64(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_BER_THRESHOLD_RUN, threshold),
				3, "Threshold Run");
	}
	else
	{
		viCheckParm(RsCore_InvalidViInt32Value(instrSession, thresholdType), 3, "Treshold Type");
	}

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}


/*****************************************************************************
 * Function: Configure Multi Eval Measurement IQ Constellation
 * Purpose:  This function configures I/Q constellation measurement.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasIQConstellation(ViSession instrSession,
                                                            ViInt32   instance,
                                                            ViInt32   rotation,
                                                            ViInt32   filter)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_IQ_ROTATION, rotation),
			3, "Rotation");

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_IQ_FILTER, filter),
			4, "Filter");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement Spectrum Modulation Freq Offset
 * Purpose:  This function defines the frequency offsets to be used for spectrum
 *           modulation measurements. The offsets are defined relative to
 *           the analyzer frequency. Up to 20 offsets can be defined and enabled.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasSpecModFOffset(ViSession instrSession,
                                                           ViInt32   instance,
                                                           ViInt32   arraySize,
                                                           ViBoolean status[],
                                                           ViReal64  frequencyOffset[])
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *p2buf;
	ViInt32  i;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, arraySize, 1, 32001),
			3, "Array Size");

	p2buf = cmd;
	p2buf += snprintf(p2buf, RS_MAX_MESSAGE_BUF_SIZE, "CONF:GSM:MEAS%ld:MEV:SMOD:OFR ", instance);
	for (i = 0; i < arraySize; i++)
	{
		if (status[i] == VI_TRUE)
			p2buf += sprintf(p2buf, "%.12lG,", frequencyOffset[i]);
		else
			p2buf += sprintf(p2buf, "OFF,");
	}

	*--p2buf = '\0';
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement Spectrum Modulation Evaluation Areas
 * Purpose:  This function defines the time intervals (evaluation areas) to be
 *           used for spectrum modulation measurements.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasSpecModEvalAreas(ViSession instrSession,
                                                             ViInt32   instance,
                                                             ViInt32   arraySize,
                                                             ViBoolean status[],
                                                             ViInt32   start_ofArea[],
                                                             ViInt32   stop_ofArea[])
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *p2buffer = cmd;
	ViInt32  i;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, arraySize, 1, 32001),
			3, "Array Size");

	p2buffer += sprintf(p2buffer, "CONF:GSM:MEAS%ld:MEV:SMOD:EAR ", instance);
	for (i = 0; i < arraySize; i++)
	{
		p2buffer += sprintf(p2buffer, "%s,%ld,%ld,", status[i] ? "ON" : "OFF", start_ofArea[i], stop_ofArea[i]);
	}
	*--p2buffer = '\0';
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement Spectrum Modulation Time Freq Offset
 * Purpose:  This function defines the offset frequency for the
 *           "Spectrum Modulation Time" diagram. The diagram shows the measured
 *           power vs. time at the selected offset frequency.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasSpecModTimeFreqOffset(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   frequencyOffset)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instance_mevMeasRepCap(instrSession, instance, 2, repCap));

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_OFFSET_FREQUENCY, frequencyOffset),
			3, "Frequency Offset");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement Spectrum Modulation Power Average
 *           Mode
 * Purpose:  This function defines the power average mode.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasSpecModPwrAverMode(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   powerAverageMode)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instance_mevMeasRepCap(instrSession, instance, 2, repCap));

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_POWER_AVERAGE, powerAverageMode),
			3, "Power Average Mode");

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement Spectrum Switching Freq Offset
 * Purpose:  This function defines the frequency offsets to be used for spectrum
 *           switching measurements. The offsets are defined relative to
 *           the analyzer frequency. Up to 20 offsets can be defined and enabled.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasSpecSwitchFOffset(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViBoolean status[],
							ViReal64  frequencyOffset[])
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *p2buffer = cmd;
	ViInt32  i;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, arraySize, 1, 32001),
			3, "Array Size");

	p2buffer += sprintf(p2buffer, "CONF:GSM:MEAS%ld:MEV:SSW:OFR ", instance);
	for (i = 0; i < arraySize; i++)
	{
		if (status[i] == VI_TRUE)
		{
			p2buffer += sprintf(p2buffer, "%.12lf,", frequencyOffset[i]);
		}
		else
		{
			p2buffer += sprintf(p2buffer, "OFF,");
		}
	}
	*--p2buffer = '\0';
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement Spectrum Switching
 * Purpose:  This function defines settings for spectrum switching frequency
 *           measurements.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasSpecSwitching(ViSession instrSession,
                                                          ViInt32   instance,
                                                          ViInt32   frequencyOffset,
                                                          ViInt32   holdMode)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instance_mevMeasRepCap(instrSession, instance, 3, repCap));

	snprintf(repCap, RS_REPCAP_BUF_SIZE, "i%ld,mevMeasSSW", instance);

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_OFFSET_FREQUENCY, frequencyOffset),
			3, "Frequency Offset");

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_PEAK_HOLD_MODE, holdMode),
			4, "Hold Mode");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}


/*****************************************************************************
 * Function: Configure Multi Eval Measurement List Mode State
 * Purpose:  Enables or disables the list mode.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMultiEvalMeasurementListModeState(
							ViSession instrSession,
							ViInt32   instance,
							ViBoolean state)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViBoolean(instrSession, repCap, RSCMWGM_ATTR_LIST_MODE_STATE, state),
			3, "State");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/// HIFN  Specifies whether a trigger event will initiate a measurement of the entire measurement interval
/// HIFN (comprising the number of segments defined via CONFigure:GSM:MEAS:MEValuation:LIST:LRANge) or a measurement of single segment.
/// HIRET Document return value here.
/// HIRET You may use multiple lines for documentation.
/// HIPAR instrSession/Document parameter here.
/// HIPAR instance/Document parameter here.
/// HIPAR triggerMode/Document parameter here.
ViStatus _VI_FUNC rscmwgm_ConfigureMultiEvalMeasurementListModeTriggerMode(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   triggerMode)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_LIST_MODE_TRIGGER_MODE, triggerMode),
			3, "Trigger Mode");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/// HIFN Defines the CMWS connector settings.
/// HIRET Document return value here.
/// HIRET You may use multiple lines for documentation.
/// HIPAR instrSession/Document parameter here.
/// HIPAR instance/Document parameter here.
/// HIPAR segment/Document parameter here.
/// HIPAR connectorMode/Document parameter here.
/// HIPAR CMWSConnector/Document parameter here.
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalMeasListModeCMWSConnector(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   connectorMode,
							ViInt32   CMWSConnector)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 2000),
			3, "Segment");

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));
	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_LIST_MODE_CONNECTOR_MODE, connectorMode),
			4, "Connector Mode");

	snprintf(repCap, RS_REPCAP_BUF_SIZE, "i%ld,seg%ld", instance, segment);
	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_LIST_MODE_CMWS_CONNECTOR, CMWSConnector),
			5, "CMWS Connector");

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN Select a range of measured segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR firstMeasuredSegments/First measured segment in the range of configured segments
/// HIPAR numberOfSegments/Relative number within the range of measured segments
ViStatus _VI_FUNC rscmwgm_ConfigureMultiEvalMeasurementListModeRange(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   firstMeasuredSegment,
							ViInt32   numberOfSegments)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, firstMeasuredSegment, 1, 2000),
			3, "First Measured Segment");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberOfSegments, 1, 512),
			4, "Number of Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CONF:GSM:MEAS%ld:MEV:LIST:LRAN %ld,%ld", instance, firstMeasuredSegment, numberOfSegments);
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement List Mode Segment Count
 * Purpose:  Configures list mode number of measured and displayed segments.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMultiEvalMeasurementListModeSegmentCount(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   measuredSegments,
							ViInt32   displayedSegments)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, displayedSegments, 1, 200),
			4, "Displayed Segments");

	if (displayedSegments > measuredSegments)
	{
		viCheckParm(RsCore_InvalidViInt32Value(instrSession, displayedSegments), 4, "Displayed Segments");
	}

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_LIST_MODE_COUNT, measuredSegments),
			3, "Measured Segments");

	if (displayedSegments == 0)
	{
		snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CONFigure:GSM:MEAS%ld:MEValuation:LIST:OSINdex OFF", instance);
		checkErr(RsCore_Write(instrSession, cmd));
	}
	else
	{
		snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CONFigure:GSM:MEAS%ld:MEValuation:LIST:OSINdex %ld", instance, displayedSegments);
		checkErr(RsCore_Write(instrSession, cmd));
	}

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/// HIFN  Configures whether idle frames are ignored or cause a "Signal low"
/// HIFN  error.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR ignore/Enables or disables the list mode.
ViStatus _VI_FUNC rscmwgm_ConfigureMultiEvalMeasurementListModeStateIgnoreIdleFrames(
							ViSession instrSession,
							ViInt32   instance,
							ViBoolean ignore)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViBoolean(instrSession, repCap, RSCMWGM_ATTR_LIST_MODE_IGNORE_IDLE_FRAMES, ignore),
			3, "Ignore");

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement List Mode Step Length
 * Purpose:  Selects the step length, i.e. the time difference between two
 *           measured TDMA timeslots. A step length of 1 means that every
 *           slot is measured, a step length of 8 means that a single
 *           timeslot per TDMA frame is measured.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMultiEvalMeasurementListModeStepLength(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   stepLength,
							ViBoolean stepLengthState)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	if (stepLengthState == VI_TRUE)
	{
		viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, repCap, RSCMWGM_ATTR_LIST_MODE_STEP_LENGTH, stepLength),
				3, "Step Length");
	}
	else
	{
		viCheckParm(rscmwgm_SetAttributeViBoolean(instrSession, repCap, RSCMWGM_ATTR_LIST_MODE_STEP_LENGTH_STATE, stepLengthState),
				3, "Step Length State");
	}

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement List Mode Segment
 * Purpose:  Defines the length, the analyzer settings, the expected PCL,
 *           retrigger setting and evaluation offset of a selected segment.
 *           In general this must be configured for all measured segments.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMultiEvalMeasurementListModeSegment(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   steps,
							ViReal64  nominalPower,
							ViReal64  frequency,
							ViInt32   powerControlLevel,
							ViInt32   retrigger,
							ViInt32   offset)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 2000),
			3, "Segment");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, steps, 1, 3000),
			4, "Steps");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, nominalPower, -47.0, 55.0), 5, "Nominal Power");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, frequency, 100.0E+6, 6000.0E+6), 6, "Frequency");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, powerControlLevel, 0, 31),
			7, "Power Control Level");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, retrigger, RSCMWGM_VAL_RETRIGGER_OFF, RSCMWGM_VAL_RETRIGGER_IFPOWER),
			8, "Retrigger");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, offset, 0, 1000),
			9, "Offset");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CONF:GSM:MEAS%ld:MEV:LIST:SEGM%ld:SET %ld,%.12lG,%.12lG,%ld,%s,%ld", instance,
	         segment,
	         steps,
	         nominalPower,
	         frequency,
	         powerControlLevel,
	         retriggerArr[retrigger],
	         offset);
	checkErr(RsCore_Write(instrSession, cmd));
	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement List Mode Modulation Results
 * Purpose:  Selects the modulation results to be measured in selected
 *           segment.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMultiEvalMeasurementListModeModulationResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   statistics,
							ViBoolean EVMState,
							ViBoolean magnitudeErrorState,
							ViBoolean phaseErrorState,
							ViBoolean AMPMDelayState,
							ViString  framePattern)
{
	ViStatus  error = VI_SUCCESS;
	ViChar    repCap[RS_REPCAP_BUF_SIZE] = "";
	ViChar    cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar    *p2buffer = cmd;
	ViBoolean framePatternState = VI_FALSE;

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 2000),
			3, "Segment");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, statistics, 1, 3000),
			4, "Statistics");

	p2buffer += sprintf(p2buffer,
	                    "CONFigure:GSM:MEAS%ld:MEValuation:LIST:SEGMent%ld:MODulation %ld",
	                    instance, segment, statistics);

	if (EVMState == VI_TRUE)
	{
		p2buffer += sprintf(p2buffer, ",ON");
	}
	else
	{
		p2buffer += sprintf(p2buffer, ",OFF");
	}

	if (magnitudeErrorState == VI_TRUE)
	{
		p2buffer += sprintf(p2buffer, ",ON");
	}
	else
	{
		p2buffer += sprintf(p2buffer, ",OFF");
	}

	if (phaseErrorState == VI_TRUE)
	{
		p2buffer += sprintf(p2buffer, ",ON");
	}
	else
	{
		p2buffer += sprintf(p2buffer, ",OFF");
	}

	if (AMPMDelayState == VI_TRUE)
	{
		p2buffer += sprintf(p2buffer, ",ON");
	}
	else
	{
		p2buffer += sprintf(p2buffer, ",OFF");
	}

	viCheckParm(rscmwgm_GetAttributeViBoolean(instrSession, repCap, RSCMWGM_ATTR_LIST_MODE_STEP_LENGTH_STATE, &framePatternState),
			7, "Frame Pattern");

	if (framePatternState == VI_FALSE)
	{
		sprintf(p2buffer, ",%s", framePattern);
	}

	checkErr(RsCore_Write(instrSession, cmd));
	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement List Mode PVT Results
 * Purpose:  Selects the power vs. time results to be measured in selected
 *           segment.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMultiEvalMeasurementListModePVTResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   statistics,
							ViBoolean PVTState,
							ViString  framePattern)
{
	ViStatus  error = VI_SUCCESS;
	ViChar    repCap[RS_REPCAP_BUF_SIZE] = "";
	ViChar    cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar    *p2buffer = cmd;
	ViBoolean framePatternState = VI_FALSE;

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 2000),
			3, "Segment");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, statistics, 1, 3000),
			4, "Statistics");

	p2buffer += sprintf(p2buffer,
	                    "CONFigure:GSM:MEAS%ld:MEValuation:LIST:SEGMent%ld:PVTime %ld",
	                    instance, segment, statistics);

	if (PVTState == VI_TRUE)
	{
		p2buffer += sprintf(p2buffer, ",ON");
	}
	else
	{
		p2buffer += sprintf(p2buffer, ",OFF");
	}

	viCheckParm(rscmwgm_GetAttributeViBoolean(instrSession, repCap, RSCMWGM_ATTR_LIST_MODE_STEP_LENGTH_STATE, &framePatternState),
			7, "Frame Pattern");

	if (framePatternState == VI_FALSE)
	{
		sprintf(p2buffer, ",%s", framePattern);
	}

	checkErr(RsCore_Write(instrSession, cmd));
	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement List Mode Spectrum Modulation Results
 * Purpose:  Selects the spectrum due to modulation results to be measured in
 *           selected segment.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMultiEvalMeasurementListModeSpectrumModulationResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   statistics,
							ViBoolean spectrumModulationState,
							ViString  framePattern)
{
	ViStatus  error = VI_SUCCESS;
	ViChar    repCap[RS_REPCAP_BUF_SIZE] = "";
	ViChar    cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar    *p2buffer = cmd;
	ViBoolean framePatternState = VI_FALSE;

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 2000),
			3, "Segment");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, statistics, 1, 3000),
			4, "Statistics");

	p2buffer += sprintf(p2buffer,
	                    "CONFigure:GSM:MEAS%ld:MEValuation:LIST:SEGMent%ld:SMODulation %ld",
	                    instance, segment, statistics);

	if (spectrumModulationState == VI_TRUE)
	{
		p2buffer += sprintf(p2buffer, ",ON");
	}
	else
	{
		p2buffer += sprintf(p2buffer, ",OFF");
	}

	viCheckParm(rscmwgm_GetAttributeViBoolean(instrSession, repCap, RSCMWGM_ATTR_LIST_MODE_STEP_LENGTH_STATE, &framePatternState),
			7, "Frame Pattern");

	if (framePatternState == VI_FALSE)
	{
		sprintf(p2buffer, ",%s", framePattern);
	}

	checkErr(RsCore_Write(instrSession, cmd));
	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement List Mode Spectrum Switching Results
 * Purpose:  Selects the spectrum due to switching results to be measured in
 *           selected segment.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMultiEvalMeasurementListModeSpectrumSwitchingResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   statistics,
							ViBoolean spectrumSwitchingState,
							ViString  framePattern)
{
	ViStatus  error = VI_SUCCESS;
	ViChar    repCap[RS_REPCAP_BUF_SIZE] = "";
	ViChar    cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar    *p2buffer = cmd;
	ViBoolean framePatternState = VI_FALSE;

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 2000),
			3, "Segment");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, statistics, 1, 3000),
			4, "Statistics");

	p2buffer += sprintf(p2buffer,
	                    "CONFigure:GSM:MEAS%ld:MEValuation:LIST:SEGMent%ld:SSWitching %ld",
	                    instance, segment, statistics);

	if (spectrumSwitchingState == VI_TRUE)
	{
		p2buffer += sprintf(p2buffer, ",ON");
	}
	else
	{
		p2buffer += sprintf(p2buffer, ",OFF");
	}

	viCheckParm(rscmwgm_GetAttributeViBoolean(instrSession, repCap, RSCMWGM_ATTR_LIST_MODE_STEP_LENGTH_STATE, &framePatternState),
			7, "Frame Pattern");

	if (framePatternState == VI_FALSE)
	{
		sprintf(p2buffer, ",%s", framePattern);
	}

	checkErr(RsCore_Write(instrSession, cmd));
	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Measurement List Mode BER
 * Purpose:  Selects the spectrum due to switching results to be measured in
 *           selected segment.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMultiEvalMeasurementListModeBER(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   statistics,
							ViBoolean BERState,
							ViInt32   BERLoop,
							ViString  framePattern)
{
	ViStatus  error = VI_SUCCESS;
	ViChar    repCap[RS_REPCAP_BUF_SIZE] = "";
	ViChar    cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar    *p2buffer = cmd;
	ViBoolean framePatternState = VI_FALSE;

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 2000),
			3, "Segment");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, statistics, 1, 100),
			4, "Statistics");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, BERLoop, RSCMWGM_VAL_BER_LOOP_C, RSCMWGM_VAL_BER_LOOP_SRB),
			6, "BER Loop");

	p2buffer += sprintf(p2buffer,
	                    "CONF:GSM:MEAS%ld:MEV:LIST:SEGM%ld:BER %ld",
	                    instance, segment, statistics);

	if (BERState == VI_TRUE)
	{
		p2buffer += sprintf(p2buffer, ",ON");
	}
	else
	{
		p2buffer += sprintf(p2buffer, ",OFF");
	}

	p2buffer += sprintf(p2buffer, ",%s", BERLoopArr[BERLoop]);

	viCheckParm(rscmwgm_GetAttributeViBoolean(instrSession, repCap, RSCMWGM_ATTR_LIST_MODE_STEP_LENGTH_STATE, &framePatternState),
			7, "Frame Pattern");

	if (framePatternState == VI_FALSE)
	{
		sprintf(p2buffer, ",%s", framePattern);
	}

	checkErr(RsCore_Write(instrSession, cmd));
	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Limits Errot Vector Magnitude
 * Purpose:  This function defines Error Vectro Magnitude limits for results
 *           which characterize the modulation accuracy for the modulation scheme
 *           GMSK or EPSK.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalLimitsEVM(ViSession instrSession,
                                                  ViInt32   instance,
                                                  ViInt32   modulationScheme,
                                                  ViReal64  RMSLimit,
                                                  ViBoolean RMSLimitStatus[],
                                                  ViReal64  peakLimit,
                                                  ViBoolean peakLimitStatus[],
                                                  ViReal64  _95Limit,
                                                  ViBoolean _95LimitStatus)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *p2buffer = cmd;
	ViInt32  i;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, modulationScheme, 0, 2),
			3, "Modulation Scheme");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, RMSLimit, 0.0, 50.0), 4, "RMS Limit");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, peakLimit, 0.0, 50.0), 6, "Peak Limit");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, _95Limit, 0.0, 50.0), 8, "95% Limit");

	p2buffer += sprintf(p2buffer, "CONF:GSM:MEAS%ld:MEV:LIM:%s:EVM %.12le,%.12le,%.12le,",
	                    instance, modulationSchemeArr[modulationScheme], RMSLimit, peakLimit, _95Limit);
	for (i = 0; i < 3; i++)
	{
		p2buffer += sprintf(p2buffer, "%s,", RMSLimitStatus[i] ? "ON" : "OFF");
	}
	for (i = 0; i < 3; i++)
	{
		p2buffer += sprintf(p2buffer, "%s,", peakLimitStatus[i] ? "ON" : "OFF");
	}
	sprintf(p2buffer, "%s", _95LimitStatus ? "ON" : "OFF");
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Limits Magnitude Error
 * Purpose:  This function defines Magnitude Error limits for results which
 *           characterize the modulation accuracy for the modulation scheme
 *           GMSK or EPSK.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalLimitsMagnitudeError(ViSession instrSession,
                                                             ViInt32   instance,
                                                             ViInt32   modulationScheme,
                                                             ViReal64  RMSLimit,
                                                             ViBoolean RMSLimitStatus[],
                                                             ViReal64  peakLimit,
                                                             ViBoolean peakLimitStatus[],
                                                             ViReal64  _95Limit,
                                                             ViBoolean _95LimitStatus)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *p2buffer = cmd;
	ViInt32  i;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, modulationScheme, 0, 2),
			3, "Modulation Scheme");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, RMSLimit, 0.0, 50.0), 4, "RMS Limit");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, peakLimit, 0.0, 50.0), 6, "Peak Limit");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, _95Limit, 0.0, 50.0), 8, "95% Limit");

	p2buffer += sprintf(p2buffer, "CONF:GSM:MEAS%ld:MEV:LIM:%s:MERR %.12le,%.12le,%.12le,",
	                    instance, modulationSchemeArr[modulationScheme], RMSLimit, peakLimit, _95Limit);
	for (i = 0; i < 3; i++)
	{
		p2buffer += sprintf(p2buffer, "%s,", RMSLimitStatus[i] ? "ON" : "OFF");
	}
	for (i = 0; i < 3; i++)
	{
		p2buffer += sprintf(p2buffer, "%s,", peakLimitStatus[i] ? "ON" : "OFF");
	}
	sprintf(p2buffer, "%s", _95LimitStatus ? "ON" : "OFF");
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Limits Phase Error
 * Purpose:  This function defines Phase Error limits for results which
 *           characterize the modulation accuracy for the modulation scheme
 *           GMSK or EPSK.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalLimitsPhaseError(ViSession instrSession,
                                                         ViInt32   instance,
                                                         ViInt32   modulationScheme,
                                                         ViReal64  RMSLimit,
                                                         ViBoolean RMSLimitStatus[],
                                                         ViReal64  peakLimit,
                                                         ViBoolean peakLimitStatus[],
                                                         ViReal64  _95Limit,
                                                         ViBoolean _95LimitStatus)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *p2buffer = cmd;
	ViInt32  i;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, modulationScheme, 0, 2),
			3, "Modulation Scheme");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, RMSLimit, 0.0, 180.0), 4, "RMS Limit");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, peakLimit, 0.0, 180.0), 6, "Peak Limit");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, _95Limit, 0.0, 180.0), 8, "95% Limit");

	p2buffer += sprintf(p2buffer, "CONF:GSM:MEAS%ld:MEV:LIM:%s:PERR %.12le,%.12le,%.12le,",
	                    instance, modulationSchemeArr[modulationScheme], RMSLimit, peakLimit, _95Limit);
	for (i = 0; i < 3; i++)
	{
		p2buffer += sprintf(p2buffer, "%s,", RMSLimitStatus[i] ? "ON" : "OFF");
	}
	for (i = 0; i < 3; i++)
	{
		p2buffer += sprintf(p2buffer, "%s,", peakLimitStatus[i] ? "ON" : "OFF");
	}
	sprintf(p2buffer, "%s", _95LimitStatus ? "ON" : "OFF");
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Limits IQ Offset
 * Purpose:  This function defines IQ Offset limits for results which
 *           characterize the modulation accuracy for the modulation scheme
 *           GMSK or EPSK.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalLimitsIQOffset(ViSession instrSession,
                                                       ViInt32   instance,
                                                       ViInt32   modulationScheme,
                                                       ViReal64  limit,
                                                       ViBoolean limitStatus[])
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *p2buffer = cmd;
	ViInt32  i;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, modulationScheme, 0, 2),
			3, "Modulation Scheme");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, limit, -100.0, 0.0), 4, "Limit");

	p2buffer += sprintf(p2buffer, "CONF:GSM:MEAS%ld:MEV:LIM:%s:IQOF %.12le,",
	                    instance, modulationSchemeArr[modulationScheme], limit);
	for (i = 0; i < 3; i++)
	{
		p2buffer += sprintf(p2buffer, "%s,", limitStatus[i] ? "ON" : "OFF");
	}
	*--p2buffer = '\n';
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Limits IQ Imbalance
 * Purpose:  This function defines IQ Imbalance limits for results which
 *           characterize the modulation accuracy for the modulation scheme
 *           GMSK or EPSK.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalLimitsIQImbalance(ViSession instrSession,
                                                          ViInt32   instance,
                                                          ViInt32   modulationScheme,
                                                          ViReal64  limit,
                                                          ViBoolean limitStatus[])
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *p2buffer = cmd;
	ViInt32  i;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, modulationScheme, 0, 2),
			3, "Modulation Scheme");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, limit, -100.0, 0.0), 4, "Limit");

	p2buffer += sprintf(p2buffer, "CONF:GSM:MEAS%ld:MEV:LIM:%s:IQIM %.12le,",
	                    instance, modulationSchemeArr[modulationScheme], limit);
	for (i = 0; i < 3; i++)
	{
		p2buffer += sprintf(p2buffer, "%s,", limitStatus[i] ? "ON" : "OFF");
	}
	*--p2buffer = '\n';
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Limits Timing Error
 * Purpose:  This function defines Timing Error limits for results which
 *           characterize the modulation accuracy for the modulation scheme
 *           GMSK or EPSK.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalLimitsTimingError(ViSession instrSession,
                                                          ViInt32   instance,
                                                          ViInt32   modulationScheme,
                                                          ViInt32   limit,
                                                          ViBoolean limitStatus[])
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *p2buffer = cmd;
	ViInt32  i;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, modulationScheme, 0, 2),
			3, "Modulation Scheme");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, limit, -1000, 1000), 4, "Limit");

	p2buffer += sprintf(p2buffer, "CONF:GSM:MEAS%ld:MEV:LIM:%s:TERR %ld,",
	                    instance, modulationSchemeArr[modulationScheme], limit);
	for (i = 0; i < 3; i++)
	{
		p2buffer += sprintf(p2buffer, "%s,", limitStatus[i] ? "ON" : "OFF");
	}
	*--p2buffer = '\n';
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Limits Frequency Error
 * Purpose:  This function defines Frequency Error limits for results which
 *           characterize the modulation accuracy for the modulation scheme
 *           GMSK or EPSK.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalLimitsFreqError(ViSession instrSession,
                                                        ViInt32   instance,
                                                        ViInt32   modulationScheme,
                                                        ViReal64  limit,
                                                        ViBoolean limitStatus[])
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *p2buffer = cmd;
	ViInt32  i;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, modulationScheme, 0, 2),
			3, "Modulation Scheme");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, limit, 0.0, 999.0), 4, "Limit");

	p2buffer += sprintf(p2buffer, "CONF:GSM:MEAS%ld:MEV:LIM:%s:FERR %.12le,",
	                    instance, modulationSchemeArr[modulationScheme], limit);
	for (i = 0; i < 3; i++)
	{
		p2buffer += sprintf(p2buffer, "%s,", limitStatus[i] ? "ON" : "OFF");
	}
	*--p2buffer = '\n';
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Limits PVT Burst Power
 * Purpose:  This function defines limits for the average burst power, i.e.
 *           tolerances for ranges of template power control levels (TPCLs).
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalLimitsPVTBurstPower(ViSession instrSession,
                                                            ViInt32   instance,
                                                            ViInt32   groupNumber,
                                                            ViInt32   startPCL,
                                                            ViInt32   stopPCL,
                                                            ViReal64  lowerLimit,
                                                            ViReal64  upperLimit,
                                                            ViBoolean status)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, groupNumber, 1, 10),
			3, "Group Number");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, startPCL, 0, 31),
			4, "Start PCL");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, stopPCL, 0, 31),
			5, "Stop PCL");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, lowerLimit, -10.0, 0.0), 6, "Lower Limit");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, upperLimit, 0.0, 10.0), 7, "Lower Limit");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CONF:GSM:MEAS%ld:MEV:LIM:PVT:ABP%ld %ld,%ld,%.12lf,%.12lf,%s", instance, groupNumber, startPCL, stopPCL, lowerLimit, upperLimit, status ? "ON" : "OFF");
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Limits PVT Guard Period
 * Purpose:  This function defines the raising of the upper limit line in the
 *           guard period between two consecutive bursts.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalLimitsPVTGuardPeriod(ViSession instrSession,
                                                             ViInt32   instance,
                                                             ViReal64  level)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, level, 0.0, 10.0), 3, "Level");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CONF:GSM:MEAS%ld:MEV:LIM:PVT:GPL %.12lf", instance, level);
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Limits PVT Upper Lines
 * Purpose:  This function defines and activate upper limit lines for the measured power
 *           vs. time. The lines apply to the modulation schemes GMSK or 8PSK
 *           (EPSK). Each line consists of three sections: rising edge (REDGe),
 *           useful part (UPARt) and falling edge (FEDGe). Each section consists
 *           of several areas for which relative and absolute limits can be defined
 *           (if both are defined the higher limit overrules the lower one).
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalLimitsPVTUpperLines(ViSession instrSession,
                                                            ViInt32   instance,
                                                            ViInt32   modulationScheme,
                                                            ViInt32   section,
                                                            ViInt32   area,
                                                            ViReal64  time[],
                                                            ViReal64  relativeLimitLevel[],
                                                            ViReal64  absoluteLimitLevel[],
                                                            ViBoolean absoluteLimitLevelStatus[],
                                                            ViBoolean statusArea)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *p2buffer = cmd;
	ViInt32  i;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, modulationScheme, 0, 2),
			3, "Modulation Scheme");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, section, 0, 2),
			4, "Section");

	switch (section)
	{
	case RSCMWGM_VAL_LIM_SECTION_UPAR:
		viCheckParm(RsCore_InvalidViInt32Range(instrSession, area, 1, 3),
				5, "Area");
		break;
	case RSCMWGM_VAL_LIM_SECTION_REDG:
	case RSCMWGM_VAL_LIM_SECTION_FEDG:
		viCheckParm(RsCore_InvalidViInt32Range(instrSession, area, 1, 4),
				5, "Area");
		break;
	default:
		viCheckParm(RsCore_InvalidViInt32Value(instrSession, section), 4, "Section");
	}
	p2buffer += sprintf(p2buffer, "CONF:GSM:MEAS%ld:MEV:LIM:%s:PVT:UPP:%s%ld:STAT ",
	                    instance, modulationSchemeArr[modulationScheme], PVTSectionArr[section], area);
	for (i = 0; i < 2; i++)
	{
		p2buffer += sprintf(p2buffer, "%le,", time[i] * 1E-6);
	}
	for (i = 0; i < 2; i++)
	{
		p2buffer += sprintf(p2buffer, "%.12lG,", relativeLimitLevel[i]);
	}
	for (i = 0; i < 2; i++)
	{
		if (absoluteLimitLevelStatus[i] == VI_TRUE)
		{
			p2buffer += sprintf(p2buffer, "%.12lG,", absoluteLimitLevel[i]);
		}
		else
		{
			p2buffer += sprintf(p2buffer, "OFF,");
		}
	}
	sprintf(p2buffer, "%s", statusArea ? "ON" : "OFF");
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Limits PVT Lower Lines
 * Purpose:  This function define and activate lower limit lines for the
 *           measured power vs. time. The lines apply to the "useful part" of
 *           a burst for modulation schemes GMSK or 8PSK (EPSK). Each line may
 *           consist of several areas for which relative and absolute limits
 *           can be defined (if both are defined the lower limit overrules
 *           the higher one).
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalLimitsPVTLowerLines(ViSession instrSession,
                                                            ViInt32   instance,
                                                            ViInt32   modulationScheme,
                                                            ViInt32   area,
                                                            ViReal64  time[],
                                                            ViReal64  relativeLimitLevel[],
                                                            ViReal64  absoluteLimitLevel[],
                                                            ViBoolean absoluteLimitLevelStatus[],
                                                            ViBoolean statusArea)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *p2buffer = cmd;
	ViInt32  i;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, modulationScheme, 0, 2),
			3, "Modulation Scheme");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, area, 1, 5),
			4, "Area");

	p2buffer += sprintf(p2buffer, "CONF:GSM:MEAS%ld:MEV:LIM:%s:PVT:LOW:UPAR%ld:STAT ",
	                    instance, modulationSchemeArr[modulationScheme], area);
	for (i = 0; i < 2; i++)
	{
		p2buffer += sprintf(p2buffer, "%.12lf,", time[i] * 1E-6);
	}
	for (i = 0; i < 2; i++)
	{
		p2buffer += sprintf(p2buffer, "%.12lf,", relativeLimitLevel[i]);
	}
	for (i = 0; i < 2; i++)
	{
		if (absoluteLimitLevelStatus[i] == VI_TRUE)
		{
			p2buffer += sprintf(p2buffer, "%.12lf,", absoluteLimitLevel[i]);
		}
		else
		{
			p2buffer += sprintf(p2buffer, "OFF,");
		}
	}
	sprintf(p2buffer, "%s", statusArea ? "ON" : "OFF");
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Limits PVT Upper Lines Dynamic
 * Purpose:  This function defines and activate dynamic (PCL-dependent)
 *           corrections to the upper limit lines for the measured
 *           power vs. time.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalLimitsPVTUpperLinesDynamic(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   modulationScheme,
							ViInt32   section,
							ViInt32   area,
							ViInt32   range,
							ViBoolean status,
							ViInt32   startPCL,
							ViInt32   stopPCL,
							ViReal64  correction)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, modulationScheme, 0, 2),
			3, "Modulation Scheme");

	switch (section)
	{
	case RSCMWGM_VAL_LIM_SECTION_UPAR:
		viCheckParm(RsCore_InvalidViInt32Range(instrSession, area, 1, 3),
				5, "Area");
		break;
	case RSCMWGM_VAL_LIM_SECTION_REDG:
	case RSCMWGM_VAL_LIM_SECTION_FEDG:
		viCheckParm(RsCore_InvalidViInt32Range(instrSession, area, 1, 4),
				5, "Area");
		break;
	default:
		viCheckParm(RsCore_InvalidViInt32Value(instrSession, section), 4, "Section");
	}
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, range, 1, 5),
			6, "Range");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, startPCL, 0, 31),
			8, "Start PCL");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, stopPCL, 0, 31),
			9, "Stop PCL");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, correction, -100.0, 100.0), 10, "Correction");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CONF:GSM:MEAS%ld:MEV:LIM:%s:PVT:UPP:%s%ld:DYN%ld %s,%ld,%ld,%.12lf", instance, modulationSchemeArr[modulationScheme], PVTSectionArr[section], area, range,
	         status ? "ON" : "OFF", startPCL, stopPCL, correction);
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Limits PVT Lower Lines Dynamic
 * Purpose:  This function defines and activates dynamic (PCL-dependent)
 *           corrections to the lower limit lines for the measured
 *           power vs. time.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalLimitsPVTLowerLinesDynamic(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   modulationScheme,
							ViInt32   area,
							ViInt32   range,
							ViBoolean status,
							ViInt32   startPCL,
							ViInt32   stopPCL,
							ViReal64  correction)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, modulationScheme, 0, 2),
			3, "Modulation Scheme");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, area, 1, 5),
			4, "Area");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, range, 1, 5),
			5, "Range");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, startPCL, 0, 31),
			7, "Start PCL");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, stopPCL, 0, 31),
			8, "Stop PCL");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, correction, -100.0, 100.0), 9, "Correction");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CONF:GSM:MEAS%ld:MEV:LIM:%s:PVT:LOW:UPAR%ld:DYN%ld %s,%ld,%ld,%.12lf", instance, modulationSchemeArr[modulationScheme], area, range,
	         status ? "ON" : "OFF", startPCL, stopPCL, correction);
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Limits Spec Modulation Ref Power
 * Purpose:  This function defines and activates two reference power values for
 *           the modulation scheme GMSK or EPSK.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalLimitsSpecModRefPower(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   modulationScheme,
							ViReal64  lowPower,
							ViReal64  highPower)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, modulationScheme, 0, 2),
			3, "Modulation Scheme");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, lowPower, 0.0, 43.0), 4, "Low Power");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, highPower, 0.0, 43.0), 5, "High Power");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CONF:GSM:MEAS%ld:MEV:LIM:%s:SMOD:RPOW %.12lG,%.12lG", instance, modulationSchemeArr[modulationScheme], lowPower, highPower);
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Limits Spec Modulation Freq Offset
 * Purpose:  This function defines and activates a limit line for the modulation
 *           scheme GMSK for a certain frequency offset. The specified limits apply
 *           above the high power reference value and below the low power reference
 *           value defined by the previous command. Between the two reference
 *           power values the limits are determined by linear interpolation.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalLimitsSpecModFreqOffset(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   modulationScheme,
							ViInt32   frequencyOffset,
							ViReal64  lowPowerLimit,
							ViReal64  highPowerLimit,
							ViReal64  absoluteLimit,
							ViBoolean limitStatus)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, modulationScheme, 0, 2),
			3, "Modulation Scheme");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, frequencyOffset, 1, 20),
			4, "Frequency Offset");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, lowPowerLimit, -120.0, 31.5), 5, "Low Power Limit");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, highPowerLimit, -120.0, 31.5), 6, "High Power Limit");
	viCheckParm(RsCore_InvalidViReal64Range(instrSession, absoluteLimit, -120.0, 31.5), 7, "Absolute Limit");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CONF:GSM:MEAS%ld:MEV:LIM:%s:SMOD:MPO%ld %.12lG,%.12lG,%.12lG,%s", instance, modulationSchemeArr[modulationScheme], frequencyOffset,
	         lowPowerLimit, highPowerLimit, absoluteLimit, limitStatus ? "ON" : "OFF");
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Limits Spec Switching Ref Power
 * Purpose:  This function defines and activates reference power values for
 *           the modulation scheme GMSK or EPSK.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalLimitsSpecSwitchRefPower(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   modulationScheme,
							ViBoolean status[],
							ViReal64  powerLevel[])
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *p2buffer = cmd;
	ViInt32  i;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, modulationScheme, 0, 2),
			3, "Modulation Scheme");

	p2buffer += sprintf(p2buffer, "CONF:GSM:MEAS%ld:MEV:LIM:%s:SSW:PLEV ",
	                    instance, modulationSchemeArr[modulationScheme]);
	for (i = 0; i < 10; i++)
	{
		p2buffer += sprintf(p2buffer, "%s,", status[i] ? "ON" : "OFF");
	}
	for (i = 0; i < 10; i++)
	{
		p2buffer += sprintf(p2buffer, "%.12lf,", powerLevel[i]);
	}
	*--p2buffer = '\0';
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Configure Multi Eval Limits Spec Switching Freq Offset
 * Purpose:  This function defines and activates a limit line for the modulation
 *           scheme GMSK or EPSK for a certain frequency offset. The specified limits
 *           apply at the reference power values defined by the previous command.
 *           Between the reference power values the limits are determined by
 *           linear interpolation.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ConfigureMEvalLimitsSpecSwitchFreqOffset(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   modulationScheme,
							ViInt32   frequencyOffset,
							ViReal64  powerLimit[],
							ViBoolean limitStatus)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *p2buffer = cmd;
	ViInt32  i;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, modulationScheme, 0, 2),
			3, "Modulation Scheme");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, frequencyOffset, 1, 20),
			4, "Frequency Offset");

	p2buffer += sprintf(p2buffer, "CONF:GSM:MEAS%ld:MEV:LIM:%s:SSW:MPO%ld ",
	                    instance, modulationSchemeArr[modulationScheme], frequencyOffset);
	for (i = 0; i < 10; i++)
	{
		p2buffer += sprintf(p2buffer, "%.12lf,", powerLimit[i]);
	}
	sprintf(p2buffer, "%s", limitStatus ? "ON" : "OFF");
	checkErr(RsCore_Write(instrSession, cmd));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Multi Eval Measurement Init
 * Purpose:  Starts or restarts the measurement; the R&S CMW 500 enters the
 *           "RUN" state.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_MEvalMeasInit(ViSession instrSession,
                                        ViInt32   instance)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViString(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_INIT, NULL),
			3, "Multi Eval Measurement Init");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Multi Eval Measurement Abort
 * Purpose:  Causes a running measurement to stop immediately; the R&S CMW 500
 *           enters the "OFF" state.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_MEvalMeasAbort(ViSession instrSession,
                                         ViInt32   instance)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViString(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_ABORT, NULL),
			3, "Multi Eval Measurement Abort");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Multi Eval Measurement Stop
 * Purpose:  Causes a running measurement to stop after the current evaluation
 *           period is terminated and valid results are available; the R&S CMW
 *           500 enters the "RDY" state.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_MEvalMeasStop(ViSession instrSession,
                                        ViInt32   instance)
{
	ViStatus error = VI_SUCCESS;
	ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_instanceRepCap(instrSession, instance, repCap));

	viCheckParm(rscmwgm_SetAttributeViString(instrSession, repCap, RSCMWGM_ATTR_MULTI_EVAL_MEASUREMENT_STOP, NULL),
			3, "Multi Eval Measurement Stop");

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Query Multi Eval Measurement Status
 * Purpose:  Queries the main measurement state and the measurement substates.
 *           Both measurement substates are relevant for running measurements
 *           only. Use FETCh...MEASurement<instance>:STATe? to query the main
 *           measurement state only. Use INITiate..., STOP..., ABORt... to
 *           change the measurement state.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasurementStatus(ViSession instrSession,
                                                      ViInt32   instance,
                                                      ViInt32   *mainState,
                                                      ViInt32   *syncState,
                                                      ViInt32   *resourcesState)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:STAT:ALL?", instance);
	checkErr(rscmwgm_QueryMeasStatus(instrSession, cmd, mainState, syncState,
        resourcesState));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Read Multi Eval EVM Trace
 * Purpose:  Starts a new single shot measurement and returns the EVM trace
 *           results of the multi evaluation measurement for the "measured
 *           slot".
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ReadMEvalMeasEVMTrace(ViSession instrSession,
                                                ViInt32   instance,
                                                ViInt32   resultType,
                                                ViInt32   arraySize,
                                                ViInt32   *reliability,
                                                ViReal64  results[],
                                                ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 2),
			3, "Result Type");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 6, "Results");
	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "READ:GSM:MEAS%ld:MEV:TRAC:EVM:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 15000, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval EVM Trace
 * Purpose:  Returns the EVM trace results of the multi evaluation measurement
 *           for the "measured slot".
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasEVMTrace(ViSession instrSession,
                                                 ViInt32   instance,
                                                 ViInt32   resultType,
                                                 ViInt32   arraySize,
                                                 ViInt32   *reliability,
                                                 ViReal64  results[],
                                                 ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 2),
			3, "Result Type");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 6, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:TRACe:EVM:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Read Multi Eval Magnitude Error Trace
 * Purpose:  Starts a new single shot measurement and returns the magnitude
 *           error trace results of the multi evaluation measurement for the
 *           "measured slot".
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ReadMEvalMeasMagnErrTrace(ViSession instrSession,
                                                    ViInt32   instance,
                                                    ViInt32   resultType,
                                                    ViInt32   arraySize,
                                                    ViInt32   *reliability,
                                                    ViReal64  results[],
                                                    ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 2),
			3, "Result Type");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 6, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "READ:GSM:MEAS%ld:MEV:TRAC:MERR:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 15000, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval Magnitude Error Trace
 * Purpose:  Returns the magnitude error trace results of the multi evaluation
 *           measurement for the "measured slot".
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasMagnErrTrace(ViSession instrSession,
                                                     ViInt32   instance,
                                                     ViInt32   resultType,
                                                     ViInt32   arraySize,
                                                     ViInt32   *reliability,
                                                     ViReal64  results[],
                                                     ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 2),
			3, "Result Type");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 6, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:TRACe:MERR:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Read Multi Eval Magnitude Error Trace
 * Purpose:  Starts a new single shot measurement and returns the phase
 *           error trace results of the multi evaluation measurement for the
 *           "measured slot".
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ReadMEvalMeasPhaseErrTrace(ViSession instrSession,
                                                     ViInt32   instance,
                                                     ViInt32   resultType,
                                                     ViInt32   arraySize,
                                                     ViInt32   *reliability,
                                                     ViReal64  results[],
                                                     ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 2),
			3, "Result Type");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 6, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "READ:GSM:MEAS%ld:MEV:TRACe:PERR:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 15000, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval Phase Error Trace
 * Purpose:  Returns the phase error trace results of the multi evaluation
 *           measurement for the "measured slot".
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasPhaseErrTrace(ViSession instrSession,
                                                      ViInt32   instance,
                                                      ViInt32   resultType,
                                                      ViInt32   arraySize,
                                                      ViInt32   *reliability,
                                                      ViReal64  results[],
                                                      ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 2),
			3, "Result Type");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 6, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:TRACe:PERR:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Read Multi Eval Modulation Results
 * Purpose:  Starts a new single shot measurement and returns the single slot
 *           modulation results of the multi evaluation measurement.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ReadMEvalMeasModulationResults(ViSession instrSession,
                                                         ViInt32   instance,
                                                         ViInt32   resultType,
                                                         ViInt32   arraySize,
                                                         ViInt32   *reliability,
                                                         ViReal64  results[],
                                                         ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 2),
			3, "Result Type");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "READ:GSM:MEAS%ld:MEV:MOD:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 15000, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval Modulation Results
 * Purpose:  Returns the single slot modulation results of the multi evaluation
 *           measurement.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasModulationResults(ViSession instrSession,
                                                          ViInt32   instance,
                                                          ViInt32   resultType,
                                                          ViInt32   arraySize,
                                                          ViInt32   *reliability,
                                                          ViReal64  results[],
                                                          ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 2),
			3, "Result Type");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:MOD:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Query Calculate Multi Eval Modulation Check Limit Results
 * Purpose:  Returns limit check results instead, one value for each result
 *           (Fetch or Read).
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasModulationCheckLimitResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   resultType,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 2),
			3, "Result Type");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results),
    	6, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:MOD:%s?", instance, resultTypeArr[resultType]);

	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, arraySize, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Read Multi Eval Modulation Standard Deviation
 * Purpose:  Returns the standard deviation of the single slot modulation
 *           results of the multi evaluation measurement.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ReadMEvalMeasModulationStandardDeviation(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 4, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "READ:GSM:MEAS%ld:MEV:MOD:SDEV?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 15000, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval Modulation Standard Deviation
 * Purpose:  Returns the standard deviation of the single slot modulation
 *           results of the multi evaluation measurement.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasModulationStandardDeviation(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 4, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:MOD:SDEV?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Query Multi Eval Modulation Standard Deviation Limit Results
 * Purpose:  Returns limit check results instead, one value for each result
 *           (Fetch or Read).
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasModulationStandardDeviationCheckLimitResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results),
    	5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:MOD:SDEV?", instance);

	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, arraySize, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Read Multi Eval Modulation Percentile
 * Purpose:  Starts a new single shot measurement and returns the 95th
 *           percentile results of the multi evaluation measurement.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ReadMEvalMeasModulationPercentile(ViSession instrSession,
                                                            ViInt32   instance,
                                                            ViInt32   arraySize,
                                                            ViInt32   *reliability,
                                                            ViReal64  results[],
                                                            ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 4, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "READ:GSM:MEAS%ld:MEV:MOD:PERC?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 15000, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval Modulation Percentile
 * Purpose:  Returns the 95th percentile results of the multi evaluation
 *           measurement.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasModulationPercentile(ViSession instrSession,
                                                             ViInt32   instance,
                                                             ViInt32   arraySize,
                                                             ViInt32   *reliability,
                                                             ViReal64  results[],
                                                             ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 4, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:MOD:PERC?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Query Multi Eval Modulation Percentile Limit Results
 * Purpose:  Returns limit check results instead, one value for each result
 *           (Fetch or Read).
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasModulationPercentileCheckLimitResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results),
    	5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:MOD:PERC?", instance);

	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, arraySize, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns the demodulated bits of the "Measurement Slot".
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR arraySize/Sets the number of values allocated for result array.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of symbols (demodulated bits) in decimal presentaion.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMultiEvalListModeAllDemodulatedBits(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:MOD:DBIT?", instance);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));

	p2buf = strtok(ext_buf, ",\r\n");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;

	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",\r\n");
	while (p2buf)
	{
		if (i < arraySize)
		{
			results[i++] = RsCore_Convert_String_To_ViInt32(p2buf);
		}
		p2buf = strtok(NULL, ",\r\n");
	}

	*returnedValues = i;

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Read Multi Eval Power vs Time Trace
 * Purpose:  Starts a new single shot measurement and returns the
 *           power vs. time trace results of the multi evaluation measurement
 *           for the "measured slot"
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ReadMEvalMeasPVTTrace(ViSession instrSession,
                                                ViInt32   instance,
                                                ViInt32   resultType,
                                                ViInt32   arraySize,
                                                ViInt32   *reliability,
                                                ViReal64  results[],
                                                ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 2),
			3, "Result Type");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 6, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "READ:GSM:MEAS%ld:MEV:TRAC:PVT:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 15000, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval Power vs Time Trace
 * Purpose:  Returns the power vs. time trace results of the multi evaluation
 *           measurement  for the "measured slot"
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasPVTTrace(ViSession instrSession,
                                                 ViInt32   instance,
                                                 ViInt32   resultType,
                                                 ViInt32   arraySize,
                                                 ViInt32   *reliability,
                                                 ViReal64  results[],
                                                 ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 2),
			3, "Result Type");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 6, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:TRAC:PVT:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Read Multi Eval Power vs Time Results
 * Purpose:  Starts a new single shot measurement and returns the statistical
 *           power vs. time results of the multi evaluation measurement.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ReadMEvalMeasPVTResults(ViSession instrSession,
                                                  ViInt32   instance,
                                                  ViInt32   arraySize,
                                                  ViInt32   *reliability,
                                                  ViReal64  results[],
                                                  ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 4, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "READ:GSM:MEAS%ld:MEV:PVT?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 15000, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval Power vs Time Results
 * Purpose:  Returns the statistical power vs. time results of the multi
 *           evaluation measurement.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasPVTResults(ViSession instrSession,
                                                   ViInt32   instance,
                                                   ViInt32   arraySize,
                                                   ViInt32   *reliability,
                                                   ViReal64  results[],
                                                   ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 4, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:PVT?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Query Multi Eval Power vs Time Limit Results
 * Purpose:  Returns the statistical power vs. time limit check results of
 *           the multi evaluation measurement.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasPVTCheckLimitResults(ViSession instrSession,
                                                             ViInt32   instance,
                                                             ViInt32   arraySize,
                                                             ViInt32   *reliability,
                                                             ViInt32   results[],
                                                             ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results),
    	5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:PVT:ALL?", instance);

	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, arraySize, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval Power vs Time Subvector
 * Purpose:  Returns special burst power values for the "Measure Slot".
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasPVTSubvector(ViSession instrSession,
                                                     ViInt32   instance,
                                                     ViInt32   resultType,
                                                     ViInt32   arraySize,
                                                     ViInt32   *reliability,
                                                     ViReal64  results[],
                                                     ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 3),
			3, "Result Type");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:PVT:%s:SVEC?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Query Multi Eval Power vs Time Subvector Limit Results
 * Purpose:  Returns special burst power values for the "Measure Slot" limits
 *           results.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasPVTSubvectorCheckLimitResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   resultType,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 3),
			3, "Result Type");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results),
    	6, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:PVT:%s:SVEC?", instance, resultTypeArr[resultType]);

	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, arraySize, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval Power vs Time Burst Type
 * Purpose:  Returns the detected burst type for all measured timeslots. 8
 *           values are returned, irrespective of the "No. of Slots"
 *           measured. If "No. of Slots" < 8, some of the returned values are
 *           NAN.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasPVTBurstType(ViSession instrSession,
                                                     ViInt32   instance,
                                                     ViInt32   arraySize,
                                                     ViInt32   *reliability,
                                                     ViInt32   results[],
                                                     ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViString burstTypeArr[] = {"OFF", "GMSK", "EPSK", "ACC", "Q16", NULL};
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:PVT:BTYP", instance);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",\n");
	while (p2buf != NULL)
	{
		if (i < arraySize)
		{
			results[i++] = RsCore_FindStringIndex(burstTypeArr, p2buf);
		}
		p2buf = strtok(NULL, ",\n");
	}
	if (returnedValues) *returnedValues = i;

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval Power vs Time Realtive Slot Timing
 * Purpose:  Returns the slot timing for all measured timeslots, relative to
 *           the timing of the "Measurement Slot". The relative slot timing
 *           of the "Measurement Slot" is always zero. The relative slot
 *           timing of the other timeslots is the deviation of the measured
 *           relative timing from the nominal timing, which is based on a
 *           timeslot length of 156.25 symbol durations.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasPVTRealtiveSlotTiming(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:PVT:RST", instance);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",\n");
	while (p2buf != NULL)
	{
		if (i < arraySize)
		{
			results[i++] = RsCore_Convert_String_To_ViInt32(p2buf);
		}
		p2buf = strtok(NULL, ",\n");
	}
	if (returnedValues) *returnedValues = i;

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval Power vs Time Training Sequence Code
 * Purpose:  Returns the detected Training Sequence Code (TSC) and burst type
 *           for all measured timeslots. 8 values are returned, irrespective
 *           of the "No. of Slots" measured.
 *****************************************************************************/

ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasPVTTrainingSequenceCode(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViString TSCArr[] = {"OFF", "NB0", "NB1", "NB2", "NB3", "NB4", "NB5", "NB6", "NB7", "DUMM", "AB0", "AB1", "AB2", NULL};
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:PVT:TSC", instance);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",\n");
	while (p2buf != NULL)
	{
		if (i < arraySize)
		{
			results[i++] = RsCore_FindStringIndex(TSCArr, p2buf);
		}
		p2buf = strtok(NULL, ",\n");
	}
	if (returnedValues) *returnedValues = i;

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Read Multi Eval Spectrum Modulation Burst Power Results
 * Purpose:  Starts a new single shot measurement and returns the average
 *           burst power measured at a series of frequencies. The frequencies
 *           are determined by the offset values defined via the command
 *           CONFigure:GSM:MEAS:MEValuation:SMODulation:FOFFsets. All
 *           defined offset values are considered (irrespective of their
 *           activation status).
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ReadMEvalMeasSpecModulationBurstPowerResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "READ:GSM:MEAS%ld:MEV:SMOD:FREQ?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 15000, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval Spectrum Modulation Burst Power Results
 * Purpose:  Returns the average burst power measured at a series of frequencies.
 *           The frequencies are determined by the offset values defined via
 *           the command CONFigure:GSM:MEAS:MEValuation:SMODulation:FOFFsets.
 *           All defined offset values are considered (irrespective of their
 *           activation status).
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasSpecModulationBurstPowerResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:SMOD:FREQ?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Query Multi Eval Spectrum Modulation Burst Power Limit Results
 * Purpose:  Returns the limits check average burst power measured at a series of
 *           frequencies. The frequencies are determined by the offset values
 *           defined via the command
 *           CONFigure:GSM:MEAS:MEValuation:SMODulation:FOFFsets.
 *           All defined offset values are considered (irrespective of their
 *           activation status).
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasSpecModulationBurstPowerCheckLimitResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results),
    	5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:SMOD:FREQ?", instance);

	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, arraySize, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Read Multi Eval Spectrum Modulation Frequency Limits Results
 * Purpose:  Starts a new single shot measurement and returns  the limits for
 *			 spectrum modulation frequency.
 *
 * 			 See also CONFigure:GSM:MEAS:MEValuation:SMODulation:OFRequence
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ReadMEvalMeasSpecModulationFrequencyLimitsResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "READ:GSM:MEAS%ld:MEV:SMOD:FREQ:LIM?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 15000, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval Spectrum Modulation Frequency Limits Results
 * Purpose:  Queries the limits for spectrum modulation frequency.
 *
 *			 See also CONFigure:GSM:MEAS:MEValuation:SMODulation:OFRequence
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasSpecModulationFrequencyLimitsResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:SMOD:FREQ:LIM?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Read Multi Eval Spectrum Modulation Trace
 * Purpose:  Starts a new single shot measurement and returns the spectrum
 *           due to modulation trace values measured at a selected offset
 *           frequency.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ReadMEvalMeasSpecModulationTrace(ViSession instrSession,
                                                           ViInt32   instance,
                                                           ViInt32   arraySize,
                                                           ViInt32   *reliability,
                                                           ViReal64  results[],
                                                           ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "READ:GSM:MEAS%ld:MEV:TRAC:SMOD:TIME:CURR?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 15000, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval Spectrum Modulation Trace
 * Purpose:  Returns the spectrum due to modulation trace values measured at
 *           a selected offset frequency.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasSpecModulationTrace(ViSession instrSession,
                                                            ViInt32   instance,
                                                            ViInt32   arraySize,
                                                            ViInt32   *reliability,
                                                            ViReal64  results[],
                                                            ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:TRAC:SMOD:TIME:CURR?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Read Multi Eval Spectrum Modulation Result
 * Purpose:  Starts a new single shot measurement and returns general spectrum
 *           modulation frequency results.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ReadMEvalMeasSpecModulationResult(ViSession instrSession,
                                                            ViInt32   instance,
                                                            ViInt32   arraySize,
                                                            ViInt32   *reliability,
                                                            ViReal64  results[],
                                                            ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 4, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "READ:GSM:MEAS%ld:MEV:SMOD?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 15000, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval Spectrum Modulation Result
 * Purpose:  Returns general spectrum modulation frequency results.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasSpecModulationResult(ViSession instrSession,
                                                             ViInt32   instance,
                                                             ViInt32   arraySize,
                                                             ViInt32   *reliability,
                                                             ViReal64  results[],
                                                             ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 4, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:SMOD?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Query Multi Eval Spectrum Modulation Limit Result
 * Purpose:  Returns general spectrum modulation frequency limit check
 *           results.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasSpecModulationCheckLimitResult(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results),
    	5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:SMOD?", instance);

	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, arraySize, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Read Multi Eval Spectrum Switching Burst Power Results
 * Purpose:  Starts a new single shot measurement and returns the maximum
 *           burst power measured at a series of frequencies. The frequencies
 *           are determined by the offset values defined via the command
 *           CONFigure:GSM:MEAS:MEValuation:SSWitching:FOFFsets. All
 *           defined offset values are considered (irrespective of their
 *           activation status).
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ReadMEvalMeasSpecSwitchBurstPowerResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "READ:GSM:MEAS%ld:MEV:SSW:FREQ?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 15000, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval Spectrum Switching Burst Power Results
 * Purpose:  Returns the maximum burst power measured at a series of frequencies.
 *           The frequencies are determined by the offset values defined via
 *           the command CONFigure:GSM:MEAS:MEValuation:SSWitching:FOFFsets.
 *           All defined offset values are considered (irrespective of their
 *           activation status).
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasSpecSwitchBurstPowerResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:SSW:FREQ?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Query Multi Eval Spectrum Switching Burst Power Limit Results
 * Purpose:  Returns the limits check of maximum burst power measured at a
 *           series of frequencies. The frequencies are determined by the
 *           offset values defined via the command
 *           CONFigure:GSM:MEASurement<instance>:MEValuation:SSWitching:FOFF
 *           sets. All defined offset values are considered (irrespective of
 *           their activation status).
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasSpecSwitchBurstPowerCheckLimitResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results),
    	5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:SSW:FREQ?", instance);

	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, arraySize, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Read Multi Eval Spectrum Switching Frequency Limits Results
 * Purpose:  Starts a new single shot measurement and returns the maximum
 *           burst power measured at a series of frequencies. The frequencies
 *           are determined by the offset values defined via the command
 *           CONFigure:GSM:MEAS:MEValuation:SSWitching:FOFFsets. All
 *           defined offset values are considered (irrespective of their
 *           activation status).
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ReadMEvalMeasSpecSwitchFrequencyLimitsResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "READ:GSM:MEAS%ld:MEV:SSW:FREQ:LIM?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 15000, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval Spectrum Switching Frequency Limits Results
 * Purpose:  Returns the maximum burst power measured at a series of frequencies.
 *           The frequencies are determined by the offset values defined via
 *           the command CONFigure:GSM:MEAS:MEValuation:SSWitching:FOFFsets.
 *           All defined offset values are considered (irrespective of their
 *           activation status).
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasSpecSwitchFrequencyLimitsResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:SSW:FREQ:LIM?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Read Multi Eval Spectrum Switching Trace
 * Purpose:  Starts a new single shot measurement and returns the spectrum
 *           due to switching trace values measured at a selected offset
 *           frequency.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ReadMEvalMeasSpecSwitchTrace(ViSession instrSession,
                                                       ViInt32   instance,
                                                       ViInt32   arraySize,
                                                       ViInt32   *reliability,
                                                       ViReal64  results[],
                                                       ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "READ:GSM:MEAS%ld:MEV:TRAC:SSW:TIME:CURR?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 15000, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval Spectrum Switching Trace
 * Purpose:  Returns the spectrum due to switching trace values measured at
 *           a selected offset frequency.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasSpecSwitchTrace(ViSession instrSession,
                                                        ViInt32   instance,
                                                        ViInt32   arraySize,
                                                        ViInt32   *reliability,
                                                        ViReal64  results[],
                                                        ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:TRAC:SSW:TIME:CURR?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Read Multi Eval Spectrum Switching Result
 * Purpose:  Starts a new single shot measurement and returns general spectrum
 *           switching frequency results.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ReadMEvalMeasSpecSwitchResult(ViSession instrSession,
                                                        ViInt32   instance,
                                                        ViInt32   arraySize,
                                                        ViInt32   *reliability,
                                                        ViReal64  results[],
                                                        ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 4, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "READ:GSM:MEAS%ld:MEV:SSW?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 15000, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval Spectrum Switching Result
 * Purpose:  Returns general spectrum switching frequency results.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasSpecSwitchResult(ViSession instrSession,
                                                         ViInt32   instance,
                                                         ViInt32   arraySize,
                                                         ViInt32   *reliability,
                                                         ViReal64  results[],
                                                         ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results), 4, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:SSW?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, arraySize, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Query Multi Eval Spectrum Switching Limit Result
 * Purpose:  Returns general limits check of the spectrum switching frequency
 *           results.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasSpecSwitchCheckLimitResult(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, results),
    	5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:SSW?", instance);

	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, arraySize, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval IQ Constellation Result
 * Purpose:  Returns the results in the I/Q constellation diagram.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasIQConstellationResult(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViReal64  i[],
							ViReal64  q[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViReal64 *loc_res = NULL;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, i), 5, "I");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, q), 6, "Q");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:TRAC:IQ?", instance);
	loc_res = (ViReal64 *)malloc(2 * arraySize * sizeof(ViReal64));
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, arraySize, reliability, loc_res, returnedValues));

	memcpy(i, loc_res, sizeof(ViReal64) * arraySize);
	memcpy(q, &loc_res[arraySize], sizeof(ViReal64) * arraySize);

Error:
	if (loc_res != NULL)
		free(loc_res);

	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Read Multi Eval IQ Constellation Result
 * Purpose: Starts a new single shot measurement and returns the results in
 *          the I/Q constellation diagram.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ReadMEvalMeasIQConstellationResult(ViSession instrSession,
                                                             ViInt32   instance,
                                                             ViInt32   arraySize,
                                                             ViInt32   *reliability,
                                                             ViReal64  i[],
                                                             ViReal64  q[],
                                                             ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViReal64 *loc_res = NULL;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, i), 5, "I");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, q), 6, "Q");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "READ:GSM:MEAS%ld:MEV:TRAC:IQ?", instance);
	loc_res = (ViReal64 *)malloc(2 * arraySize * sizeof(ViReal64));
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 15000, arraySize, reliability, loc_res, returnedValues));

	memcpy(i, loc_res, sizeof(ViReal64) * arraySize);
	memcpy(q, &loc_res[arraySize], sizeof(ViReal64) * arraySize);

Error:
	if (loc_res != NULL)
		free(loc_res);

	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval BER Result
 * Purpose:  Returns measured Bit Error Rate.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasBERResult(ViSession instrSession,
                                                  ViInt32   instance,
                                                  ViInt32   *reliability,
                                                  ViInt32   *errors,
                                                  ViInt32   *bursts)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViReal64 results[2];
	ViInt32  returned = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, bursts), 5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:BER?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, 2, reliability, results, &returned));
	*errors = (ViInt32)results[0];
	*bursts = (ViInt32)results[1];

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Read Multi Eval BER Result
 * Purpose:  Starts a new single shot measurement and returns measured Bit
 *           Error Rate.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ReadMEvalMeasBERResult(ViSession instrSession,
                                                 ViInt32   instance,
                                                 ViInt32   *reliability,
                                                 ViInt32   *errors,
                                                 ViInt32   *bursts)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViReal64 results[2];
	ViInt32  returned = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, bursts), 5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "READ:GSM:MEAS%ld:MEV:BER?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 15000, 2, reliability, results, &returned));
	*errors = (ViInt32)results[0];
	*bursts = (ViInt32)results[1];

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Query Multi Eval BER Limit Result
 * Purpose:  Starts a new single shot measurement and returns measured Bit
 *           Error Rate and returns limit results.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasBERCheckLimitResult(ViSession instrSession,
                                                            ViInt32   instance,
                                                            ViInt32   *reliability,
                                                            ViInt32   *errors,
                                                            ViInt32   *bursts)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViReal64 results[2];
	ViInt32  returned = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, bursts), 5, "Results");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:BER?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 15000, 2, reliability, results, &returned));
	*errors = (ViInt32)results[0];
	*bursts = (ViInt32)results[1];

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval List Mode Modulation Results
 * Purpose:  Returns the modulation results for selected segment in list
 *           mode.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeModulationResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   resultType,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   *returnCode,
							ViReal64  *statisticsExpired,
							ViInt32   *burstType,
							ViBoolean *slotStatistics,
							ViReal64  *out_ofTolerance,
							ViReal64  *results,
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 200),
			3, "Segment");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 2),
			4, "Result Type");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:SEGM%ld:MOD:%s", instance, segment, resultTypeArr[resultType]);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",");
	*returnCode = RsCore_Convert_String_To_ViInt32(p2buf);
	p2buf = strtok(NULL, ",");
	*statisticsExpired = RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");
	*burstType = RsCore_FindStringIndex(listModeBurstTypeArr, p2buf);
	p2buf = strtok(NULL, ",");
	*slotStatistics = RsCore_Convert_String_To_Boolean(p2buf);
	p2buf = strtok(NULL, ",");
	*out_ofTolerance = RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");

	while (p2buf != NULL)
	{
		if (i < arraySize)
		{
			results[i++] = RsCore_Convert_String_To_ViReal64(p2buf);
		}

		p2buf = strtok(NULL, ",\n");
	}
	if (returnedValues) *returnedValues = i;

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}


/// HIFN Returns the standard deviation limit state of the modulation results for selected segment in list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR segment/Selects the segment.
/// HIPAR arraySize/Sets the number of values allocated for result array.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement result
/// HIPAR returnCode/Returns the code limit state.
/// HIPAR statisticsExpired/Returns relative number of slots limit state.
/// HIPAR burstType/Returns the limit state for detected burst type of the last measured burst.
/// HIPAR slotStatistics/Returns the slot statistics limit state.
/// HIPAR outOfTolerance/Returns the limit state for percentage of measured bursts with failed limit check.
/// HIPAR results/Returns the array of results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasListModeCheckLimitModulationStandardDeviation(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   *returnCode,
							ViInt32   *statisticsExpired,
							ViInt32   *burstType,
							ViInt32   *slotStatistics,
							ViInt32   *outOfTolerance,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViInt32  i;
	ViInt32  codes[20];

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 200),
			3, "Segment");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:SEGM%ld:MOD:SDEV", instance, segment);
	checkErr(rscmwgm_GetMeasErrorCodesQuery(instrSession, cmd,reliability,4+arraySize,codes,returnCode,returnedValues));

	*statisticsExpired = codes[0];
	*burstType = codes[1];
	*slotStatistics = codes[2];
	*outOfTolerance = codes[3];

	for (i = 0; i < arraySize; i++)
	{
		results[i] = codes[4 + i];
	}

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval List Mode Modulation Percentile
 * Purpose:  Returns the modulation results for selected segment in list
 *           mode.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeModulationPercentile(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   *returnCode,
							ViReal64  *statisticsExpired,
							ViInt32   *burstType,
							ViBoolean *slotStatistics,
							ViReal64  *out_ofTolerance,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 200),
			3, "Segment");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:SEGM%ld:MOD:PERC", instance, segment);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",");
	*returnCode = RsCore_Convert_String_To_ViInt32(p2buf);
	p2buf = strtok(NULL, ",");
	*statisticsExpired = RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");
	*burstType = RsCore_FindStringIndex(listModeBurstTypeArr, p2buf);
	p2buf = strtok(NULL, ",");
	*slotStatistics = RsCore_Convert_String_To_Boolean(p2buf);
	p2buf = strtok(NULL, ",");
	*out_ofTolerance = RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");

	while (p2buf != NULL)
	{
		if (i < arraySize)
		{
			results[i++] = RsCore_Convert_String_To_ViReal64(p2buf);
		}
		p2buf = strtok(NULL, ",\n");
	}
	if (returnedValues) *returnedValues = i;

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}


/// HIFN Returns the 95th percentile limit state of the modulation results for selected segment in list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR segment/Selects the segment.
/// HIPAR arraySize/Sets the number of values allocated for result array.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement result
/// HIPAR returnCode/Returns the code limit state.
/// HIPAR statisticsExpired/Returns relative number of slots limit state.
/// HIPAR burstType/Returns the limit state for detected burst type of the last measured burst.
/// HIPAR slotStatistics/Returns the slot statistics limit state.
/// HIPAR outOfTolerance/Returns the limit state for percentage of measured bursts with failed limit check.
/// HIPAR results/Returns the array of results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasListModeCheckLimitModulationPercentile(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   *returnCode,
							ViInt32   *statisticsExpired,
							ViInt32   *burstType,
							ViInt32   *slotStatistics,
							ViInt32   *outOfTolerance,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViInt32  i;
	ViInt32  codes[20];

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 200),
			3, "Segment");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:SEGM%ld:MOD:PERC", instance, segment);
	checkErr(rscmwgm_GetMeasErrorCodesQuery(instrSession, cmd,reliability,4+arraySize,codes,returnCode,returnedValues));

	*statisticsExpired = codes[0];
	*burstType = codes[1];
	*slotStatistics = codes[2];
	*outOfTolerance = codes[3];

	for (i = 0; i < arraySize; i++)
	{
		results[i] = codes[4 + i];
	}

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval List Mode Spectrum Modulation
 * Purpose:  Returns the spectrum due to modulation results for selected segment
 *           in list mode.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeSpectrumModulation(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   *returnCode,
							ViReal64  *statisticsExpired,
							ViInt32   *burstType,
							ViBoolean *slotStatistics,
							ViReal64  *out_ofTolerance,
							ViReal64  *carrierPower,
							ViReal64  powerOffset[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 200),
			3, "Segment");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:SEGM%ld:SMOD", instance, segment);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",");
	*returnCode = RsCore_Convert_String_To_ViInt32(p2buf);
	p2buf = strtok(NULL, ",");
	*statisticsExpired = RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");
	*burstType = RsCore_FindStringIndex(listModeBurstTypeArr, p2buf);
	p2buf = strtok(NULL, ",");
	*slotStatistics = RsCore_Convert_String_To_Boolean(p2buf);
	p2buf = strtok(NULL, ",");
	*out_ofTolerance = RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");
	*carrierPower = RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");

	while (p2buf != NULL)
	{
		if (i < arraySize)
		{
			powerOffset[i++] = RsCore_Convert_String_To_ViReal64(p2buf);
		}
		p2buf = strtok(NULL, ",\n");
	}
	if (returnedValues) *returnedValues = i;

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}


/// HIFN Returns the limit state for spectrum due to modulation results for selected segment in list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR segment/Selects the segment.
/// HIPAR arraySize/Sets the number of values allocated for result array.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement result
/// HIPAR returnCode/Returns the code limit state.
/// HIPAR statisticsExpired/Returns relative number of slots limit state.
/// HIPAR burstType/Returns the limit state for detected burst type of the last measured burst.
/// HIPAR slotStatistics/Returns the slot statistics limit state.
/// HIPAR outOfTolerance/Returns the limit state for percentage of measured bursts with failed limit check.
/// HIPAR carrierPower/Returns the limit state of measured carrier output power.
/// HIPAR powerOffset/Returns the array of average burst power at the carrier frequency +/- the frequency offset.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasListModeCheckLimitSpectrumModulation(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   *returnCode,
							ViInt32   *statisticsExpired,
							ViInt32   *burstType,
							ViInt32   *slotStatistics,
							ViInt32   *outOfTolerance,
							ViInt32   *carrierPower,
							ViInt32   powerOffset[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViInt32  i;
	ViInt32  codes[50];

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 200),
			3, "Segment");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:SEGM%ld:SMOD", instance, segment);
	checkErr(rscmwgm_GetMeasErrorCodesQuery(instrSession, cmd,reliability,5+arraySize,codes,returnCode,returnedValues));

	*statisticsExpired = codes[0];
	*burstType = codes[1];
	*slotStatistics = codes[2];
	*outOfTolerance = codes[3];
	*carrierPower = codes[4];
	for (i = 0; i < arraySize; i++)
	{
		powerOffset[i] = codes[5 + i];
	}

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval List Mode Spectrum Switching
 * Purpose:  Returns the spectrum due to switching results for selected segment
 *           in list mode.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeSpectrumSwitching(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   *returnCode,
							ViReal64  *statisticsExpired,
							ViInt32   *burstType,
							ViBoolean *slotStatistics,
							ViReal64  *out_ofTolerance,
							ViReal64  *carrierPower,
							ViReal64  powerOffset[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 200),
			3, "Segment");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:SEGM%ld:SSW", instance, segment);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",");
	*returnCode = RsCore_Convert_String_To_ViInt32(p2buf);
	p2buf = strtok(NULL, ",");
	*statisticsExpired = RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");
	*burstType = RsCore_FindStringIndex(listModeBurstTypeArr, p2buf);
	p2buf = strtok(NULL, ",");
	*slotStatistics = RsCore_Convert_String_To_Boolean(p2buf);
	p2buf = strtok(NULL, ",");
	*out_ofTolerance = RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");
	*carrierPower = RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");

	while (p2buf != NULL)
	{
		if (i < arraySize)
		{
			powerOffset[i++] = RsCore_Convert_String_To_ViReal64(p2buf);
		}
		p2buf = strtok(NULL, ",\n");
	}
	if (returnedValues) *returnedValues = i;

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/// HIFN Returns the limit state for spectrum due to switching results for selected segment in list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR segment/Selects the segment.
/// HIPAR arraySize/Sets the number of values allocated for result array.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement result
/// HIPAR returnCode/Returns the code limit state.
/// HIPAR statisticsExpired/Returns relative number of slots limit state.
/// HIPAR burstType/Returns the limit state for detected burst type of the last measured burst.
/// HIPAR slotStatistics/Returns the slot statistics limit state.
/// HIPAR outOfTolerance/Returns the limit state for percentage of measured bursts with failed limit check.
/// HIPAR carrierPower/Returns the limit state of measured carrier output power.
/// HIPAR powerOffset/Returns the array of average burst power limit state at the carrier frequency +/- the frequency offset.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasListModeCheckLimitSpectrumSwitching(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   *returnCode,
							ViInt32   *statisticsExpired,
							ViInt32   *burstType,
							ViInt32   *slotStatistics,
							ViInt32   *outOfTolerance,
							ViInt32   *carrierPower,
							ViInt32   powerOffset[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViInt32  i;
	ViInt32  codes[50];

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 200),
			3, "Segment");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:SEGM%ld:SSW", instance, segment);
	checkErr(rscmwgm_GetMeasErrorCodesQuery(instrSession, cmd,reliability,5+arraySize,codes,returnCode,returnedValues));

	*statisticsExpired = codes[0];
	*burstType = codes[1];
	*slotStatistics = codes[2];
	*outOfTolerance = codes[3];
	*carrierPower = codes[4];
	for (i = 0; i < arraySize; i++)
	{
		powerOffset[i] = codes[5 + i];
	}

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval List Mode Spectrum BER
 * Purpose:  Returns the BER results for segment no. <Seg> in list mode
 *           (CONFigure:GSM:MEAS:MEValuation:LIST ON). To define the
 *           statistical length and enable the calculation of the BER
 *           results, use the command
 *           CONFigure:GSM:MEAS:MEValuation:LIST:SEGMent<Seg>:BER
 *           <Statistics>, <BER Enable>.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeSpectrumBER(ViSession instrSession,
                                                            ViInt32   instance,
                                                            ViInt32   segment,
                                                            ViInt32   *reliability,
                                                            ViInt32   *returnCode,
                                                            ViReal64  *statisticsExpired,
                                                            ViInt32   *burstType,
                                                            ViBoolean *slotStatistics,
                                                            ViInt32   *ber,
                                                            ViInt32   *errors,
                                                            ViInt32   *bursts)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 200),
			3, "Segment");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:SEGM%ld:BER", instance, segment);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",");
	*returnCode = RsCore_Convert_String_To_ViInt32(p2buf);
	p2buf = strtok(NULL, ",");
	*statisticsExpired = RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");
	*burstType = RsCore_FindStringIndex(listModeBurstTypeArr, p2buf);
	p2buf = strtok(NULL, ",");
	*slotStatistics = RsCore_Convert_String_To_Boolean(p2buf);
	p2buf = strtok(NULL, ",");
	*ber = (ViInt32)RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");
	*errors = (ViInt32)RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");
	*bursts = (ViInt32)RsCore_Convert_String_To_ViReal64(p2buf);

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}


/// HIFN Returns the BER limit check results for segment no. <Seg> in list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR segment/Selects the segment.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement result
/// HIPAR returnCode/Returns the code limit state.
/// HIPAR statisticsExpired/Returns relative number of slots limit state.
/// HIPAR burstType/Returns the limit state for detected burst type of the last measured burst.
/// HIPAR slotStatistics/Returns the slot statistics limit state.
/// HIPAR ber/Returns percentage of Bit Error Rate limit state.
/// HIPAR errors/Returns lmit state for total number of detected bit errors.
/// HIPAR bursts/Returns limit state for total number of measured bursts.
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasListModeCheckLimitSpectrumBER(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   *reliability,
							ViInt32   *returnCode,
							ViInt32   *statisticsExpired,
							ViInt32   *burstType,
							ViInt32   *slotStatistics,
							ViInt32   *ber,
							ViInt32   *errors,
							ViInt32   *bursts)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViInt32  codes[20];
	ViInt32  *returnedValues = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 200),
			3, "Segment");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:SEGM%ld:BER", instance, segment);
	checkErr(rscmwgm_GetMeasErrorCodesQuery(instrSession, cmd,reliability,6,codes,returnCode, returnedValues));

	*statisticsExpired = codes[0];
	*burstType = codes[1];
	*slotStatistics = codes[2];
	*ber = codes[3];
	*errors = codes[4];
	*bursts = codes[5];

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval List Mode Power vs Time Results
 * Purpose:  Returns the power vs. time results for selected segment in list
 *           mode.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModePVTResults(ViSession instrSession,
                                                           ViInt32   instance,
                                                           ViInt32   segment,
                                                           ViInt32   resultType,
                                                           ViInt32   *reliability,
                                                           ViInt32   *returnCode,
                                                           ViReal64  *statisticsExpired,
                                                           ViInt32   *burstType,
                                                           ViBoolean *slotStatistics,
                                                           ViReal64  *out_ofTolerance,
                                                           ViReal64  *averageBurstPower)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 200),
			3, "Segment");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 1),
			4, "Result Type");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:SEGM%ld:PVT:%s", instance, segment, resultTypeArr[resultType]);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",");
	*returnCode = RsCore_Convert_String_To_ViInt32(p2buf);
	p2buf = strtok(NULL, ",");
	*statisticsExpired = RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");
	*burstType = RsCore_FindStringIndex(listModeBurstTypeArr, p2buf);
	p2buf = strtok(NULL, ",");
	*slotStatistics = RsCore_Convert_String_To_Boolean(p2buf);
	p2buf = strtok(NULL, ",");
	*out_ofTolerance = RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");
	*averageBurstPower = RsCore_Convert_String_To_ViReal64(p2buf);

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/// HIFN Returns the power vs. time results limit state for selected segment in list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR segment/Selects the segment.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement result
/// HIPAR returnCode/Returns the code limit state.
/// HIPAR statisticsExpired/Returns relative number of slots limit state.
/// HIPAR burstType/Returns the limit state for detected burst type of the last measured burst.
/// HIPAR slotStatistics/Returns the slot statistics limit state.
/// HIPAR outOfTolerance/Returns the limit state for percentage of measured bursts with failed limit check.
/// HIPAR averageBurstPower/Returns the average burst power limit state.
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasListModeCheckLimitPVTResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   resultType,
							ViInt32   *reliability,
							ViInt32   *returnCode,
							ViInt32   *statisticsExpired,
							ViInt32   *burstType,
							ViInt32   *slotStatistics,
							ViInt32   *outOfTolerance,
							ViInt32   *averageBurstPower)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViInt32  codes[20];
	ViInt32  *returnedValues = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 200),
			3, "Segment");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 1),
			4, "Result Type");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:SEGM%ld:PVT:%s", instance, segment, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasErrorCodesQuery(instrSession, cmd,reliability,5,codes,returnCode,returnedValues));

	*statisticsExpired = codes[0];
	*burstType = codes[1];
	*slotStatistics = codes[2];
	*outOfTolerance = codes[3];
	*averageBurstPower = codes[4];

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval List Mode Power vs Time Subvector
 * Purpose:  Returns special burst power values for the "Measure Slot" for
 *           selected segment in list mode.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModePVTSubvector(ViSession instrSession,
                                                             ViInt32   instance,
                                                             ViInt32   segment,
                                                             ViInt32   resultType,
                                                             ViInt32   arraySize,
                                                             ViInt32   *reliability,
                                                             ViInt32   *returnCode,
                                                             ViReal64  *statisticsExpired,
                                                             ViInt32   *burstType,
                                                             ViBoolean *slotStatistics,
                                                             ViReal64  *outofTolerance,
                                                             ViReal64  *usefulPartMin,
                                                             ViReal64  *usefulPartMax,
                                                             ViReal64  subvector[],
                                                             ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 200),
			3, "Segment");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 3),
			4, "Result Type");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEValuation:LIST:SEGMent%ld:PVT:%s:SVEC", instance, segment, resultTypeArr[resultType]);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",");
	*returnCode = RsCore_Convert_String_To_ViInt32(p2buf);
	p2buf = strtok(NULL, ",");
	*statisticsExpired = RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");
	*burstType = RsCore_FindStringIndex(listModeBurstTypeArr, p2buf);
	p2buf = strtok(NULL, ",");
	*slotStatistics = RsCore_Convert_String_To_Boolean(p2buf);
	p2buf = strtok(NULL, ",");
	*outofTolerance = RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");
	*usefulPartMin = RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");
	*usefulPartMax = RsCore_Convert_String_To_ViReal64(p2buf);

	while (p2buf != NULL)
	{
		if (i < arraySize)
		{
			subvector[i++] = RsCore_Convert_String_To_ViReal64(p2buf);
		}
		p2buf = strtok(NULL, ",\n");
	}
	if (returnedValues) *returnedValues = i;

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

ViStatus _VI_FUNC rscmwgm_MEvalMeasListModeAll(ViSession instrSession,
                                               ViString  cmd,
                                               ViInt32   numberofActiveSegments,
                                               ViInt32   numberofResults,
                                               ViInt32   *reliability,
                                               ViInt32   returnCode[],
                                               ViReal64  statisticsExpired[],
                                               ViInt32   burstType[],
                                               ViBoolean slotStatistics[],
                                               ViReal64  outofTolerance[],
                                               ViReal64  results[],
                                               ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;
	ViInt32  j = 0;

	checkErr(RsCore_LockSession(instrSession));

	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",\n");

	while (p2buf != NULL && i < numberofActiveSegments)
	{
		p2buf = strtok(NULL, ",");
		returnCode[i] = RsCore_Convert_String_To_ViInt32(p2buf);
		p2buf = strtok(NULL, ",");
		statisticsExpired[i] = RsCore_Convert_String_To_ViReal64(p2buf);
		p2buf = strtok(NULL, ",");
		burstType[i] = RsCore_FindStringIndex(listModeBurstTypeArr, p2buf);
		p2buf = strtok(NULL, ",");
		slotStatistics[i] = RsCore_Convert_String_To_Boolean(p2buf);
		p2buf = strtok(NULL, ",");
		outofTolerance[i] = RsCore_Convert_String_To_ViReal64(p2buf);
		p2buf = strtok(NULL, ",");

		for (j = 0; j < numberofResults; j++)
		{
			*(results + numberofResults * i + j) = RsCore_Convert_String_To_ViReal64(p2buf);
			p2buf = strtok(NULL, ",\n");
		}
		i++;
	}
	if (returnedValues) *returnedValues = i;

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

ViStatus _VI_FUNC rscmwgm_MEvalMeasListModeAllQuery(ViSession instrSession,
                                                    ViString  cmd,
                                                    ViInt32   numberofActiveSegments,
                                                    ViInt32   numberofResults,
                                                    ViInt32   *reliability,
                                                    ViInt32   returnCode[],
                                                    ViInt32   statisticsExpired[],
                                                    ViInt32   burstType[],
                                                    ViInt32   slotStatistics[],
                                                    ViInt32   outofTolerance[],
                                                    ViInt32   results[],
                                                    ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;
	ViInt32  j = 0;
	static   ViString list[] = {"OK", "ULEU", "ULEL", "OFL", "UFL", "INV", "NAV",NULL};

	checkErr(RsCore_LockSession(instrSession));

	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	while (p2buf != NULL && i < numberofActiveSegments)
	{
		p2buf = strtok(NULL, ",");
		returnCode[i] = atol(p2buf);
		p2buf = strtok(NULL, ",");
		statisticsExpired[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		burstType[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		slotStatistics[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		outofTolerance[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");

		for (j = 0; j < numberofResults; j++)
		{
			*(results + numberofResults * i + j) = RsCore_FindStringIndex(list, p2buf);
			p2buf = strtok(NULL, ",\n");
		}
		i++;
	}
	if (returnedValues) *returnedValues = i;

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);
	return error;
}


/// HIFN Returns the power vs. time results limit state for selected segment in list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR segment/Selects the segment.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR arraySize/Sets the number of values allocated for result array.
/// HIPAR arraySize/Number of active segments.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement result
/// HIPAR returnCode/Returns the code limit state.
/// HIPAR statisticsExpired/Returns relative number of slots limit state.
/// HIPAR burstType/Returns the limit state for detected burst type of the last measured burst.
/// HIPAR slotStatistics/Returns the slot statistics limit state.
/// HIPAR outOfTolerance/Returns the limit state for percentage of measured bursts with failed limit check.
/// HIPAR usefulPartMin/Returns the limit state of minimum power across the useful part of the "Measure Slot".
/// HIPAR usefulPartMax/Returns the limit state of maximum power across the useful part of the "Measure Slot"
/// HIPAR subvector/Returns the Burst power limit state at -28 us, -18 us, -10 us, 0, 2 us, 4 us,
/// HIPAR subvector/538.2 us, 540.2 us, 542.8 us, 552.8 us, 560.8 us, 570.8 us.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasListModeCheckLimitPVTSubvector(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   resultType,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   *returnCode,
							ViInt32   *statisticsExpired,
							ViInt32   *burstType,
							ViInt32   *slotStatistics,
							ViInt32   *outOfTolerance,
							ViInt32   *usefulPartMin,
							ViInt32   *usefulPartMax,
							ViInt32   subvector[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViInt32  i;
	ViInt32  codes[210];

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 200),
			3, "Segment");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 3),
			4, "Result Type");
	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEValuation:LIST:SEGMent%ld:PVT:%s:SVEC", instance, segment, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasErrorCodesQuery(instrSession, cmd, reliability, 6+arraySize, codes, returnCode, returnedValues));

	*statisticsExpired = codes[0];
	*burstType = codes[1];
	*slotStatistics = codes[2];
	*outOfTolerance = codes[3];
	*usefulPartMin = codes[4];
	*usefulPartMax = codes[5];
	for (i = 0; i < arraySize; i++)
	{
		subvector[i] = codes[6 + i];
	}

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}


/*****************************************************************************
 * Function: Fetch Multi Eval List Mode All Modulation Results
 * Purpose:  Returns the modulation results for each active segment in list
 *           mode.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeAllModulationResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   returnCode[],
							ViReal64  statisticsExpired[],
							ViInt32   burstType[],
							ViBoolean slotStatistics[],
							ViReal64  outofTolerance[],
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 2),
			4, "Result Type");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:MOD:%s", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_MEvalMeasListModeAll(instrSession, cmd, numberofActiveSegments, arraySize,
                                            reliability, returnCode, statisticsExpired,
                                            burstType, slotStatistics, outofTolerance,results,returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}


/// HIFN Returns the limit state of modulation results for each active segment in list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberOfActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR arraySize/Sets the number of values allocated for result array.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement result.
/// HIPAR returnCode/Returns the code limit state.
/// HIPAR statisticsExpired/Returns relative number of slots limit state.
/// HIPAR burstType/Returns the limit state for detected burst type of the last measured burst.
/// HIPAR slotStatistics/Returns the slot statistics limit state.
/// HIPAR outOfTolerance/Returns limit state the percentage of measured bursts with failed limit check.
/// HIPAR results/Returns the array of limit check results. The array has to contain number of elements
/// HIPAR results/greater or equal to value passed to 12 * 'Number of Active Segments'.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasListModeAllCheckLimitModulationResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberOfActiveSegments,
							ViInt32   resultType,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   returnCode[],
							ViInt32   statisticsExpired[],
							ViInt32   burstType[],
							ViInt32   slotStatistics[],
							ViInt32   outOfTolerance[],
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberOfActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 2),
			4, "Result Type");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:MOD:%s", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_MEvalMeasListModeAllQuery(instrSession, cmd, numberOfActiveSegments, arraySize,
                                            reliability, returnCode, statisticsExpired,
                                            burstType, slotStatistics, outOfTolerance,results,returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval List Mode All Modulation Standard Deviation
 * Purpose:  Returns the standard deviation of the modulation results for each
 *           segment in list mode.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeAllModulationStandardDeviation(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   returnCode[],
							ViReal64  statisticsExpired[],
							ViInt32   burstType[],
							ViBoolean slotStatistics[],
							ViReal64  outofTolerance[],
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	         arraySize;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:MOD:SDEV", instance);
	checkErr(rscmwgm_MEvalMeasListModeAll(instrSession, cmd, numberofActiveSegments, 12,
                                            reliability, returnCode, statisticsExpired,
                                            burstType, slotStatistics, outofTolerance,results,returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}


/// HIFN Returns limit state for the standard deviation of the modulation results for each segment in list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberOfActiveSegments/Number of active segments.
/// HIPAR arraySize/Sets the number of values allocated for result array.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement result.
/// HIPAR returnCode/Returns the code limit state.
/// HIPAR statisticsExpired/Returns relative number of slots limit state.
/// HIPAR burstType/Returns the limit state for detected burst type of the last measured burst.
/// HIPAR slotStatistics/Returns the slot statistics limit state.
/// HIPAR outOfTolerance/Returns limit state the percentage of measured bursts with failed limit check.
/// HIPAR results/Returns the array of limit check results. The array has to contain number of elements
/// HIPAR results/greater or equal to value passed to 12 * 'Number of Active Segments'.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasListModeAllCheckLimitModulationStandardDeviation(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberOfActiveSegments,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   returnCode[],
							ViInt32   statisticsExpired[],
							ViInt32   burstType[],
							ViInt32   slotStatistics[],
							ViInt32   outOfTolerance[],
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberOfActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:MOD:SDEV", instance);
	checkErr(rscmwgm_MEvalMeasListModeAllQuery(instrSession, cmd, numberOfActiveSegments, arraySize,
                                            reliability, returnCode, statisticsExpired,
                                            burstType, slotStatistics, outOfTolerance, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval List Mode All Modulation Percentile
 * Purpose:  Returns the 95th percentile of the modulation results for each
 *           segment in list mode.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeAllModulationPercentile(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   returnCode[],
							ViReal64  statisticsExpired[],
							ViInt32   burstType[],
							ViBoolean slotStatistics[],
							ViReal64  outofTolerance[],
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	         arraySize;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:MOD:PERC", instance);
	checkErr(rscmwgm_MEvalMeasListModeAll(instrSession, cmd, numberofActiveSegments, 3,
                                            reliability, returnCode, statisticsExpired,
                                            burstType, slotStatistics, outofTolerance,results,returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}


/// HIFN Returns limit state for the 95th percentile of the modulation results for each segment in list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberOfActiveSegments/Number of active segments.
/// HIPAR arraySize/Sets the number of values allocated for result array.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement result.
/// HIPAR returnCode/Returns the code limit state.
/// HIPAR statisticsExpired/Returns relative number of slots limit state.
/// HIPAR burstType/Returns the limit state for detected burst type of the last measured burst.
/// HIPAR slotStatistics/Returns the slot statistics limit state.
/// HIPAR outOfTolerance/Returns limit state the percentage of measured bursts with failed limit check.
/// HIPAR results/Returns the array of limit check results. The array has to contain number of elements
/// HIPAR results/greater or equal to value passed to 12 * 'Number of Active Segments'.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasListModeAllCheckLimitModulationPercentile(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberOfActiveSegments,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   returnCode[],
							ViInt32   statisticsExpired[],
							ViInt32   burstType[],
							ViInt32   slotStatistics[],
							ViInt32   outOfTolerance[],
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberOfActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:MOD:PERC", instance);
	checkErr(rscmwgm_MEvalMeasListModeAllQuery(instrSession, cmd, numberOfActiveSegments, arraySize,
                                            reliability, returnCode, statisticsExpired,
                                            burstType, slotStatistics, outOfTolerance,results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval List Mode All Spectrum Modulation
 * Purpose:  Returns the spectrum due to modulation results for each segment
 *           in list mode.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeAllSpectrumModulation(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberOfActiveSegments,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   returnCode[],
							ViReal64  statisticsExpired[],
							ViInt32   burstType[],
							ViBoolean slotStatistics[],
							ViReal64  outOfTolerance[],
							ViReal64  carrierPower[],
							ViReal64  powerOffset[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;
	ViInt32  j = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberOfActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:SMOD", instance);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",");

	while (p2buf != NULL && i < numberOfActiveSegments)
	{
		returnCode[i] = RsCore_Convert_String_To_ViInt32(p2buf);
		p2buf = strtok(NULL, ",");
		statisticsExpired[i] = (ViInt32)RsCore_Convert_String_To_ViReal64(p2buf);
		p2buf = strtok(NULL, ",");
		burstType[i] = RsCore_FindStringIndex(listModeBurstTypeArr, p2buf);
		p2buf = strtok(NULL, ",");
		slotStatistics[i] = RsCore_Convert_String_To_Boolean(p2buf);
		p2buf = strtok(NULL, ",");
		outOfTolerance[i] = RsCore_Convert_String_To_ViReal64(p2buf);
		p2buf = strtok(NULL, ",");
		carrierPower[i] = RsCore_Convert_String_To_ViReal64(p2buf);
		p2buf = strtok(NULL, ",");
		for (j = 0; j < arraySize; j++)
		{
			*(powerOffset + arraySize * i + j) = RsCore_Convert_String_To_ViReal64(p2buf);
			p2buf = strtok(NULL, ",\n");
		}
		i++;
	}
	if (returnedValues) *returnedValues = i;

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}


/// HIFN Returns the limit state for spectrum due to modulation results for each segment in list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberOfActiveSegments/Number of active segments.
/// HIPAR arraySize/Sets the number of values allocated for result array.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement result.
/// HIPAR returnCode/Returns the code limit state.
/// HIPAR statisticsExpired/Returns relative number of slots limit state.
/// HIPAR burstType/Returns the limit state for detected burst type of the last measured burst.
/// HIPAR slotStatistics/Returns the slot statistics limit state.
/// HIPAR outOfTolerance/Returns limit state the percentage of measured bursts with failed limit check.
/// HIPAR carrierPower/Returns the measured carrier output power (reference power) limit state.
/// HIPAR powerOffset/Returns the array of average burst power limit state at the carrier frequency +/- the frequency offset.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasListModeAllCheckLimitSpectrumModulation(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberOfActiveSegments,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   returnCode[],
							ViInt32   statisticsExpired[],
							ViInt32   burstType[],
							ViInt32   slotStatistics[],
							ViInt32   outOfTolerance[],
							ViInt32   carrierPower[],
							ViInt32   powerOffset[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;
	ViInt32  j = 0;
	static   ViString list[] = {"OK", "ULEU", "ULEL", "OFL", "UFL", "INV", "NAV",NULL};

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberOfActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:SMOD", instance);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",");
	while (p2buf != NULL && i < numberOfActiveSegments)
	{
		returnCode[i] = RsCore_Convert_String_To_ViInt32(p2buf);
		p2buf = strtok(NULL, ",");
		statisticsExpired[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		burstType[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		slotStatistics[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		outOfTolerance[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		carrierPower[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		for (j = 0; j < arraySize; j++)
		{
			*(powerOffset + arraySize * i + j) = RsCore_FindStringIndex(list, p2buf);
			p2buf = strtok(NULL, ",");
		}
		i++;
	}
	if (returnedValues) *returnedValues = i;

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval List Mode All Spectrum Switching
 * Purpose:  Returns the spectrum due to switching results for each segment in
 *           list mode.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeAllSpectrumSwitching(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   returnCode[],
							ViReal64  statisticsExpired[],
							ViInt32   burstType[],
							ViBoolean slotStatistics[],
							ViReal64  outofTolerance[],
							ViReal64  carrierPower[],
							ViReal64  powerOffset[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;
	ViInt32  j = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:SSW", instance);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",");
	while (p2buf != NULL && i < numberofActiveSegments)
	{
		returnCode[i] = RsCore_Convert_String_To_ViInt32(p2buf);
		p2buf = strtok(NULL, ",");
		statisticsExpired[i] = RsCore_Convert_String_To_ViReal64(p2buf);
		p2buf = strtok(NULL, ",");
		burstType[i] = RsCore_FindStringIndex(listModeBurstTypeArr, p2buf);
		p2buf = strtok(NULL, ",");
		slotStatistics[i] = RsCore_Convert_String_To_Boolean(p2buf);
		p2buf = strtok(NULL, ",");
		outofTolerance[i] = RsCore_Convert_String_To_ViReal64(p2buf);
		p2buf = strtok(NULL, ",");
		carrierPower[i] = RsCore_Convert_String_To_ViReal64(p2buf);
		p2buf = strtok(NULL, ",");
		for (j = 0; j < arraySize; j++)
		{
			*(powerOffset + arraySize * i + j) = RsCore_Convert_String_To_ViReal64(p2buf);
			p2buf = strtok(NULL, ",");
		}
		i++;
	}
	if (returnedValues) *returnedValues = i;

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/// HIFN Returns the limit state for spectrum due to modulation results for each segment in list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberOfActiveSegments/Number of active segments.
/// HIPAR arraySize/Sets the number of values allocated for result array.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement result.
/// HIPAR returnCode/Returns the code limit state.
/// HIPAR statisticsExpired/Returns relative number of slots limit state.
/// HIPAR burstType/Returns the limit state for detected burst type of the last measured burst.
/// HIPAR slotStatistics/Returns the slot statistics limit state.
/// HIPAR outOfTolerance/Returns limit state the percentage of measured bursts with failed limit check.
/// HIPAR carrierPower/Returns the measured carrier output power (reference power) limit state.
/// HIPAR powerOffset/Returns the array of average burst power limit state at the carrier frequency +/- the frequency offset.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasListModeAllCheckLimitSpectrumSwitching(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberOfActiveSegments,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   returnCode[],
							ViInt32   statisticsExpired[],
							ViInt32   burstType[],
							ViInt32   slotStatistics[],
							ViInt32   outOfTolerance[],
							ViInt32   carrierPower[],
							ViInt32   powerOffset[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;
	ViInt32  j = 0;
	static   ViString list[] = {"OK", "ULEU", "ULEL", "OFL", "UFL", "INV", "NAV",NULL};

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberOfActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:SSW", instance);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",");
	while (p2buf != NULL && i < numberOfActiveSegments)
	{
		returnCode[i] = RsCore_Convert_String_To_ViInt32(p2buf);
		p2buf = strtok(NULL, ",");
		statisticsExpired[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		burstType[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		slotStatistics[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		outOfTolerance[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		carrierPower[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		for (j = 0; j < arraySize; j++)
		{
			*(powerOffset + arraySize * i + j) = RsCore_FindStringIndex(list, p2buf);
			p2buf = strtok(NULL, ",");
		}
		i++;
	}
	if (returnedValues) *returnedValues = i;

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval List Mode All Spectrum BER
 * Purpose:  Returns the BER results in list mode. To define the statistical
 *           length and enable the calculation of the BER results, use the
 *           command CONFigure:GSM:MEAS:MEValuation:LIST:SEGMent<Seg>:BER
 *           <Statistics>, <BER Enable>.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeAllSpectrumBER(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   *reliability,
							ViInt32   returnCode[],
							ViReal64  statisticsExpired[],
							ViInt32   burstType[],
							ViBoolean slotStatistics[],
							ViInt32   ber[],
							ViInt32   errors[],
							ViInt32   bursts[])
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:BER", instance);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	for (i = 0; i < numberofActiveSegments; i++)
	{
		p2buf = strtok(NULL, ",");
		returnCode[i] = RsCore_Convert_String_To_ViInt32(p2buf);
		p2buf = strtok(NULL, ",");
		statisticsExpired[i] = RsCore_Convert_String_To_ViReal64(p2buf);
		p2buf = strtok(NULL, ",");
		burstType[i] = RsCore_FindStringIndex(listModeBurstTypeArr, p2buf);
		p2buf = strtok(NULL, ",");
		slotStatistics[i] = RsCore_Convert_String_To_Boolean(p2buf);
		p2buf = strtok(NULL, ",");
		ber[i] = (ViInt32)RsCore_Convert_String_To_ViReal64(p2buf);
		p2buf = strtok(NULL, ",");
		errors[i] = (ViInt32)RsCore_Convert_String_To_ViReal64(p2buf);
		p2buf = strtok(NULL, ",");
		bursts[i] = (ViInt32)RsCore_Convert_String_To_ViReal64(p2buf);
		if (p2buf == NULL) break;
	}

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/// HIFN Returns the BER limiti check results in list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberOfActiveSegments/Number of active segments.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement result.
/// HIPAR returnCode/Returns the code limit state.
/// HIPAR statisticsExpired/Returns relative number of slots limit state.
/// HIPAR burstType/Returns the limit state for detected burst type of the last measured burst.
/// HIPAR slotStatistics/Returns the slot statistics limit state.
/// HIPAR BER/Returns the limit state for percentage of Bit Error Rate.
/// HIPAR errors/Returns limit state for total number of detected bit errors.
/// HIPAR bursts/Returns limit state for total number of measured bursts.
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasListModeAllCheckLimitSpectrumBER(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberOfActiveSegments,
							ViInt32   *reliability,
							ViInt32   returnCode[],
							ViInt32   statisticsExpired[],
							ViInt32   burstType[],
							ViInt32   slotStatistics[],
							ViInt32   BER[],
							ViInt32   errors[],
							ViInt32   bursts[])
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;
	static   ViString list[] = {"OK", "ULEU", "ULEL", "OFL", "UFL", "INV", "NAV",NULL};

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberOfActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:BER", instance);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	for (i = 0; i < numberOfActiveSegments; i++)
	{
		p2buf = strtok(NULL, ",");
		returnCode[i] = RsCore_Convert_String_To_ViInt32(p2buf);
		p2buf = strtok(NULL, ",");
		statisticsExpired[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		burstType[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		slotStatistics[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		BER[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		errors[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		bursts[i] = RsCore_FindStringIndex(list, p2buf);
		if (p2buf == NULL) break;
	}

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval List Mode All Power vs Time Results
 * Purpose:  Returns the power vs. time results for each segment in list mode.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeAllPVTResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   *reliability,
							ViInt32   returnCode[],
							ViReal64  statisticsExpired[],
							ViInt32   burstType[],
							ViBoolean slotStatistics[],
							ViReal64  outofTolerance[],
							ViReal64  averageBurstPower[])
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 1),
			4, "Result Type");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:PVT:%s", instance, resultTypeArr[resultType]);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	for (i = 0; i < numberofActiveSegments; i++)
	{
		p2buf = strtok(NULL, ",");
		returnCode[i] = RsCore_Convert_String_To_ViInt32(p2buf);
		p2buf = strtok(NULL, ",");
		statisticsExpired[i] = RsCore_Convert_String_To_ViReal64(p2buf);
		p2buf = strtok(NULL, ",");
		burstType[i] = RsCore_FindStringIndex(listModeBurstTypeArr, p2buf);
		p2buf = strtok(NULL, ",");
		slotStatistics[i] = RsCore_Convert_String_To_Boolean(p2buf);
		p2buf = strtok(NULL, ",");
		outofTolerance[i] = RsCore_Convert_String_To_ViReal64(p2buf);
		p2buf = strtok(NULL, ",");
		averageBurstPower[i] = RsCore_Convert_String_To_ViReal64(p2buf);
	}

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/// HIFN Returns the limit state of the power vs. time results for each segment in list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberOfActiveSegments/Number of active segments.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement result.
/// HIPAR returnCode/Returns the code limit state.
/// HIPAR statisticsExpired/Returns relative number of slots limit state.
/// HIPAR burstType/Returns the limit state for detected burst type of the last measured burst.
/// HIPAR slotStatistics/Returns the slot statistics limit state.
/// HIPAR outOfTolerance/Returns the limit for percentage of measured bursts with failed limit check.
/// HIPAR averageBurstPower/Returns the average burst power limit state.
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasListModeAllCheckLimitPVTResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberOfActiveSegments,
							ViInt32   resultType,
							ViInt32   *reliability,
							ViInt32   returnCode[],
							ViInt32   statisticsExpired[],
							ViInt32   burstType[],
							ViInt32   slotStatistics[],
							ViInt32   outOfTolerance[],
							ViInt32   averageBurstPower[])
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;
	static   ViString list[] = {"OK", "ULEU", "ULEL", "OFL", "UFL", "INV", "NAV",NULL};

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberOfActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 1),
			4, "Result Type");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:PVT:%s", instance, resultTypeArr[resultType]);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	for (i = 0; i < numberOfActiveSegments; i++)
	{
		p2buf = strtok(NULL, ",");
		returnCode[i] = RsCore_Convert_String_To_ViInt32(p2buf);
		p2buf = strtok(NULL, ",");
		statisticsExpired[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		burstType[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		slotStatistics[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		outOfTolerance[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		averageBurstPower[i] = RsCore_FindStringIndex(list, p2buf);
	}

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval List Mode All Power vs Time Subvector
 * Purpose:  Returns special burst power values for the "Measure Slot" for
 *           each segment in list mode.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeAllPVTSubvector(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberOfActiveSegments,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   returnCode[],
							ViReal64  statisticsExpired[],
							ViInt32   burstType[],
							ViBoolean slotStatistics[],
							ViReal64  outOfTolerance[],
							ViReal64  usefulPartMin[],
							ViReal64  usefulPartMax[],
							ViReal64  subvector[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;
	ViInt32  j = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberOfActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:PVT:CURR:SVEC", instance);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",");
	while (p2buf != NULL && i < numberOfActiveSegments)
	{
		returnCode[i] = RsCore_Convert_String_To_ViInt32(p2buf);
		p2buf = strtok(NULL, ",");
		statisticsExpired[i] = (ViInt32)RsCore_Convert_String_To_ViReal64(p2buf);
		p2buf = strtok(NULL, ",");
		burstType[i] = RsCore_FindStringIndex(listModeBurstTypeArr, p2buf);
		p2buf = strtok(NULL, ",");
		slotStatistics[i] = RsCore_Convert_String_To_Boolean(p2buf);
		p2buf = strtok(NULL, ",");
		outOfTolerance[i] = RsCore_Convert_String_To_ViReal64(p2buf);
		p2buf = strtok(NULL, ",");
		usefulPartMin[i] = RsCore_Convert_String_To_ViReal64(p2buf);
		p2buf = strtok(NULL, ",");
		usefulPartMax[i] = RsCore_Convert_String_To_ViReal64(p2buf);
		p2buf = strtok(NULL, ",");
		for (j = 0; j < arraySize; j++)
		{
			*(subvector + arraySize * i + j) = RsCore_Convert_String_To_ViReal64(p2buf);
			p2buf = strtok(NULL, ",");
		}
		i++;
	}
	if (returnedValues) *returnedValues = i;

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}


/// HIFN Returns the limit state of modulation results for selected segment in list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR segment/Selects the segment.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR arraySize/Sets the number of values allocated for result array.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement result
/// HIPAR returnCode/Returns the code limit state.
/// HIPAR statisticsExpired/Returns relative number of slots limit state.
/// HIPAR burstType/Returns the limit state for detected burst type of the last measured burst.
/// HIPAR slotStatistics/Returns the slot statistics limit state.
/// HIPAR outOfTolerance/Returns the limit state for percentage of measured bursts with failed limit check.
/// HIPAR results/Returns the array of results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasListModeCheckLimitModulationResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   resultType,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   *returnCode,
							ViInt32   *statisticsExpired,
							ViInt32   *burstType,
							ViInt32   *slotStatistics,
							ViInt32   *outOfTolerance,
							ViInt32   *results,
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViInt32  i;
	ViInt32  codes[20];

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 200),
			3, "Segment");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, 0, 2),
			4, "Result Type");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:SEGM%ld:MOD:%s", instance, segment, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasErrorCodesQuery(instrSession, cmd,reliability,4+arraySize,codes,returnCode,returnedValues));

	*statisticsExpired = codes[0];
	*burstType = codes[1];
	*slotStatistics = codes[2];
	*outOfTolerance = codes[3];

	for (i = 0; i < arraySize; i++)
	{
		results[i] = codes[4 + i];
	}

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Fetch Multi Eval List Mode Modulation Standard Deviation
 * Purpose:  Returns the standard deviation of the modulation results for
 *           selected segment in list mode.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeModulationStandardDeviation(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   segment,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   *returnCode,
							ViReal64  *statisticsExpired,
							ViInt32   *burstType,
							ViBoolean *slotStatistics,
							ViReal64  *out_ofTolerance,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, segment, 1, 200),
			3, "Segment");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:SEGM%ld:MOD:SDEV", instance, segment);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",");
	*returnCode = RsCore_Convert_String_To_ViInt32(p2buf);
	p2buf = strtok(NULL, ",");
	*statisticsExpired = RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");
	*burstType = RsCore_FindStringIndex(listModeBurstTypeArr, p2buf);
	p2buf = strtok(NULL, ",");
	*slotStatistics = RsCore_Convert_String_To_Boolean(p2buf);
	p2buf = strtok(NULL, ",");
	*out_ofTolerance = RsCore_Convert_String_To_ViReal64(p2buf);
	p2buf = strtok(NULL, ",");

	while (p2buf != NULL)
	{
		if (i < arraySize)
		{
			results[i++] = RsCore_Convert_String_To_ViReal64(p2buf);
		}
		p2buf = strtok(NULL, ",\n");
	}
	if (returnedValues) *returnedValues = i;

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/// HIFN Returns the limit state of special burst power values for the "Measure Slot" for each segment in list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberOfActiveSegments/Number of active segments.
/// HIPAR arraySize/Sets the number of values allocated for result array.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement result.
/// HIPAR returnCode/Returns the code limit state.
/// HIPAR statisticsExpired/Returns relative number of slots limit state.
/// HIPAR burstType/Returns the limit state for detected burst type of the last measured burst.
/// HIPAR slotStatistics/Returns the slot statistics limit state.
/// HIPAR outOfTolerance/Returns limit state the percentage of measured bursts with failed limit check.
/// HIPAR usefulPartMin/Returns the limit state of minimum power across the useful part of the "Measure Slot".
/// HIPAR usefulPartMax/Returns the limit state of maximum power across the useful part of the "Measure Slot".
/// HIPAR subvector/Returns the Burst power limit state at -28 us, -18 us, -10 us, 0, 2 us, 4 us, 538.2 us, 540.2 us, 542.8 us, 552.8 us, 560.8 us, 570.8 us.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasListModeAllCheckLimitPVTSubvector(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberOfActiveSegments,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   returnCode[],
							ViInt32   statisticsExpired[],
							ViInt32   burstType[],
							ViInt32   slotStatistics[],
							ViInt32   outOfTolerance[],
							ViInt32   usefulPartMin[],
							ViInt32   usefulPartMax[],
							ViInt32   subvector[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;
	ViInt32  j = 0;
	static   ViString list[] = {"OK", "ULEU", "ULEL", "OFL", "UFL", "INV", "NAV",NULL};

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberOfActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:PVT:CURR:SVEC", instance);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",");
	while (p2buf != NULL && i < numberOfActiveSegments)
	{
		returnCode[i] = RsCore_Convert_String_To_ViInt32(p2buf);
		p2buf = strtok(NULL, ",");
		statisticsExpired[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		burstType[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		slotStatistics[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		outOfTolerance[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		usefulPartMin[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		usefulPartMax[i] = RsCore_FindStringIndex(list, p2buf);
		p2buf = strtok(NULL, ",");
		for (j = 0; j < arraySize; j++)
		{
			*(subvector + arraySize * i + j) = RsCore_FindStringIndex(list, p2buf);
			p2buf = strtok(NULL, ",");
		}
		i++;
	}
	if (returnedValues) *returnedValues = i;

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/// HIFN  Returns the segment reliability for all measured list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberOfActiveSegments/Number of active segments.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of integer values of the reliability indicator for all measured
/// HIPAR results/list mode segments.
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeAllReliabilityResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberOfActiveSegments,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedResults)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:SREL", instance);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;

	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",");
	while ((p2buf != NULL) && (i < numberOfActiveSegments))
	{
		results[i++] = RsCore_Convert_String_To_ViInt32(p2buf);
		p2buf = strtok(NULL, ",");
	}
	*returnedResults = i;

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN Returns all single results for each segment in list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberOfActiveSegments/Number of active segments.
/// HIPAR arraySize/Array size.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement result.
/// HIPAR reliability/Document parameter here.
/// HIPAR returnCode/Returns the code.
/// HIPAR results/Returns the array of results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeAllOverviewResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberOfActiveSegments,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   returnCode[],
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;
	ViInt32  j = 0;

	returnedValues;


	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:OVER", instance);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",");
	while (p2buf != NULL && i < numberOfActiveSegments)
	{
		returnCode[i] = RsCore_Convert_String_To_ViInt32(p2buf);
		p2buf = strtok(NULL, ",");
		for (j = 0; j < arraySize; j++)
		{
			*(results + j + arraySize * i) = RsCore_Convert_String_To_ViReal64(p2buf);
			p2buf = strtok(NULL, ",");
		}
		i++;
	}

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/// HIFN Returns all single results limit state for each segment in list mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberOfActiveSegments/Number of active segments.
/// HIPAR arraySize/Array size.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement result.
/// HIPAR reliability/Document parameter here.
/// HIPAR returnCode/Returns the code.
/// HIPAR results/Returns the array of results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMEvalMeasListModeAllCheckLimitOverviewResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberOfActiveSegments,
							ViInt32   arraySize,
							ViInt32   *reliability,
							ViInt32   returnCode[],
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViChar   *ext_buf = NULL;
	ViChar   *p2buf;
	ViInt32  i = 0;
	ViInt32  j = 0;
	static   ViString list[] = {"OK", "ULEU", "ULEL", "OFL", "UFL", "INV", "NAV",NULL};

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:OVER", instance);
	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &ext_buf));
	checkErr(rscmwgm_CheckStatus(instrSession));

	p2buf = strtok(ext_buf, ",");
	if (p2buf == NULL)
		return RSCMWGM_ERROR_INCORRECT_RESPONSE;
	if (reliability != NULL)
		*reliability = atol(p2buf);

	p2buf = strtok(NULL, ",");
	while (p2buf != NULL && i < numberOfActiveSegments)
	{
		returnCode[i] = RsCore_Convert_String_To_ViInt32(p2buf);
		p2buf = strtok(NULL, ",");
		for (j = 0; j < arraySize; j++)
		{
			*(results + j + arraySize * i) = RsCore_FindStringIndex(list, p2buf);
			p2buf = strtok(NULL, ",");
		}
		i++;
	}

	if (returnedValues) *returnedValues = i;

Error:
	if (ext_buf) free(ext_buf);
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/// HIFN  Returns IQ origin offset for all measured list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of I/Q origin offset  values.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMultiEvalListModeAllIQOffsetResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_SDEV) || (resultType == RSCMWGM_VAL_RESULT_MIN), 3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:MOD:IQOF:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns IQ origin limit check results for all measured list mode
/// HIFN  segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of limit check results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMultiEvalListModeAllCheckLimitIQOffsetResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_MAX),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:MOD:IQOF:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, numberofActiveSegments, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns IQ Imbalance for all measured list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of IQ Imbalance values.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMultiEvalListModeAllIQImbalanceResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_SDEV) || (resultType == RSCMWGM_VAL_RESULT_MIN), 3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:MOD:IQIM:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns IQ imbalance limit check results for all measured list mode
/// HIFN  segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of limit check results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMultiEvalListModeAllCheckLimitIQImbalanceResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_MAX),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:MOD:IQIM:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, numberofActiveSegments, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns  Carrier Frequency Error for all measured list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of Carrier Frequency Error values.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMultiEvalListModeAllCarrierFrequencyErrorResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_SDEV) || (resultType == RSCMWGM_VAL_RESULT_MIN), 3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:MOD:FERR:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns carrier frequency error limit check results for all measured
/// HIFN  list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of limit check results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMultiEvalListModeAllCheckLimitCarrierFrequencyErrorResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_MAX),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:MOD:FERR:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, numberofActiveSegments, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns  transmit time error for all measured list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of transmit time error values.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMultiEvalListModeAllTransmitTimeErrorResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_SDEV) || (resultType == RSCMWGM_VAL_RESULT_MIN), 3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:MOD:TERR:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasIntValues(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns transmit time error limit check results for all measured list
/// HIFN  mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of limit check results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMultiEvalListModeAllCheckLimitTransmitTimeErrorResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_MAX),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:MOD:TERR:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, numberofActiveSegments, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns  burst power for all measured list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of burst power values.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMultiEvalListModeAllBurstPowerResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_SDEV) || (resultType == RSCMWGM_VAL_RESULT_MIN), 3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:MOD:BPOW:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns burst power limit check results for all measured list mode
/// HIFN  segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of limit check results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMultiEvalListModeAllCheckLimitBurstPowerResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_MAX),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:MOD:BPOW:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, numberofActiveSegments, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns  AM-PM delay for all measured list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of AM-PM delay values.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMultiEvalListModeAllAMPMDelayResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_SDEV) || (resultType == RSCMWGM_VAL_RESULT_MIN), 3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:MOD:APD:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns AM-PM delay limit check results for all measured list mode
/// HIFN  segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of limit check results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMultiEvalListModeAllCheckLimitAMPMDelayResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_MAX),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:MOD:APD:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, numberofActiveSegments, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns error vector magnitude values for all measured list mode
/// HIFN  segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR type/Denotes the result type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of error vector magnitude values.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMultiEvalListModeAllEVMResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   type,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_SDEV) || (resultType == RSCMWGM_VAL_RESULT_MIN), 3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, type, RSCMWGM_VAL_TYPE_RMS, RSCMWGM_VAL_TYPE_PEAK),
			5, "Type");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:MOD:EVM:%s:%s?", instance, detectorTypeArr[type], resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns error vector magnitude limit check results for all measured
/// HIFN  list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR type/Denotes the result type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of limit check results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMultiEvalListModeAllCheckLimitEVMResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   type,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_MAX),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, type, RSCMWGM_VAL_TYPE_RMS, RSCMWGM_VAL_TYPE_PEAK),
			5, "Type");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:MOD:EVM:%s:%s?", instance, detectorTypeArr[type], resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, numberofActiveSegments, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Return error vector magnitude 95th percentile values for all measured
/// HIFN  list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of error vector magnitude percentile values.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMultiEvalListModeAllEVMPercentile(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:MOD:EVM:PERC?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns error vector magnitude 95th percentile limit check results for
/// HIFN  all measured list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of limit check results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMultiEvalListModeAllCheckLimitEVMPercentile(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:MOD:EVM:PERC?", instance);
	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, numberofActiveSegments, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns magnitude error values for all measured list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR type/Denotes the result type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of magnitude error values.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMultiEvalListModeAllMagnitudeErrorResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   type,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_SDEV) || (resultType == RSCMWGM_VAL_RESULT_MIN), 3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, type, RSCMWGM_VAL_TYPE_RMS, RSCMWGM_VAL_TYPE_PEAK),
			5, "Type");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:MOD:MERR:%s:%s?", instance, detectorTypeArr[type], resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns magnitude error limit check results for all measured list mode
/// HIFN  segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR type/Denotes the result type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of limit check results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMultiEvalListModeAllCheckLimitMagnitudeErrorResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   type,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_MAX),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, type, RSCMWGM_VAL_TYPE_RMS, RSCMWGM_VAL_TYPE_PEAK),
			5, "Type");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:MOD:MERR:%s:%s?", instance, detectorTypeArr[type], resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, numberofActiveSegments, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns magnitude error 95th percentile values for all measured list
/// HIFN  mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of magnitude error percentile values.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMultiEvalListModeAllMagnitudeErrorPercentile(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:MOD:MERR:PERC?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns magnitude error 95th percentile limit check results for all
/// HIFN  measured list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of limit check results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMultiEvalListModeAllCheckLimitMagnitudeErrorPercentile(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:MOD:MERR:PERC?", instance);
	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability,numberofActiveSegments, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns phase error values for all measured list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR type/Denotes the result type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of phase error values.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMultiEvalListModeAllPhaseErrorResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   type,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_SDEV) || (resultType == RSCMWGM_VAL_RESULT_MIN), 3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, type, RSCMWGM_VAL_TYPE_RMS, RSCMWGM_VAL_TYPE_PEAK),
			5, "Type");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:MOD:PERR:%s:%s?", instance, detectorTypeArr[type], resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns phase error limit check results for all measured list mode
/// HIFN  segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR type/Denotes the result type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of limit check results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMultiEvalListModeAllCheckLimitPhaseErrorResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   type,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_MAX),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, type, RSCMWGM_VAL_TYPE_RMS, RSCMWGM_VAL_TYPE_PEAK),
			5, "Type");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:MOD:PERR:%s:%s?", instance, detectorTypeArr[type], resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, numberofActiveSegments, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns phase error 95th percentile values for all measured list mode
/// HIFN  segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of phase error percentile values.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMultiEvalListModeAllPhaseErrorPercentile(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:MOD:PERR:PERC?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns phase error 95th percentile limit check results for all
/// HIFN  measured list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of limit check results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMultiEvalListModeAllCheckLimitPhaseErrorPercentile(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:MOD:PERR:PERC?", instance);
	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, numberofActiveSegments, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns carrier output power values for all measured list mode
/// HIFN  segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of carrier output power values.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMultiEvalListModeAllSpectrumModulationCarrierPowerResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:SMOD:CPOW?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns carrier power limit check results for all measured list mode
/// HIFN  segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of limit check results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMultiEvalListModeAllCheckLimitSpectrumModulationCarrierPower(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:SMOD:CPOW?", instance);
	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, numberofActiveSegments, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns burst power values at the carrier frequency minus/plus a
/// HIFN  selected frequency offset for all measured list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR frequencyOffset/Index selecting one of the configured frequency offsets.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of the burst power at the carrier frequency minus/plus a
/// HIPAR results/selected frequency offset.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMultiEvalListModeAllSpectrumModulationBurstPowerResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   frequencyOffset,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, frequencyOffset, -20, 20), 4, "Frequency Offset");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:SMOD:POFF%ld?", instance, frequencyOffset + 21);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns burst power limit check results for all measured list mode
/// HIFN  segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR frequencyOffset/Index selecting one of the configured frequency offsets.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of limit check results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMultiEvalListModeAllCheckLimitSpectrumModulationBurstPower(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   frequencyOffset,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, frequencyOffset, -20, 20), 4, "Frequency Offset");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:SMOD:POFF%ld?", instance, frequencyOffset + 21);
	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, numberofActiveSegments, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns carrier output power values for all measured list mode
/// HIFN  segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of carrier output power values.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMultiEvalListModeAllSpectrumSwitchingCarrierPowerResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:SSW:CPOW?", instance);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns carrier power limit check results for all measured list mode
/// HIFN  segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of limit check results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMultiEvalListModeAllCheckLimitSpectrumSwitchingCarrierPower(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:SSW:CPOW?", instance);
	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, numberofActiveSegments, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns burst power values at the carrier frequency minus/plus a
/// HIFN  selected frequency offset for all measured list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR frequencyOffset/Index selecting one of the configured frequency offsets.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of the burst power values at the carrier frequency minus/plus a
/// HIPAR results/selected frequency offset.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMultiEvalListModeAllSpectrumSwitchingBurstPowerResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   frequencyOffset,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, frequencyOffset, -20, 20), 4, "Frequency Offset");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:SSW:POFF%ld?", instance, frequencyOffset + 21);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns burst power limit check results for all measured list mode
/// HIFN  segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR frequencyOffset/Index selecting one of the configured frequency offsets.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of limit check results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMultiEvalListModeAllCheckLimitSpectrumSwitchingBurstPower(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   frequencyOffset,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, frequencyOffset, -20, 20), 4, "Frequency Offset");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:SSW:POFF%ld?", instance, frequencyOffset + 21);
	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, numberofActiveSegments, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns the BER values for all measured list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Select required result.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of BER values.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeAllBERResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViString BERResultArr[] = {"BER", "ABS", "COUN"};

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_BER, RSCMWGM_VAL_RESULT_BER_BURSTS),
			4, "Result Type");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:BER:%s?", instance, BERResultArr[resultType]);
	checkErr(rscmwgm_GetMeasIntValues(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns average burst power for all measured list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of burst power values.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeAllPVTAverageBurstPowerResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_AVERAGE),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:PVT:ABP:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns average burst power limit check results for all measured list
/// HIFN  mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of limit check results.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_QueryMultiEvalListModeAllCheckLimitPVTAverageBurstPowerResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   *reliability,
							ViInt32   results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_AVERAGE),
			3, "Number of Active Segments");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:GSM:MEAS%ld:MEV:LIST:PVT:ABP:%s?", instance, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasErrorCodes(instrSession, cmd, reliability, numberofActiveSegments, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns power across the useful part of the burst for all measured
/// HIFN  list mode segments.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR powerType/Selects maximum or minimum power values.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of burst power values.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeAllPVTUsefulPartResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   powerType,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_MIN),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, powerType, RSCMWGM_VAL_RESULT_MAX, RSCMWGM_VAL_RESULT_MIN),
			5, "Power Type");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:PVT:SVEC:U%s:%s?", instance, resultTypeArr[powerType], resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Returns burst power at a specific burst position for all measured list
/// HIFN  mode segments
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR numberofActiveSegments/Number of active segments.
/// HIPAR resultType/Denotes the statistics type.
/// HIPAR burstPosition/Selects burst position.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR results/Array of burst power values.
/// HIPAR returnedValues/Returns number of results returned from the instrument.
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasListModeAllPVTSubvectorAtPositionResults(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   numberofActiveSegments,
							ViInt32   resultType,
							ViInt32   burstPosition,
							ViInt32   *reliability,
							ViReal64  results[],
							ViInt32   *returnedValues)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, numberofActiveSegments, 1, 200),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, resultType, RSCMWGM_VAL_RESULT_CURRENT, RSCMWGM_VAL_RESULT_MIN),
			3, "Number of Active Segments");
	viCheckParm(RsCore_InvalidViInt32Range(instrSession, burstPosition, RSCMWGM_VAL_BURST_POSITION_NEG_28, RSCMWGM_VAL_BURST_POSITION_570_8),
			5, "Burst Position");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:LIST:PVT:SVEC:SUBV%ld:%s?", instance, burstPosition, resultTypeArr[resultType]);
	checkErr(rscmwgm_GetMeasTrace(instrSession, cmd, 0, numberofActiveSegments, reliability, results, returnedValues));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN Returns the modulation view throughput, i.e. the percentage of
/// HIFN measurement intervals where the detected burst pattern was found
/// HIFN to correspond to the Modulation View settings.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR instance/No. of generator firmware application.
/// HIPAR reliability/Integer value of the reliability indicator for the entire measurement
/// HIPAR reliability/result; see Reliability Indicator.
/// HIPAR result/Returns the result of the measurement.
ViStatus _VI_FUNC rscmwgm_FetchMEvalMeasModulationViewThroughput(
							ViSession instrSession,
							ViInt32   instance,
							ViInt32   *reliability,
							ViReal64  *result)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
	ViReal64 aux;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, instance, 1, 16),
			2, "Instance");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, reliability), 3, "Reliability");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, result), 4, "Result");

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:GSM:MEAS%ld:MEV:MVTH", instance);
	checkErr(RsCore_QueryTupleViReal64(instrSession, cmd, &aux, result, NULL, NULL));

	*reliability = (ViInt32)aux;

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: rscmwgm_SetAttribute<type> Functions
 * Purpose:  These functions enable the instrument driver user to set
 *           attribute values directly.  There are typesafe versions for
 *           ViInt32, ViReal64, ViString, ViBoolean, and ViSession datatypes.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_SetAttributeViInt32(ViSession instrSession,
                                              ViString  channelName,
                                              ViUInt32  attributeId,
                                              ViInt32   value)
{
	ViStatus error = VI_SUCCESS;

	(void)RsCore_LockSession(instrSession);

	error = RsCore_SetAttributeViInt32(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/* ViReal64 */
ViStatus _VI_FUNC rscmwgm_SetAttributeViReal64(ViSession instrSession,
                                               ViString  channelName,
                                               ViUInt32  attributeId,
                                               ViReal64  value)
{
	ViStatus error = VI_SUCCESS;

	(void)RsCore_LockSession(instrSession);

	error = RsCore_SetAttributeViReal64(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/* ViString */
ViStatus _VI_FUNC rscmwgm_SetAttributeViString(ViSession instrSession,
                                               ViString  channelName,
                                               ViUInt32  attributeId,
                                               ViString  value)
{
	ViStatus error = VI_SUCCESS;

	(void)RsCore_LockSession(instrSession);

	error = RsCore_SetAttributeViString(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/* ViBoolean */
ViStatus _VI_FUNC rscmwgm_SetAttributeViBoolean(ViSession instrSession,
                                                ViString  channelName,
                                                ViUInt32  attributeId,
                                                ViBoolean value)
{
	ViStatus error = VI_SUCCESS;

	(void)RsCore_LockSession(instrSession);

	error = RsCore_SetAttributeViBoolean(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/* ViSession */
ViStatus _VI_FUNC rscmwgm_SetAttributeViSession(ViSession instrSession,
                                                ViString  channelName,
                                                ViUInt32  attributeId,
                                                ViSession value)
{
	ViStatus error = VI_SUCCESS;

	(void)RsCore_LockSession(instrSession);

	error = RsCore_SetAttributeViSession(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: rscmwgm_GetAttribute<type> Functions
 * Purpose:  These functions enable the instrument driver user to get
 *           attribute values directly.  There are typesafe versions for
 *           ViInt32, ViReal64, ViString, ViBoolean, and ViSession attributes.
 *****************************************************************************/

/* ViInt32 */
ViStatus _VI_FUNC rscmwgm_GetAttributeViInt32(ViSession instrSession,
                                              ViString  channelName,
                                              ViUInt32  attributeId,
                                              ViInt32   *value)
{
	ViStatus error = VI_SUCCESS;

	(void)RsCore_LockSession(instrSession);
	error = RsCore_GetAttributeViInt32(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/* ViReal64 */
ViStatus _VI_FUNC rscmwgm_GetAttributeViReal64(ViSession instrSession,
                                               ViString  channelName,
                                               ViUInt32  attributeId,
                                               ViReal64  *value)
{
	ViStatus error = VI_SUCCESS;

	(void)RsCore_LockSession(instrSession);

	error = RsCore_GetAttributeViReal64(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/* ViString */
ViStatus _VI_FUNC rscmwgm_GetAttributeViString(ViSession instrSession,
                                               ViString  channelName,
                                               ViUInt32  attributeId,
                                               ViInt32   bufSize,
                                               ViChar    value[])
{
	ViStatus error = VI_SUCCESS;

	(void)RsCore_LockSession(instrSession);

	error = RsCore_GetAttributeViString(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, bufSize,
	                                    value);

	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/* ViBoolean */
ViStatus _VI_FUNC rscmwgm_GetAttributeViBoolean(ViSession instrSession,
                                                ViString  channelName,
                                                ViUInt32  attributeId,
                                                ViBoolean *value)
{
	ViStatus error = VI_SUCCESS;

	(void)RsCore_LockSession(instrSession);

	error = RsCore_GetAttributeViBoolean(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/* ViSession */
ViStatus _VI_FUNC rscmwgm_GetAttributeViSession(ViSession instrSession,
                                                ViString  channelName,
                                                ViUInt32  attributeId,
                                                ViSession *value)
{
	ViStatus error = VI_SUCCESS;

	(void)RsCore_LockSession(instrSession);

	error = RsCore_GetAttributeViSession(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN Function used by Hi-level functions to obtain a repCap name from:
/// HIFN - attributeID
/// HIFN - repCapNameID
/// HIFN - 0-based index in the repCapNames
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR attributeId/Pass the ID of an attribute.
/// HIPAR repCapNameId/RepCap Name ID.
/// HIPAR index/Zero based index in the repCapNames.
/// HIPAR bufferSize/Pass the number of bytes in the string you allocate for the repCapName.
/// HIPAR repCapName/String returning RepCap name
ViStatus _VI_FUNC rscmwgm_GetAttributeRepCapName(ViSession instrSession,
                                                 ViUInt32  attributeId,
                                                 ViString  repCapNameId,
                                                 ViInt32   index,
                                                 ViInt32   bufferSize,
                                                 ViChar    repCapName[])
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	checkErr(RsCore_GetAttributeRepCapName(instrSession, attributeId, repCapNameId, index, bufferSize, repCapName));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Rs_GetAttributeRepCapNameId
 * Purpose:  This function returns the repeated capability id(s) that
 *           belongs to the attribute you specify. More then one id is listed
 *           in comma separated string.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_GetAttributeRepeatedCapabilityIds(ViSession instrSession,
                                                            ViUInt32  attributeID,
                                                            ViInt32   bufferSize,
                                                            ViChar    _VI_FAR repeatedCapabilityIds[])
{
	ViStatus error = VI_SUCCESS;

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, bufferSize, 1, RS_MAX_SHORT_MESSAGE_BUF_SIZE),
			3, "Buffer Size");

	checkErr(RsCore_GetAttributeRepCapNameIds(instrSession, attributeID, bufferSize, repeatedCapabilityIds));

Error:
	return error;
}

/*****************************************************************************
 * Function: Get Attribute Repeated Capability Id Name(s)
 * Purpose:  This function returns the repeated capability name(s) that
 *           belongs to the attribute and single repeated capability id you
 *           may specify. More then one name is listed in comma separated
 *           string.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_GetAttributeRepeatedCapabilityIdNames(
							ViSession instrSession,
							ViUInt32  attributeID,
							ViString  repeatedCapabilityId,
							ViInt32   bufferSize,
							ViChar    _VI_FAR repeatedCapabilityIdNames[])
{
	ViStatus error = VI_SUCCESS;

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, bufferSize, 1, RS_MAX_SHORT_MESSAGE_BUF_SIZE),
			4, "Buffer Size");

	checkErr(RsCore_GetAttributeRepCapNamesAll(instrSession,
        attributeID,
        repeatedCapabilityId,
        bufferSize,
        repeatedCapabilityIdNames));

Error:
	return error;
}

/*****************************************************************************
 * Function: rscmwgm_CheckAttribute<type> Functions
 * Purpose:  These functions enable the instrument driver user to check
 *           attribute values directly.  These functions check the value you
 *           specify even if you set the rscmwgm_ATTR_RANGE_CHECK
 *           attribute to VI_FALSE.  There are typesafe versions for ViInt32,
 *           ViReal64, ViString, ViBoolean, and ViSession datatypes.
 *****************************************************************************/

/* ViInt32 */
ViStatus _VI_FUNC rscmwgm_CheckAttributeViInt32(ViSession instrSession,
                                                ViString  channelName,
                                                ViUInt32  attributeId,
                                                ViInt32   value)
{
	ViStatus error = VI_SUCCESS;

	(void)RsCore_LockSession(instrSession);

	error = RsCore_CheckAttributeViInt32(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/* ViReal64 */
ViStatus _VI_FUNC rscmwgm_CheckAttributeViReal64(ViSession instrSession,
                                                 ViString  channelName,
                                                 ViUInt32  attributeId,
                                                 ViReal64  value)
{
	ViStatus error = VI_SUCCESS;

	(void)RsCore_LockSession(instrSession);

	error = RsCore_CheckAttributeViReal64(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/* ViString */
ViStatus _VI_FUNC rscmwgm_CheckAttributeViString(ViSession instrSession,
                                                 ViString  channelName,
                                                 ViUInt32  attributeId,
                                                 ViString  value)
{
	ViStatus error = VI_SUCCESS;

	(void)RsCore_LockSession(instrSession);

	error = RsCore_CheckAttributeViString(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/* ViBoolean */
ViStatus _VI_FUNC rscmwgm_CheckAttributeViBoolean(ViSession instrSession,
                                                  ViString  channelName,
                                                  ViUInt32  attributeId,
                                                  ViBoolean value)
{
	ViStatus error = VI_SUCCESS;

	(void)RsCore_LockSession(instrSession);

	error = RsCore_CheckAttributeViBoolean(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/* ViSession */
ViStatus _VI_FUNC rscmwgm_CheckAttributeViSession(ViSession instrSession,
                                                  ViString  channelName,
                                                  ViUInt32  attributeId,
                                                  ViSession value)
{
	ViStatus error = VI_SUCCESS;

	(void)RsCore_LockSession(instrSession);

	error = RsCore_CheckAttributeViSession(instrSession, channelName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Reset
 * Purpose:  This function resets the instrument.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_reset(ViSession instrSession)
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	if (RsCore_Simulating(instrSession) == VI_FALSE)
	{
		checkErr(rscmwgm_ClearStatus(instrSession));
		checkErr(RsCore_Write(instrSession, "*RST")); // IGNORE: Check double write
		checkErr(RsCore_QueryViStringShort(instrSession, "*OPC?", NULL));
	}

	checkErr(rscmwgm_DefaultInstrSetup(instrSession));
	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);

	return error;
}

/*****************************************************************************
 * Function: Self Test
 * Purpose:  This function executes the instrument self-test and returns the
 *           result.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_self_test(ViSession instrSession,
                                    ViInt16   *testResult,
                                    ViChar    testMessage[])
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidNullPointer(instrSession, testResult), 2, "Null address for Test Result");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, testMessage), 3, "Null address for Test Message");

	checkErr(RsCore_SelfTest(instrSession, testResult, 48, testMessage));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Self Test Results
 * Purpose:  This function executes the instrument self-test and returns the
 *           result.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_SelfTestResults(ViSession instrSession,
                                          ViInt32   bufferSize,
                                          ViChar    results[])
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	checkErr(RsCore_QueryViStringUnknownLengthToUserBufferWithOpc(instrSession, "DIAG:SERV:STES:RES?", 0, bufferSize, results, NULL));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Revision Query
 * Purpose:  This function returns the driver and instrument revisions.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_revision_query(ViSession instrSession,
                                         ViChar    driverRev[],
                                         ViChar    instrRev[])
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidNullPointer(instrSession, driverRev), 2, "Null address for Instrument Driver Revision");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, instrRev), 3, "Null address for Instrument Revision");

	checkErr(RsCore_GetAttributeViString(instrSession, NULL, RS_ATTR_SPECIFIC_DRIVER_REVISION, 0, 256, driverRev));
	checkErr(RsCore_GetAttributeViString(instrSession, NULL, RS_ATTR_INSTRUMENT_FIRMWARE_REVISION, 0, 256, instrRev));
	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function:    System Version
 * Purpose:     This function queries the number of the SCPI version, which
 *              is relevant for the instrument.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_SystemVersion(ViSession instrSession,
                                        ViInt32   length,
                                        ViChar    systemVersion[])
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(rscmwgm_GetAttributeViString(instrSession, "", RSCMWGM_ATTR_SYSTEM_VERSION,
			length, systemVersion),
			2, "System Version");

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Error Query
 * Purpose:  This function queries the instrument error queue and returns
 *           the result.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_error_query(ViSession instrSession,
                                      ViInt32   *errorCode,
                                      ViChar    errorMessage[])
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidNullPointer(instrSession, errorCode), 2, "Null address for Error Code");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, errorMessage), 3, "Null address for Error Message");

	checkErr(RsCore_ErrorQueryAll(instrSession, errorCode, 1024, errorMessage));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Error List
 * Purpose:  This function reads all system messages and returns a list of
 *           comma separated strings.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ErrorList(ViSession instrSession,
                                    ViInt32   bufferSize,
                                    ViChar    errors[])
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, bufferSize, 1, 32001),
			2, "Buffer Size");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, errors), 3, "Null address for Errors");

	checkErr(rscmwgm_GetAttributeViString( instrSession, "", RSCMWGM_ATTR_SYSTEM_ALL_ERRORS,
        bufferSize, errors));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Error Message
 * Purpose:  This function translates the error codes returned by this
 *           instrument driver into user-readable strings.
 *
 *           Note:  The caller can pass NULL for the vi parameter.  This
 *           is useful if one of the init functions fail.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_error_message(ViSession instrSession,
                                        ViStatus  errorCode,
                                        ViChar    errorMessage[])
{
	ViStatus error = VI_SUCCESS;
	static   RsCoreStringValueTable errorTable =
	         {
	         RSCMWGM_ERROR_CODES_AND_MSGS,
	{0,      NULL}
	         };

	(void)(RsCore_LockSession(instrSession));

	// all VISA and RS error codes are handled as well as codes in the table
	viCheckParm(RsCore_InvalidNullPointer(instrSession, errorMessage), 3, "Null address for Error Message");

	checkErr(RsCore_GetSpecificDriverStatusDesc(instrSession, errorCode, errorMessage, 256, errorTable));

Error:
	(void)RsCore_UnlockSession(instrSession); // IGNORE: ERROR!!! Missing Lock
	return error;
}

/*****************************************************************************
 * Function: Reliability Indicator
 * Purpose:  This function converts a reliability code returned by an
 *           instrument driver function into a user-readable string.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_ReliabilityIndicator(ViSession instrSession,
                                               ViInt32   reliabilityCode,
                                               ViChar    reliabilityMessage[])
{
	ViStatus  error = VI_SUCCESS;
	ViInt16   cnt = 0;
	ViBoolean found = VI_FALSE;

	static RsCoreStringValueTable reliabilityTable =
	{
		{0, "OK"},
		{1, "Measurement Timeout"},
		{2, "Capture Buffer Overflow"},
		{3, "Overdriven"},
		{4, "Underdriven"},
		{6, "Trigger Timeout"},
		{7, "Acquisition Error"},
		{8, "Sync Error"},
		{9, "Uncal"},
		{15, "Reference Frequency Error"},
		{16, "RF Not Available"},
		{17, "RF Level not Settled"},
		{18, "RF Frequency not Settled"},
		{19, "Call not Established"},
		{20, "Call Type not Usable"},
		{21, "Call Lost"},
		{23, "Missing Option"},
		{24, "Invalid RF Setting"},
		{26, "Resource Conflict"},
		{27, "No Sensor Connected"},
		{28, "Unexpected Parameter Change"},
		{30, "File not Found"},
		{31, "No DTM reply"},
		{32, "ACL Disconnected"},
		{40, "ARB File CRC Error"},
		{42, "ARB Header Tag Invalid"},
		{43, "ARB Segment Overflow"},
		{44, "ARB File not Found"},
		{45, "ARB Memory Overflow"},
		{46, "ARB Sample Rate out of Range"},
		{47, "ARB Cycles out of Range"},
		{50, "Startup Error"},
		{51, "No Reply"},
		{52, "Connection Error"},
		{53, "Configuration Error"},
		{54, "Filesystem Error"},
		{60, "Invalid RF-Connector Setting"},
		{93, "OCXO Oven Temperature too low"},
		{101, "Firmware Error"},
		{102, "Unidentified Error"},
		{103, "Parameter Error"},
		{104, "Not Functional"},
	};

	(void)RsCore_LockSession(instrSession);

	for (cnt = 0; reliabilityTable[cnt].string; cnt++)
	{
		if (reliabilityTable[cnt].value == reliabilityCode)
		{
			sprintf(reliabilityMessage, "%s", reliabilityTable[cnt].string);
			if (reliabilityCode)
				error = RSCMWGM_ERROR_RELIABILITY_INDICATOR;
			found = VI_TRUE;
		}
	}

	if (!found)
	{
		sprintf(reliabilityMessage, "Unrecognized reliability indicator! Report problem to R&S support!");
		error = RS_WARN_UNKNOWN_INSTRUMENT_STATUS;
	}

	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Preset
 * Purpose:  This function initiates an instrument reset.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_Preset(ViSession instrSession)
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_SetAttributeViString(instrSession, "", RSCMWGM_ATTR_SYSTEM_PRESET, NULL));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN This function sets the OPC timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR timeout/Timeout value in ms
ViStatus _VI_FUNC rscmwgm_SetTimeout(ViSession instrSession,
                                     ViInt32   timeout)
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(rscmwgm_SetAttributeViInt32(instrSession, "", RS_ATTR_OPC_TIMEOUT, timeout),
			3, "Timeout");

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN This function gets the OPC timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or rscmwgm_InitWithOptions function.
/// HIPAR timeout/Timeout value in ms
ViStatus _VI_FUNC rscmwgm_GetTimeout(ViSession instrSession,
                                     ViInt32   *timeout)
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(rscmwgm_GetAttributeViInt32(instrSession, "", RS_ATTR_OPC_TIMEOUT, timeout),
			3, "Timeout");

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  This function sets the OPC timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR opcTimeout/This control sets the timeout value.
ViStatus _VI_FUNC rscmwgm_SetOPCTimeout(ViSession instrSession,
                                        ViUInt32  opcTimeout)
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_SetAttributeViInt32(instrSession, NULL, RS_ATTR_OPC_TIMEOUT, 0, (ViInt32) opcTimeout));

Error:
	return error;
}

/// HIFN  This function returns the OPC timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR timeout/This control returns the timeout value.
ViStatus _VI_FUNC rscmwgm_GetOPCTimeout(ViSession instrSession,
                                        ViUInt32  *timeout)
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_GetAttributeViInt32(instrSession, NULL, RS_ATTR_OPC_TIMEOUT, 0, (ViInt32*) timeout));

Error:
	return error;
}

/// HIFN This function specifies the minimum timeout value to use (in
/// HIFN milliseconds) when accessing the device associated with the
/// HIFN given session.
/// HIFN Note that the actual timeout used may be higher than the one
/// HIFN requested.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR VISATimeout/Specifies the minimum timeout value to use (in milliseconds)
/// HIPAR VISATimeout/when accessing the device associated with the given session.
/// HIPAR VISATimeout/Note that the actual timeout used may be higher than the one
/// HIPAR VISATimeout/requested.
ViStatus _VI_FUNC rscmwgm_SetVISATimeout(ViSession instrSession,
                                         ViUInt32  VISATimeout)
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	checkErr(RsCore_SetVisaTimeout(instrSession, VISATimeout));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN This function queries the minimum timeout value to use (in
/// HIFN milliseconds) when accessing the device associated with the
/// HIFN given session.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR VISATimeout/Queries the minimum timeout value to use (in milliseconds)
/// HIPAR VISATimeout/when accessing the device associated with the given session.
ViStatus _VI_FUNC rscmwgm_GetVISATimeout(ViSession instrSession,
                                         ViUInt32  *VISATimeout)
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	checkErr(RsCore_GetVisaTimeout(instrSession, VISATimeout));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  This function clears status.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
ViStatus _VI_FUNC rscmwgm_ClearStatus(ViSession instrSession)
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	checkErr(RsCore_ClearStatus(instrSession));
	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  This function returns the ID Query response string.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR bufferSize/Size of IDQueryResponse buffer
/// HIPAR idQueryResponse/Returns the ID Query response string. The array should consist of at
/// HIPAR idQueryResponse/least 256 elements.
ViStatus _VI_FUNC rscmwgm_IDQueryResponse(ViSession instrSession,
                                          ViUInt32  bufferSize,
                                          ViChar    IDQueryResponse[])
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_QueryViString(instrSession, "*IDN?", bufferSize, IDQueryResponse));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  Stops further command processing until all commands sent before *WAI
/// HIFN  have been executed.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
ViStatus _VI_FUNC rscmwgm_ProcessAllPreviousCommands(ViSession instrSession)
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_WriteInstrData(instrSession, "*WAI"));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN  This function queries the OPC.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR opc/Queries the OPC.
ViStatus _VI_FUNC rscmwgm_QueryOPC(ViSession instrSession,
                                   ViInt32   *opc)
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	checkErr(rscmwgm_QueryViInt32(instrSession, "*OPC?", opc));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN This function switches ON/OFF various error checkings performed inside of the driver.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR optionChecking/This control switches option checking On or Off.
/// HIPAR rangeChecking/This control enables or disables range checking.
/// HIPAR statusChecking/This control enables or disables instrument state checking
ViStatus _VI_FUNC rscmwgm_ConfigureErrorChecking(ViSession instrSession,
                                                 ViBoolean optionChecking,
                                                 ViBoolean rangeChecking,
                                                 ViBoolean statusChecking)
{
	ViStatus         error = VI_SUCCESS;
	RsCoreSessionPtr rsSession = NULL;

	checkErr(RsCore_GetRsSession(instrSession, &rsSession));
	rsSession->optionChecking = optionChecking;
	checkErr(RsCore_SetAttributeViBoolean(instrSession, "", RS_ATTR_QUERY_INSTRUMENT_STATUS, 0, statusChecking));
	checkErr(RsCore_SetAttributeViBoolean(instrSession, "", RS_ATTR_RANGE_CHECK, 0, rangeChecking));

Error:
	return error;
}

/// HIFN If TRUE (default after init is FALSE), the driver calls SYST:ERR? automatically in CheckStatus() when Instrument Status Error is detected.
/// HIFN Use the function rscmwgm_GetError afterwards to get the detailed information about the instrument error(s).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR autoSystErrQuery/If TRUE (default after init is FALSE), the driver calls SYST:ERR? automatically in CheckStatus() when Instrument Status Error is detected.
ViStatus _VI_FUNC rscmwgm_ConfigureAutoSystemErrQuery(ViSession instrSession,
                                                      ViBoolean autoSystErrQuery)
{
	ViStatus         error = VI_SUCCESS;
	RsCoreSessionPtr rsSession = NULL;

	checkErr(RsCore_GetRsSession(instrSession, &rsSession));
	rsSession->autoSystErrQuery = autoSystErrQuery;

Error:
	return error;
}

/// HIFN Session-related mutex switch.
/// HIFN If TRUE (default after init is FALSE), the driver sets a mutual-exclusion mechanism for every driver's hi-level function and attribute access.
/// HIFN Set this value to TRUE if you plan to use one session in more than one thread.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rscmwgm_init or
/// HIPAR instrSession/rscmwgm_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR multiThreadLocking/Session-related mutex switch
ViStatus _VI_FUNC rscmwgm_ConfigureMultiThreadLocking(ViSession instrSession,
                                                      ViBoolean multiThreadLocking)
{
	ViStatus         error = VI_SUCCESS;
	RsCoreSessionPtr rsSession = NULL;

	checkErr(RsCore_GetRsSession(instrSession, &rsSession));
	rsSession->multiThreadLocking = multiThreadLocking;

Error:
	return error;
}

/*****************************************************************************
 * Function: Get Error and Clear Error Functions
 * Purpose:  These functions enable the instrument driver user to
 *           get or clear the error information the driver associates with the
 *           RS session.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_GetError(ViSession instrSession,
                                   ViStatus  *errorCode,
                                   ViInt32   bufferSize,
                                   ViChar    description[])
{
	ViStatus error = VI_SUCCESS;

	// Lock Session - do not jump to the end, even with invalid instrSession the function continues further
	(void)(RsCore_LockSession(instrSession));

	// Test for nulls and acquire error data
	viCheckParm(RsCore_InvalidNullPointer(instrSession, errorCode), 2, "Null address for Error");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, description), 4, "Null address for Description");

	checkErr(RsCore_GetErrorCompleteDescription(instrSession, &rscmwgm_error_message, errorCode, bufferSize, description)
    );

Error:
	// Unlock Session
	// IGNORE: ERROR!!! Missing Lock
	return error;
}

ViStatus _VI_FUNC rscmwgm_ClearError(ViSession instrSession)
{
	return RsCore_ClearErrorInfo(instrSession);
}

/// HIFN This function is used to read data from the instrument and write
/// HIFN it to a user specified file on the host computer.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR source/This control selects file for the data transfer from instrument
/// HIPAR source/to control computer.
/// HIPAR destination/This control defines destination file to which the data transfer
/// HIPAR destination/from instrument to control computer applies.
ViStatus _VI_FUNC rscmwgm_ReadToFileFromInstrument(ViSession instrSession,
                                                   ViString  source,
                                                   ViString  destination)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

	checkErr(RsCore_LockSession(instrSession));

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "MMEM:DATA? '%s'", source);
	checkErr(RsCore_QueryBinaryDataBlockToFile(instrSession, cmd, destination, RS_VAL_TRUNCATE));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN This function is used to read data from the host computer and
/// HIFN write it to a user specified file in the instrument.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR source/This control selects file for the data transfer from control
/// HIPAR source/computer to instrument.
/// HIPAR destination/This control defines destination file to which the data transfer
/// HIPAR destination/from control computer to instrument applies.
ViStatus _VI_FUNC rscmwgm_WriteFromFileToInstrument(ViSession instrSession,
                                                    ViString  source,
                                                    ViString  destination)
{
	ViStatus error = VI_SUCCESS;
	ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

	checkErr(RsCore_LockSession(instrSession));

	snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "MMEM:DATA '%s',", destination);
	checkErr(RsCore_WriteBinaryDataFromFile(instrSession, cmd, source));
	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: WriteInstrData and ReadInstrData Functions
 * Purpose:  These functions enable the instrument driver user to
 *           write and read commands directly to and from the instrument.
 *
 *           Note:  These functions bypass the RS attribute state caching.
 *                  WriteInstrData invalidates the cached values for all
 *                  attributes.
 *****************************************************************************/

ViStatus _VI_FUNC rscmwgm_WriteInstrData(ViSession instrSession,
                                         ViString  writeBuffer)
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	error = RsCore_Write(instrSession, writeBuffer);

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

ViStatus _VI_FUNC rscmwgm_ReadInstrData(ViSession instrSession,
                                        ViInt32   numBytes,
                                        ViChar    rdBuf[],
                                        ViUInt32  *bytesRead)
{
	ViStatus error = VI_SUCCESS;
	ViChar   *responseString = NULL;

	checkErr(RsCore_LockSession(instrSession));

	checkErr(RsCore_ReadViStringUnknownLength(instrSession, &responseString));
	checkErr(RsCore_CopyToUserBufferAsciiData(rdBuf, numBytes, responseString));
	*bytesRead = (ViUInt32)strlen(responseString);

Error:
	if (responseString)
		free(responseString);

	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN Sends a command to the instrument synchronised with OPC-polling.
/// HIFN Use this function to wait for an instrument operation that can
/// HIFN take a long time to complete, e.g. self-alignment.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR writeBuffer/Pass the string to be written to the instrument.
/// HIPAR timeout/This control sets the timeout value.
ViStatus _VI_FUNC rscmwgm_WriteCommandWithOPCSync(ViSession instrSession,
                                                  ViString  writeBuffer,
                                                  ViUInt32  timeout)
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	checkErr(RsCore_WriteWithOpc(instrSession, writeBuffer, timeout));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/// HIFN Sends a query to the instrument synchronised with OPC-polling.
/// HIFN Use this function to wait for a response that can take a long
/// HIFN time to complete e.g. self-alignment, calibration or self-test.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR writeBuffer/Pass the string to be written to the instrument.
/// HIPAR timeout/This control sets the timeout value.
/// HIPAR bufferSize/Pass the maximum number of bytes to read from the instruments.
/// HIPAR readBuffer/After this function executes, this parameter contains the data
/// HIPAR readBuffer/that was read from the instrument.
/// HIPAR numBytesRead/Returns the number of bytes actually read from the instrument
/// HIPAR numBytesRead/and stored in the Read Buffer.
ViStatus _VI_FUNC rscmwgm_QueryWithOPCSync(ViSession instrSession,
                                           ViString  writeBuffer,
                                           ViUInt32  timeout,
                                           ViInt32   bufferSize,
                                           ViChar    readBuffer[],
                                           ViUInt32  *numBytesRead)
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, (ViInt32)strlen(writeBuffer), 1, 2147483647),
    	2, "Write Buffer (string length)");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, readBuffer),
    	5, "Read Buffer");

	checkErr(RsCore_QueryViStringUnknownLengthToUserBufferWithOpc(instrSession, writeBuffer, timeout, bufferSize, readBuffer, (ViInt32*)numBytesRead));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Query ViBoolean
 * Purpose:  This function queries the ViBoolean value.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_QueryViBoolean(ViSession instrSession,
                                         ViString  cmd,
                                         ViBoolean *value)
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, (ViInt32)strlen(cmd), 1, 2147483647), 2,
        "Command (null string length)");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, value), 3, "Value");

	checkErr(RsCore_QueryViBoolean(instrSession, cmd, value));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Query ViInt32
 * Purpose:  This function queries the ViInt32 value.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_QueryViInt32(ViSession instrSession,
                                       ViString  cmd,
                                       ViInt32   *value)
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, (ViInt32)strlen(cmd), 1, 2147483647), 2,
        "Command (null string length)");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, value), 3, "Value");

	checkErr(RsCore_QueryViInt32(instrSession, cmd, value));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: Query ViReal64
 * Purpose:  This function queries the ViReal64 value.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_QueryViReal64(ViSession instrSession,
                                        ViString  cmd,
                                        ViReal64  *value)
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, (ViInt32)strlen(cmd), 1, 2147483647), 2,
        "Command (null string length)");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, value), 3, "Value");

	checkErr(RsCore_QueryViReal64(instrSession, cmd, value));

	checkErr(rscmwgm_CheckStatus(instrSession));

Error:
	(void)RsCore_UnlockSession(instrSession);
	return error;
}


/*****************************************************************************
 * Function: Query ViString
 * Purpose:  This function queries the ViString value.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_QueryViString(ViSession instrSession,
                                        ViString  cmd,
                                        ViInt32   bufferSize,
                                        ViChar    _VI_FAR value[])
{
	ViStatus error = VI_SUCCESS;
	ViChar   *pbuffer = NULL;

	checkErr(RsCore_LockSession(instrSession));

	viCheckParm(RsCore_InvalidViInt32Range(instrSession, (ViInt32)strlen(cmd), 1, 2147483647), 2,
        "Command (null string length)");
	viCheckParm(RsCore_InvalidNullPointer(instrSession, value), 3, "Value");

	checkErr(RsCore_QueryViStringUnknownLength(instrSession, cmd, &pbuffer));

	checkErr(rscmwgm_CheckStatus(instrSession));

	checkErr(RsCore_CopyToUserBufferAsciiData(value, bufferSize, pbuffer));

Error:
	if (pbuffer)
		free(pbuffer);

	(void)RsCore_UnlockSession(instrSession);
	return error;
}

/*****************************************************************************
 * Function: rscmwgm_close
 * Purpose:  This function closes the instrument.
 *
 *           Note:  This function must unlock the session before calling
 *           Rs_Dispose.
 *****************************************************************************/
ViStatus _VI_FUNC rscmwgm_close(ViSession instrSession)
{
	ViStatus error = VI_SUCCESS;

	checkErr(RsCore_LockSession(instrSession));

	checkErr(RsCore_ViClose(instrSession));

Error:
	(void)RsCore_Dispose(instrSession);
	(void)RsCore_UnlockSession(instrSession);
	return error;
}
